/*
 * File         : ExportedClassNameManager.java
 * Author       : Han-Ku Lee, 
 * Created      : Thu May 14 12:03:02 1998
 * Revision     : $Revision: 1.12 $
 * Updated      : $Date: 1999/11/11 01:28:43 $
 */

package hpjava.compiler.abstractexp;
 
import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;
 
import hpjava.compiler.reflection.*;

import java.util.*;
import java.io.*;

public abstract class AbstractExpression extends AbstractNode {
  
  abstract public void accept(AbstractExpressionVisitor v);

  public final int CONDITIONAL_OR = 0;

  public final int CONDITIONAL_AND = 1;

  public final int INCLUSIVE_OR = 2;

  public final int EXCLUSIVE_OR = 3;

  public final int AND = 4;

  public final int EQUALITY = 5;

  public final int INEQUALITY = 6;

  public final int INSTANCE_OF = 7;

  public final int GREATER = 8;

  public final int LESS = 9;

  public final int GREATER_THAN_EQ = 10;

  public final int LESS_THAN_EQ = 11;

  public final int LEFT_SHIFT = 12;

  public final int RIGHT_SHIFT = 13;

  public final int EXTENSION_RIGHT_SHIFT = 14;

  public final int ADD = 15;

  public final int SUBTRACT = 16;

  public final int MULTIPLE = 17;

  public final int DIVIDE = 18;

  public final int MODULE = 19;

  public final int PLUS = 20;

  public final int MINUS = 21;

  public final int PRE_INCREMENT = 22;

  public final int PRE_DECREMENT = 23;

  public final int COMPLEMENT = 24;

  public final int LOGICAL_NOT = 25;

  public final int POSTFIX = 26;

  public final int CAST = 27;

  public final int GROUP = 28;

  public final int POSTFIX_ADDITION = 29;

  public final int POSTFIX_SUBTRACTION = 30;

  public final int POSTFIX_INDEX_LOCATION = 31;

  
  private AbstractExpression parent = null;
  private Expression expression = null;
  private StatementExpression statement_expression = null;
  private PrimaryExpression primary_expression = null;

  /* protected */ public void setParent(AbstractExpression parent){
    this.parent = parent;
  }

  public AbstractExpression getParent(){
    return parent;
  }
  
  public void setExpression(Expression exp){
    expression = exp;
  }
  
  public void setStatementExpression(StatementExpression sexp){
    statement_expression = sexp;
  }
  
  public void setPrimaryExpression(PrimaryExpression pexp){
    primary_expression = pexp;
  }
  
  public AssignmentNode assignment(AbstractExpression r){
    //return new AssignmentNode(this, r);
    AssignmentNode assign = new AssignmentNode(this, r);

    this.setParent(assign);
    r.setParent(assign);
    
    return assign;  
  }

  public MultiplicationAssignmentNode multiplicationAssignment(AbstractExpression r){
    //return new MultiplicationAssignmentNode(this, r);
    MultiplicationAssignmentNode multiplication = new MultiplicationAssignmentNode(this, r);

    this.setParent(multiplication);
    r.setParent(multiplication);
    
    return multiplication;  
  }

  public DivisionAssignmentNode divisionAssignment(AbstractExpression r){
    //return new DivisionAssignmentNode(this, r);
    DivisionAssignmentNode division = new DivisionAssignmentNode(this, r);

    this.setParent(division);
    r.setParent(division);
    
    return division;    
  }

  public ModulusAssignmentNode modulusAssignment(AbstractExpression r){
    //return new ModulusAssignmentNode(this, r);
    ModulusAssignmentNode modulus = new ModulusAssignmentNode(this, r);

    this.setParent(modulus);
    r.setParent(modulus);
    
    return modulus;
  }

  public AdditionAssignmentNode additionAssignment(AbstractExpression r){
    //return new AdditionAssignmentNode(this, r);
    AdditionAssignmentNode addition = new AdditionAssignmentNode(this, r);

    this.setParent(addition);
    r.setParent(addition);
    
    return addition;
  }

  public SubtractionAssignmentNode subtractionAssignment(AbstractExpression r){
    //return new SubtractionAssignmentNode(this, r);
    SubtractionAssignmentNode subtraction = new SubtractionAssignmentNode(this, r);

    this.setParent(subtraction);
    r.setParent(subtraction);
    
    return subtraction;
  }

  public LeftShiftAssignmentNode leftShiftAssignment(AbstractExpression r){
    //return new LeftShiftAssignmentNode(this, r);
    LeftShiftAssignmentNode leftShift = new LeftShiftAssignmentNode(this, r);
    
    this.setParent(leftShift); 
    r.setParent(leftShift);
    
    return leftShift;        
  }

  public RightShiftAssignmentNode rightShiftAssignment(AbstractExpression r){
    //return new RightShiftAssignmentNode(this, r);
    RightShiftAssignmentNode rightShift = new RightShiftAssignmentNode(this, r);
    
    this.setParent(rightShift); 
    r.setParent(rightShift);
    
    return rightShift;        
  }

  public ExtensionRightShiftAssignmentNode extensionRightShiftAssignment(AbstractExpression r){
    //return new ExtensionRightShiftAssignmentNode(this, r);
    ExtensionRightShiftAssignmentNode extensionRightShift = new ExtensionRightShiftAssignmentNode(this, r);
    
    this.setParent(extensionRightShift); 
    r.setParent(extensionRightShift);
    
    return extensionRightShift;        
  }

  public AndAssignmentNode andAssignment(AbstractExpression r){
    //return new AndAssignmentNode(this, r);
    AndAssignmentNode and = new AndAssignmentNode(this, r);
    
    this.setParent(and); 
    r.setParent(and);
    
    return and;        
  }

  public ExclusiveOrAssignmentNode exclusiveOrAssignment(AbstractExpression r){
    //return new ExclusiveOrAssignmentNode(this, r);
    ExclusiveOrAssignmentNode exclusiveOr = new ExclusiveOrAssignmentNode(this, r);

    this.setParent(exclusiveOr); 
    r.setParent(exclusiveOr);
    
    return exclusiveOr;    
  }

  public InclusiveOrAssignmentNode inclusiveOrAssignment(AbstractExpression r){
    //return new InclusiveOrAssignmentNode(this, r);
    InclusiveOrAssignmentNode inclusiveOr = new InclusiveOrAssignmentNode(this, r);
    this.setParent(inclusiveOr); 
    r.setParent(inclusiveOr);
    
    return inclusiveOr;
  }

  public ConditionalNode conditional(AbstractExpression r1, AbstractExpression r2) {
    //return new ConditionalNode(this, r1, r2);
    ConditionalNode condition = new ConditionalNode(this, r1, r2);

    this.setParent(condition); 
    r1.setParent(condition);
    r2.setParent(condition);
   
    return condition;
  }
  
  public CastNode cast(TypeNode t){
    //return new CastNode(this, t);

    CastNode _cast = new CastNode(this, t);

    this.setParent(_cast);
    ((AbstractExpression)t).setParent(_cast);
    return _cast;
  }

  //public PostfixSubtractionNode postfixSubtraction(){
  //  return new PostfixSubtractionNode(this);
  //}

  public BinExprNode binExpr(int bOperator, AbstractExpression rightOp){
    //return new BinExprNode ( bOperator, this , rightOp );
    BinExprNode binary = new BinExprNode (bOperator, this , rightOp);
    this.setParent(binary);
    rightOp.setParent(binary);

    return binary;
  }

  public UnaryExprNode unaryExpr(int uOperator ) {
    //return new UnaryExprNode( uOperator, this );
    UnaryExprNode unary = new UnaryExprNode(uOperator, this);
    this.setParent(unary);

    return unary;
  }

  public MethodCallNode methodCall(IdentifierNode methodId, AbstractExpression argList){
    //return new MethodCallNode( this, methodId, argList );
    MethodCallNode method = new MethodCallNode(this, methodId, argList);
    
    this.setParent(method);
    methodId.setParent(method);
    argList.setParent(method);

    return method;
  }

  public ArraySectNode arraySect(DistArraySecRefDimListNode dimList ){
    //return new ArraySectNode(this, dimList);
    ArraySectNode arraySect = new ArraySectNode(this, dimList);

    this.setParent(arraySect);
    dimList.setParent(arraySect);

    return arraySect;
  }

  public ArrayElemRefNode arrayElemRef(TripletOrExpressionListNode eList){
    //return new ArraySectNode(this, dimList);
    ArrayElemRefNode arrayElemRef = new ArrayElemRefNode(this, eList);

    this.setParent(arrayElemRef);
    eList.setParent(arrayElemRef);

    return arrayElemRef;
  }

  public ClassFieldNode classField (IdentifierNode fieldName ) {
    //return new ClassFieldNode(this, fieldName);
    ClassFieldNode field = new ClassFieldNode(this, fieldName);

    this.setParent(field);
    fieldName.setParent(field);

    return field;
  }

  // Han-Ku Lee added two methods inner_this and inner_allocationExpr

  public InnerThisNode inner_this(ThisNode fieldName){
    //return new InnerThisNode(this, fieldName);
    InnerThisNode innerThis = new InnerThisNode(this, fieldName);

    this.setParent(innerThis);
    fieldName.setParent(innerThis);

    return innerThis;
  }

  public InnerAllocationExpressionNode inner_allocationExpr(AllocationExpressionNode fieldName){
    //return new InnerAllocationExpressionNode(this, fieldName);
    InnerAllocationExpressionNode innerAllocationExpressionNode = 
      new InnerAllocationExpressionNode(this, fieldName);
    
    this.setParent(innerAllocationExpressionNode);
    fieldName.setParent(innerAllocationExpressionNode);

    return innerAllocationExpressionNode;
  }
  /*
  public Expression toExpression() {
    Expression empty = new Expression(null, new NodeOptional());

    ExpressionManager.register(empty, this);

    return empty;
  }

  public StatementExpression toStatementExpression() {
    StatementExpression empty = new StatementExpression(new NodeChoice(null, -1));

    ExpressionManager.register(empty, this);

    return empty;
  }
  */
  public Expression toExpression() {
    return expression;
  }

  public StatementExpression toStatementExpression() {
    return statement_expression;
  }

  public PrimaryExpression toPrimaryExpression() {
    return primary_expression;
  }

  public void typeChecking() {
    // the function is also called from manager directly ...

    if (!this.typeChecked) {
      AbstractExpressionInspector inspector = new AbstractExpressionInspector();

      this.accept(inspector);
    }
  }

  public void replaceWith( AbstractExpression n ) {
    // if(this.parent == null )
    //   System.out.println("Testing parent: null !");

    if (parent instanceof AssignmentNode ) {
      if (((AssignmentNode)parent).left == this ) {
        ((AssignmentNode)parent).left = n;
      } 
      else if (((AssignmentNode)parent).right == this ) {
        ((AssignmentNode)parent).right = n ;
      } 
      else { throw new Error(); }
    } 
    else if ( parent instanceof BinExprNode ) {
      if (((BinExprNode)parent).left == this) {
        ((BinExprNode)parent).left = n;
      } else if (((BinExprNode)parent).right == this) {
        ((BinExprNode)parent).right = n;
      } else { throw new Error(); }
    } 
    else if ( parent instanceof UnaryExprNode ) {
      if (((UnaryExprNode)parent).operand == this) {
        ((UnaryExprNode)parent).operand = n;
      } else { throw new Error(); }
    } 
    else if ( parent instanceof ArraySectNode ) {
      if (((ArraySectNode)parent).arrayName == this) {
        ((ArraySectNode)parent).arrayName = n;
      } else if (((ArraySectNode)parent).dimList == this) {
        ((ArraySectNode)parent).dimList = (DistArraySecRefDimListNode)n;
      } else { throw new Error(); }
    } 
    else if ( parent instanceof ArrayElemRefNode ) {
      if (((ArrayElemRefNode)parent).arrayName == this) {
        ((ArrayElemRefNode)parent).arrayName = n;
      } else if (((ArrayElemRefNode)parent).eList == this) {
        ((ArrayElemRefNode)parent).eList = (TripletOrExpressionListNode) n;
      } else { throw new Error(); }
    } 
    else if ( parent instanceof ClassFieldNode ) {
      if (((ClassFieldNode)parent).objName == this) {
        ((ClassFieldNode)parent).objName = n;
      } else if (((ClassFieldNode)parent).fieldName == this) {
        ((ClassFieldNode)parent).fieldName = (IdentifierNode)n;
      } else { throw new Error(); }
    } 
    else if ( parent instanceof MethodCallNode ) {
      if (((MethodCallNode)parent).objName == this) {
        ((MethodCallNode)parent).objName = n;
      } else if (((MethodCallNode)parent).methodName == this) {
        ((MethodCallNode)parent).methodName = (IdentifierNode)n;
      } else if (((MethodCallNode)parent).args == this) {
        ((MethodCallNode)parent).args = n;
      } else { throw new Error(); }
    } 
    else if ( parent instanceof ArgumentsNode ) {
    
    }
    else if( expression != null && n instanceof ArrayElemRefNode){
      //System.out.println("........ " + this.expression.getClass().getName());
      ExpressionManager.register(this.expression, n);
    }
    else if( expression != null && n instanceof AllocationExpressionNode){
      //System.out.println("........ " + this.expression.getClass().getName());
      ExpressionManager.register(this.expression, n);
    }
    
  }

  public boolean isTypeChecked() {
    return typeChecked;
  }
  
}






