/*
 * File         : AbstractExpressionVisitor.java
 * Author       : Han-Ku Lee
 * Created      : Fri Oct  9 15:20:09 1998
 * Revision     : $Revision: 1.10 $
 * Updated      : $Date: 1999/11/11 02:49:11 $
 */

//
// Generated by JTB 1.1pre2
//

package hpjava.compiler.abstractexp;

import hpjava.compiler.syntaxtree.*;
import java.util.*;

public class AbstractExpressionVisitor {

   // f0 -> "="
   //       | "*="
   //       | "/="
   //       | "%="
   //       | "+="
   //       | "-="
   //       | "<<="
   //       | ">>="
   //       | ">>>="
   //       | "&="
   //       | "^="
   //       | "|="
   //
   public void visit(AssignmentNode n) {
      n.left.accept(this);
      n.right.accept(this);
   }

   public void visit(AdditionAssignmentNode n) {
      n.left.accept(this);
      n.right.accept(this);
   }

   public void visit(SubtractionAssignmentNode n) {
      n.left.accept(this);
      n.right.accept(this);
   }

   public void visit(MultiplicationAssignmentNode n) {
      n.left.accept(this);
      n.right.accept(this);
   }

   public void visit(DivisionAssignmentNode n) {
      n.left.accept(this);
      n.right.accept(this);
   }

   public void visit(AndAssignmentNode n) {
      n.left.accept(this);
      n.right.accept(this);
   }

   public void visit(InclusiveOrAssignmentNode n) {
      n.left.accept(this);
      n.right.accept(this);
   }

   public void visit(ExclusiveOrAssignmentNode n) {
      n.left.accept(this);
      n.right.accept(this);
   }

   public void visit(LeftShiftAssignmentNode n) {
      n.left.accept(this);
      n.right.accept(this);
   }

   public void visit(RightShiftAssignmentNode n) {
      n.left.accept(this);
      n.right.accept(this);
   }

   public void visit(ModulusAssignmentNode n) {
      n.left.accept(this);
      n.right.accept(this);
   }

   public void visit(ExtensionRightShiftAssignmentNode n) {
      n.left.accept(this);
      n.right.accept(this);
   }

   //
   // f0 -> ConditionalOrExpression()
   // f1 -> [ "?" Expression() ":" ConditionalExpression() ]
   //
   public void visit(ConditionalNode n) {
      n.condition.accept(this);
      n.trueResult.accept(this);
      n.falseResult.accept(this);
   }

   //
   // Following parts are added by Qiang Zheng
   // 
   //
   // f0 -> Operator
   // f1 -> AbstractExpression()
   // f2 -> AbstractExpression()
   //
   public void visit(BinExprNode n) {
      n.left.accept(this);
      n.right.accept(this);
   }

   public void visit(UnaryExprNode n) {
      n.operand.accept(this);
   }

   public void visit (ArraySectNode n) {
	  n.arrayName.accept(this);
	  n.dimList.accept(this);
   }

   public void visit (MethodCallNode n) {
	  n.objName.accept(this);
	  n.methodName.accept(this);
	  n.args.accept(this);
   }

   public void visit (ClassFieldNode n) {
	  n.objName.accept(this);
	  n.fieldName.accept(this);
   }

   public void visit (ArrayElemRefNode n) {
	  n.arrayName.accept(this);
	  n.eList.accept(this);
   }

   // Added by Qiang Zheng
   public void visit ( AllocatedShapeNode n ) {
     // Do nothing
   }
 
   // Han-Ku Lee added two methods, InnerThisNode and InnerAllocationExpressionNode

   public void visit (InnerThisNode n) {
	  n.objName.accept(this);
	  n.fieldName.accept(this);
   }
  
   public void visit (InnerAllocationExpressionNode n) {
	  n.objName.accept(this);
	  n.fieldName.accept(this);
   }

   public void visit (IdentifierNode n) {
	  // do nothing
   }

   public void visit (AllocationExpressionNode n) {
	  // do nothing
   }

   public void visit (LiteralNode n) {
	  // do nothing
   }

   public void visit (ExpressionNode n) {
	  // do nothing
   }

   public void visit (SuperNode n) {
	  // do nothing
   }

   public void visit (ThisNode n) {
	  // do nothing
   }

   public void visit (NullNode n) {
	  // do nothing
   }

   public void visit (ClassLiteralNode n) {
	  // do nothing
   }

   public void visit (ArgumentsNode n) {
	  // do nothing
   }

   public void visit (DistArraySecRefDimListNode n) {
	  // do nothing
   }

   public void visit (TripletOrExpressionListNode n) {
	  // do nothing
   }

   // Above are added by Qiang Zheng

   //
   // f0 -> "(" Type() ")" UnaryExpression()
   //       | "(" Type() ")" UnaryExpressionNotPlusMinus()
   //
   public void visit(CastNode n) {
      n.type.accept(this);
      n.leaf.accept(this);
   }

   //
   // f0 -> PrimaryExpression()
   // f1 -> ( "/" PrimaryExpression() )*
   //
   public void visit(GroupNode n) {
      n.left.accept(this);
      n.right.accept(this);
   }

   public void visit(TypeNode n) {
     //   n.leaf.accept(this);
   }

}

