/*
 * File         : ExpressionConversion_abstractexp.java
 * Author       : Han-Ku Lee
 * Created      : Sat Aug 14 16:52:42 1999
 * Revision     : $Revision: 1.5 $
 * Updated      : $Date: 1999/11/11 01:28:46 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1999
 */

package hpjava.compiler.abstractexp; 
 
import hpjava.compiler.util.*;  
import hpjava.compiler.syntaxtree.*;  
import hpjava.compiler.reflection.*; 
import hpjava.compiler.wrapper.*;
 
public class ExpressionConversion_abstractexp {

  protected String BOOLEAN = "boolean";
  protected String BYTE = "byte";
  protected String CHAR = "char";
  protected String SHORT = "short";
  protected String INT = "int";
  protected String LONG = "long";
  protected String FLOAT = "float";
  protected String DOUBLE = "double";
  protected String VOID = "void";

  public String binary_numeric_promotion(ClassReflector left, ClassReflector right){
    if(left.getName().equals(DOUBLE) || right.getName().equals(DOUBLE))
      return DOUBLE;
    else if(left.getName().equals(FLOAT) || right.getName().equals(FLOAT))
      return FLOAT;
    else if(left.getName().equals(LONG) || right.getName().equals(LONG))
      return LONG;

    return INT;
  }

  public String string_conversion(ClassReflector left, ClassReflector right){
    if(left.getName().equals("java.lang.String") 
       || right.getName().equals("java.lang.String"))
      return "java.lang.String";
    else
      return binary_numeric_promotion(left, right);
  }

  public String unary_numeric_promotion(String operand){
    return operand;
  }

  public boolean widening_primitive_conversion(ClassReflector left, ClassReflector right){
    boolean flag = false;

    if(left.getName().equals(BYTE)){
      if(right.getName().equals(SHORT)
         || right.getName().equals(INT)
         || right.getName().equals(LONG)
         || right.getName().equals(FLOAT)
         || right.getName().equals(DOUBLE))
        flag = true;
    }
    else if(left.getName().equals(SHORT) ||left.getName().equals(CHAR)){
      if(right.getName().equals(INT)
         || right.getName().equals(LONG)
         || right.getName().equals(FLOAT)
         || right.getName().equals(DOUBLE))
        flag = true;
    }
    else if(left.getName().equals(INT)){
      if(right.getName().equals(LONG)
         || right.getName().equals(FLOAT)
         || right.getName().equals(DOUBLE))
        flag = true;
    }
    else if(left.getName().equals(LONG)){
      if(right.getName().equals(FLOAT)
         || right.getName().equals(DOUBLE))
        flag = true;
    }
    else if(left.getName().equals(FLOAT)){
      if(right.getName().equals(DOUBLE))
        flag = true;
    }

    return flag;
  }

  public String cast_widening_narrowing_conversion(CastNode n){
    //String converted_type = new String();
    String converted_type = null;

    if(n.type.leaf.f0.which == 0 && n.type.leaf.f1.size() == 0){
      //NodeChoice nc = (NodeChoice)n.type.leaf.f0.choice;
      PrimitiveType nc = (PrimitiveType)n.type.leaf.f0.choice;

      //System.out.println("      nc.f0.which: " + nc.f0.which);
      //System.out.println("n.leaf.cr.getName(): " + n.leaf.cr.getName());

      if(n.leaf.cr.getName().equals(BYTE)){
        if(nc.f0.which == 3)         // widening conversion
          converted_type = SHORT;
        else if(nc.f0.which == 4)    // widening conversion
          converted_type = INT;
        else if(nc.f0.which == 5)    // widening conversion
          converted_type = LONG;
        else if(nc.f0.which == 6)    // widening conversion
          converted_type = FLOAT;
        else if(nc.f0.which == 7)    // widening conversion
          converted_type = DOUBLE;
        else if(nc.f0.which == 1)    // Narrowing conversion
          converted_type = CHAR;
        else if(nc.f0.which == 2)    
          converted_type = BYTE;
      }
      else if(n.leaf.cr.getName().equals(SHORT)){
        if(nc.f0.which == 4)         // widening conversion
          converted_type = INT;
        else if(nc.f0.which == 5)    // widening conversion
          converted_type = LONG;
        else if(nc.f0.which == 6)    // widening conversion
          converted_type = FLOAT;
        else if(nc.f0.which == 7)    // widening conversion
          converted_type = DOUBLE;
        else if(nc.f0.which == 1)    // Narrowing conversion
          converted_type = CHAR;
        else if(nc.f0.which == 2)    // Narrowing conversion
          converted_type = BYTE;
        else if(nc.f0.which == 3)    
          converted_type = SHORT;
      }
      else if(n.leaf.cr.getName().equals(CHAR)){
        if(nc.f0.which == 4)         // widening conversion
          converted_type = INT;
        else if(nc.f0.which == 5)    // widening conversion
          converted_type = LONG;
        else if(nc.f0.which == 6)    // widening conversion
          converted_type = FLOAT;
        else if(nc.f0.which == 7)    // widening conversion
          converted_type = DOUBLE;
        else if(nc.f0.which == 2)    // Narrowing conversion
          converted_type = BYTE;
        else if(nc.f0.which == 3)    // Narrowing conversion
          converted_type = SHORT;
        else if(nc.f0.which == 1)    
          converted_type = CHAR;
      }
      else if(n.leaf.cr.getName().equals(INT)){
        if(nc.f0.which == 5)         // widening conversion
          converted_type = LONG;
        else if(nc.f0.which == 6)    // widening conversion
          converted_type = FLOAT;
        else if(nc.f0.which == 7)    // widening conversion
          converted_type = DOUBLE;
        else if(nc.f0.which == 1)    // Narrowing conversion
          converted_type = CHAR;
        else if(nc.f0.which == 2)    // Narrowing conversion
          converted_type = BYTE;
        else if(nc.f0.which == 3)    // Narrowing conversion
          converted_type = SHORT;
        else if(nc.f0.which == 4)    
          converted_type = INT;
      }
      else if(n.leaf.cr.getName().equals(LONG)){
        if(nc.f0.which == 6)         // widening conversion
          converted_type = FLOAT;
        else if(nc.f0.which == 7)    // widening conversion
          converted_type = DOUBLE;
        else if(nc.f0.which == 1)    // Narrowing conversion
          converted_type = CHAR;
        else if(nc.f0.which == 2)    // Narrowing conversion
          converted_type = BYTE;
        else if(nc.f0.which == 3)    // Narrowing conversion
          converted_type = SHORT;
        else if(nc.f0.which == 4)    // Narrowing conversion
          converted_type = INT;
        else if(nc.f0.which == 5)    
          converted_type = LONG;
      }
      else if(n.leaf.cr.getName().equals(FLOAT)){
        if(nc.f0.which == 7)         // widening conversion
          converted_type = DOUBLE;
        else if(nc.f0.which == 1)    // Narrowing conversion
          converted_type = CHAR;
        else if(nc.f0.which == 2)    // Narrowing conversion
          converted_type = BYTE;
        else if(nc.f0.which == 3)    // Narrowing conversion
          converted_type = SHORT;
        else if(nc.f0.which == 4)    // Narrowing conversion
          converted_type = INT;
        else if(nc.f0.which == 5)    // Narrowing conversion
          converted_type = LONG;
        else if(nc.f0.which == 6)    
          converted_type = FLOAT;
      }
      else if(n.leaf.cr.getName().equals(DOUBLE)){
        if(nc.f0.which == 1)         // Narrowing conversion
          converted_type = CHAR;
        else if(nc.f0.which == 2)    // Narrowing conversion
          converted_type = BYTE;
        else if(nc.f0.which == 3)    // Narrowing conversion
          converted_type = SHORT;
        else if(nc.f0.which == 4)    // Narrowing conversion
          converted_type = INT;
        else if(nc.f0.which == 5)    // Narrowing conversion
          converted_type = LONG;
        else if(nc.f0.which == 6)    // Narrowing conversion
          converted_type = FLOAT;
        else if(nc.f0.which == 7)    
          converted_type = DOUBLE;
      }
    }
    else{

			//System.out.println("********* n.type.cr.getName(): " + n.type.cr.getName());
			//System.out.println("********* n.leaf.cr.getName(): " + n.leaf.cr.getName());

      //if(n.type.cr.isArray())
      //  System.out.println("###### n.type.cr.isArray()");
      //else
      //  System.out.println("###### n.type.cr.is NOT Array()");

      //if(n.leaf.cr.isArray())
      //  System.out.println("###### n.leaf.cr.isArray()");
      //else
      //  System.out.println("###### n.leaf.cr.is NOT Array()");

      // widening or narrowing reference conversion

      // identity conversion
      if(n.type.cr.equals(n.leaf.cr)){
        converted_type = n.type.cr.getName();
        //System.out.println("n.type.cr is EQUAL to n.leaf.cr");
      }
      // widening reference conversion
      else if(n.type.cr.isSuperclassOf(n.leaf.cr)){
        converted_type = n.type.cr.getName();
        //System.out.println("n.type.cr is SUPERCLASS of n.leaf.cr");
      }
      else if(n.type.cr.isInterfaceOf(n.leaf.cr)){
        converted_type = n.type.cr.getName();
        //System.out.println("n.type.cr is INTERFACE of n.leaf.cr");
      }
      else if(n.type.cr.getName().equals("java.lang.Object") && n.leaf.cr.isInterface()){
        converted_type = n.type.cr.getName();
        //System.out.println("n.type.cr is java.lang.Object");        
      }
      else if(n.type.cr.getName().equals("java.lang.Object") && n.leaf.cr.isArray()){
        converted_type = n.type.cr.getName();
        //System.out.println("n.type.cr is java.lang.Object");        
      }
      else if(n.type.cr.getName().equals("java.lang.Cloneable") && n.leaf.cr.isArray()){
        converted_type = n.type.cr.getName();
        //System.out.println("n.type.cr is java.lang.Cloneable");        
      }
      // narrowing reference conversion
      else if(n.type.cr.isSubclassOf(n.leaf.cr)){
        converted_type = n.type.cr.getName();
        //System.out.println("n.type.cr is SUBCLASS of n.leaf.cr");
      }
      else if(n.leaf.cr.isInterfaceOf(n.type.cr)){
        converted_type = n.type.cr.getName();
        //System.out.println("n.leaf.cr is INTERFACE of n.type.cr");
      }
      else if(n.leaf.cr.getName().equals("java.lang.Object") && n.type.cr.isArray()){
        converted_type = n.type.cr.getName();
        //System.out.println("n.type.cr is Array");        
      }
      else if(n.leaf.cr.getName().equals("java.lang.Object") && n.type.cr.isInterface()){
        converted_type = n.type.cr.getName();
        //System.out.println("n.type.cr is INTERFACE");        
      }
      else if(n.leaf.cr.isInterface() && !n.type.cr.isInterface()){
        converted_type = n.type.cr.getName();
        //System.out.println("n.type.cr is CLASS");
      }
      else if(n.type.cr.isInterface() && !n.leaf.cr.isInterface()){
        converted_type = n.type.cr.getName();
        //System.out.println("n.type.cr is INTERFACE");
      }
      else if(n.leaf.cr.isInterface() && n.type.cr.isInterface()){
        converted_type = n.type.cr.getName();
        //System.out.println("n.type.cr and n.leaf.cr is INTERFACE");
      }
      else if(n.type.cr.isArray() && n.leaf.cr.isArray()){
        //System.out.println("n.type.cr.isArray() && n.leaf.cr.isArray()");
        if(n.type.cr.getComponentType().getName().equals(n.leaf.cr.getComponentType().getName())){
          converted_type = n.type.cr.getName();
          //System.out.println("n.type.cr and n.leaf.cr is Array");
        }
        else if(n.type.cr.getComponentType().isSuperclassOf(n.leaf.cr.getComponentType())){
          converted_type = n.type.cr.getName();
          //System.out.println("n.type.cr.getComponentType().isSuperclassOf(n.leaf.cr.getComponentType())");
        }
        else if(n.type.cr.getComponentType().isInterfaceOf(n.leaf.cr.getComponentType())){
          converted_type = n.type.cr.getName();
          //System.out.println("n.type.cr.getComponentType().isInterfaceOf(n.leaf.cr.getComponentType())");
        }
        else if(n.leaf.cr.getComponentType().isInterfaceOf(n.type.cr.getComponentType())){
          converted_type = n.type.cr.getName();
          //System.out.println("n.leaf.cr.getComponentType().isInterfaceOf(n.type.cr.getComponentType())");
        }
        else if(n.type.cr.getComponentType().isInterface() 
                || n.type.cr.getComponentType().isArray() 
                || !n.type.cr.getComponentType().isInterface()){
          if(n.leaf.cr.getComponentType().getName().equals("hpjava.compiler.abstractexp.NullNode")){
            converted_type = n.type.cr.getName();
            //System.out.println("n.leaf.cr is NullNode");
          }
        }
        else if(n.type.cr.getComponentType().getName().equals("java.lang.Object") 
                && n.leaf.cr.getComponentType().isInterface()){
          converted_type = n.type.cr.getName();
          //System.out.println("n.type.cr is Object");
        }
        else if(n.type.cr.getComponentType().getName().equals("java.lang.Object") 
                && n.leaf.cr.getComponentType().isArray()){
          converted_type = n.type.cr.getName();
          //System.out.println("n.type.cr is Object");
        }
        else if(n.type.cr.getComponentType().getName().equals("java.lang.Cloneable") 
                && n.leaf.cr.getComponentType().isArray()){
          converted_type = n.type.cr.getName();
          //System.out.println("n.type.cr is Cloneable");
        }      
      }           
    }

    //System.out.println("@@@@@@@@@ converted_type in CastNode: " + converted_type);
    return converted_type;
  }
  
}
