/*
 * File         : ExpressionManager.java
 * Author       : Han-Ku Lee
 * Created      : Wed Aug 12 13:03:10 1998
<<<<<<< ExpressionManager.java
 * Revision     : $Revision: 1.11 $
 * Updated      : $Date: 1999/10/25 22:07:58 $
=======
 * Revision     : $Revision: 1.11 $
 * Updated      : $Date: 1999/10/25 22:07:58 $
>>>>>>> 1.15
 */

package hpjava.compiler.abstractexp;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

import hpjava.compiler.parser.*;
import hpjava.compiler.unparser.*;
import hpjava.compiler.util.*;

import java.util.*;
import java.lang.*;
import java.io.*;

public class ExpressionManager{
    
   public static Hashtable table = new Hashtable();

   public ExpressionManager(Node r){
     ExpressionFinder finder = new ExpressionFinder(table);
     r.accept(finder);
     
     //System.out.println("Number of Element in HT:  "+ table.size());
   }

  public ExpressionManager (Project prj) {
    ExpressionFinder finder = new ExpressionFinder(table);

    Enumeration enum = prj.rootFileTable.keys();

  LOOP: while (enum.hasMoreElements()) {
      Node root = (Node) enum.nextElement();
      root.accept(finder);
    }

    if (prj.isVerbose()) {
      //System.err.println("Node scope: "+nodeScopeTable+" created.");
      System.err.println("Expression table created.");
      System.err.println();
    }
  }

  public static AbstractExpression lookup(Node n){
    return (AbstractExpression)table.get(n);
  }

  public static void register(Node n, AbstractExpression exp) {
    if(table.containsKey(n)){
      table.remove(n);
      //System.out.println("Yes ...q");
    }
    //else{
      //System.out.println("No ...q");      
      //}
    
    table.put(n, exp);
  }

  public void typeChecking() {
    if (!table.isEmpty()) {
      Enumeration enum = table.elements();

    LOOP:while (enum.hasMoreElements()) {

        AbstractExpression node = 
          (AbstractExpression) enum.nextElement();

        node.typeChecking();

      }

    }
  }
}

class ExpressionFinder extends ExpressionFindingVisitor{
  protected Hashtable refHashtable;

  final int DEFAULT = 0;
  final int RECURSIVE = 1;

  private IntStack status = new IntStack();
  
  public ExpressionFinder(Hashtable table) {
    this.refHashtable = table;
    status.push(DEFAULT);
  }

   //
   // f0 -> PrimaryPrefix()
   // f1 -> ( PrimarySuffix() )*
   //
  public void visit(PrimaryExpression n) {
    ExpressionConvertor convertor = new ExpressionConvertor(this.refHashtable);

    n.accept(convertor);    
    n.f0.accept(this);

    if (n.f1.present())
      n.f1.accept(this);    
  }

   //
   // f0 -> PrimaryExpression()
   // f1 -> ( "/" PrimaryExpression() )*
   //
  public void visit(GroupExpression n){
    ExpressionConvertor convertor = new ExpressionConvertor(this.refHashtable);

    n.accept(convertor);
    n.f0.accept(this);

    if (n.f1.present())
      n.f1.accept(this);
  }

   //
   // f0 -> ConditionalExpression()
   // f1 -> [ AssignmentOperator() Expression() ]
   //
  public void visit(Expression n) {
    if (status.peek() == DEFAULT) {
      ExpressionConvertor convertor = new ExpressionConvertor(this.refHashtable);

      n.accept(convertor);

      // go on to find 
      n.f0.accept(this); 
 
      if (n.f1.present()) {
        status.push(RECURSIVE);
        n.f1.accept(this); 
        status.pop();
      }

    } 
    else if (status.peek()==RECURSIVE) {
      status.push(DEFAULT);
      n.f0.accept(this);
      status.pop();

      if (n.f1.present()) {
        n.f1.accept(this);
      }
    }
    else {
      throw new Error("Invalid status.");
    }

  }

   //
   // f0 -> PreIncrementExpression()
   //       | PreDecrementExpression()
   //       | PrimaryExpression() [ "++" | "--" | AssignmentOperator() Expression() ]
   //

  public void visit(StatementExpression n) {
    ExpressionConvertor convertor =
      new ExpressionConvertor(this.refHashtable);

    n.accept(convertor);
    n.f0.accept(this);
  }


}


class ExpressionConvertor extends DepthFirstVisitor{
   protected PrintWriter out = new PrintWriter(System.out, true);
   protected Stack nodeStack = new Stack();
   protected Hashtable refHashtable;
   protected int op = 0;
   protected Node startPoint;

   private IntStack status = new IntStack();

   final int DEFAULT = 0;
   final int RECURSIVE = 1;

   public ExpressionConvertor(Hashtable ht){
     //out = new PrintWriter(System.out, true);
     refHashtable = ht;
     status.push(DEFAULT);
   }

   public void flushWriter(){
     out.flush();
   }

  /*
   //
   // A few convenience methods overloading visit()
   //
   public void visit(NodeList n, String sep) {
      for ( Enumeration e = n.elements(); e.hasMoreElements(); ) {
         ((Node)e.nextElement()).accept(this);
         if ( e.hasMoreElements() ) out.print(sep);
      }
   }

   public void visit(NodeListOptional n, String sep) {
      if ( n.present() )
         for ( Enumeration e = n.elements(); e.hasMoreElements(); ) {
            ((Node)e.nextElement()).accept(this);
            if ( e.hasMoreElements() ) out.print(sep);
         }
   }

   public void visit(NodeOptional n, String sep) {
      if ( n.present() ) {
         n.node.accept(this);
         out.print(sep);
      }
   }

   //
   // For convenience, trusts that the node passed to it is a NodeSequence.
   // (of course, will throw an exception if it isn't).
   //
   public void visit(Node n1, String sep) {
      NodeSequence n = (NodeSequence)n1;
      for ( Enumeration e = n.elements(); e.hasMoreElements(); ) {
         ((Node)e.nextElement()).accept(this);
         if ( e.hasMoreElements() ) out.print(sep);
      }
   }

   public void visit(NodeToken n) {
      out.print(n.tokenImage + " ");
   }
  */


  // start to override visit method 

   //
   // f0 -> ConditionalExpression()
   // f1 -> [ AssignmentOperator() Expression() ]
   //
   public void visit(Expression n) {
      n.f0.accept(this);
      AbstractExpression tmp = (AbstractExpression)nodeStack.pop();

      if(n.f1.present()){

        status.push(RECURSIVE);
        n.f1.accept(this);
        status.pop();

        switch( op ){
          case 0:
             tmp = tmp.assignment((AbstractExpression)nodeStack.pop());
             break;
          case 1:
             tmp = tmp.multiplicationAssignment((AbstractExpression)nodeStack.pop());
             break;
          case 2:
             tmp = tmp.divisionAssignment((AbstractExpression)nodeStack.pop());
             break;
          case 3:
             tmp = tmp.modulusAssignment((AbstractExpression)nodeStack.pop());
             break;
          case 4:
             tmp = tmp.additionAssignment((AbstractExpression)nodeStack.pop());
             break;
          case 5:
             tmp = tmp.subtractionAssignment((AbstractExpression)nodeStack.pop());
             break;
          case 6:
             tmp = tmp.leftShiftAssignment((AbstractExpression)nodeStack.pop());
             break;
          case 7:
             tmp = tmp.rightShiftAssignment((AbstractExpression)nodeStack.pop());
             break;
          case 8:
             tmp = tmp.extensionRightShiftAssignment((AbstractExpression)nodeStack.pop());
             break;
          case 9:
             tmp = tmp.andAssignment((AbstractExpression)nodeStack.pop());
             break;
          case 10:
             tmp = tmp.exclusiveOrAssignment((AbstractExpression)nodeStack.pop());
             break;
          case 11:
             tmp = tmp.inclusiveOrAssignment((AbstractExpression)nodeStack.pop());
             break;
          default:
             break;
        }
     }
      if (status.peek()==DEFAULT) {
        refHashtable.put(n, tmp);
        tmp.setExpression(n);
      }
      else if (status.peek()==RECURSIVE) {
        nodeStack.push(tmp);
      }
      else {
        throw new Error("Invalid status.");
      }

   }

   //
   // f0 -> "="
   //       | "*="
   //       | "/="
   //       | "%="
   //       | "+="
   //       | "-="
   //       | "<<="
   //       | ">>="
   //       | ">>>="
   //       | "&="
   //       | "^="
   //       | "|="
   //
   public void visit(AssignmentOperator n) {
      n.f0.accept(this);
      op = n.f0.which;
   }

   //
   // f0 -> ConditionalOrExpression()
   // f1 -> [ "?" Expression() ":" ConditionalExpression() ]
   //
   public void visit(ConditionalExpression n) {
      n.f0.accept(this);

      if(n.f1.present()){
        AbstractExpression tmp = (AbstractExpression)nodeStack.pop();
        NodeSequence seq = (NodeSequence)n.f1.node;

        status.push(RECURSIVE);
        seq.elementAt(1).accept(this);
        status.pop();

        AbstractExpression tmp2 = (AbstractExpression)nodeStack.pop();

        seq.elementAt(3).accept(this);
        AbstractExpression tmp3 = (AbstractExpression)nodeStack.pop();

        tmp = tmp.conditional(tmp2, tmp3);
        nodeStack.push(tmp);
      }

   }

   //
   // f0 -> ConditionalAndExpression()
   // f1 -> ( "||" ConditionalAndExpression() )*
   //
   public void visit(ConditionalOrExpression n) {
      n.f0.accept(this);

      if(n.f1.present()){
        AbstractExpression tmp = (AbstractExpression)nodeStack.pop();

        for(int i=n.f1.size()-1; i>=0;i--){
          n.f1.elementAt(i).accept(this);
        }

        for(int i=0; i<n.f1.size(); i++){
          tmp = tmp.binExpr(tmp.CONDITIONAL_OR, (AbstractExpression)nodeStack.pop());
        }

        nodeStack.push(tmp);
      }

   }

   //
   // f0 -> InclusiveOrExpression()
   // f1 -> ( "&&" InclusiveOrExpression() )*
   //
   public void visit(ConditionalAndExpression n) {
      n.f0.accept(this);

      if(n.f1.present()){
        AbstractExpression tmp = (AbstractExpression)nodeStack.pop();

        for(int i=n.f1.size()-1; i>=0;i--){
          n.f1.elementAt(i).accept(this);
        }

        for(int i=0; i<n.f1.size(); i++){
          tmp = tmp.binExpr(tmp.CONDITIONAL_AND, (AbstractExpression)nodeStack.pop());
        }

        nodeStack.push(tmp);
      }

   }

   //
   // f0 -> ExclusiveOrExpression()
   // f1 -> ( "|" ExclusiveOrExpression() )*
   //
   public void visit(InclusiveOrExpression n) {
      n.f0.accept(this);

      if(n.f1.present()){
        AbstractExpression tmp = (AbstractExpression)nodeStack.pop();
        for(int i=n.f1.size()-1; i>=0;i--){
          n.f1.elementAt(i).accept(this);
        }

        for(int i=0; i<n.f1.size(); i++){
          tmp = tmp.binExpr(tmp.INCLUSIVE_OR, (AbstractExpression)nodeStack.pop());
        }

        nodeStack.push(tmp);
      }

   }

   //
   // f0 -> AndExpression()
   // f1 -> ( "^" AndExpression() )*
   //
   public void visit(ExclusiveOrExpression n) {
      n.f0.accept(this);

      if(n.f1.present()){
        AbstractExpression tmp = (AbstractExpression)nodeStack.pop();
        for(int i=n.f1.size()-1; i>=0;i--){
          n.f1.elementAt(i).accept(this);
        }

        for(int i=0; i<n.f1.size(); i++){
          tmp = tmp.binExpr(tmp.EXCLUSIVE_OR, (AbstractExpression)nodeStack.pop());
        }

        nodeStack.push(tmp);
      }

   }

   //
   // f0 -> EqualityExpression()
   // f1 -> ( "&" EqualityExpression() )*
   //
   public void visit(AndExpression n) {
      n.f0.accept(this);

      if(n.f1.present()){
        AbstractExpression tmp = (AbstractExpression)nodeStack.pop();
        for(int i=n.f1.size()-1; i>=0;i--){
          n.f1.elementAt(i).accept(this);
        }

        for(int i=0; i<n.f1.size(); i++){
          tmp = tmp.binExpr(tmp.AND, (AbstractExpression)nodeStack.pop());
        }

        nodeStack.push(tmp);
      }
   }

   //
   // f0 -> InstanceOfExpression()
   // f1 -> ( ( "==" | "!=" ) InstanceOfExpression() )*
   //
   public void visit(EqualityExpression n) {
      n.f0.accept(this);

      if(n.f1.present()){
        AbstractExpression tmp = (AbstractExpression)nodeStack.pop();
        for(int i=n.f1.size()-1; i>=0;i--){
          n.f1.elementAt(i).accept(this);
        }

        for(int i=0; i<n.f1.size(); i++){

          NodeSequence seq = (NodeSequence)n.f1.elementAt(i);

          if(((NodeChoice)seq.elementAt(0)).which == 0){
            tmp = tmp.binExpr(tmp.EQUALITY, (AbstractExpression)nodeStack.pop());
          }
          else{
            tmp = tmp.binExpr(tmp.INEQUALITY, (AbstractExpression)nodeStack.pop());
          }
        }

        nodeStack.push(tmp);
      }

   }

   //
   // f0 -> RelationalExpression()
   // f1 -> [ "instanceof" Type() ]
   //
   public void visit(InstanceOfExpression n) {
      n.f0.accept(this);
			
      if(n.f1.present()){
        AbstractExpression tmp = (AbstractExpression)nodeStack.pop();
        NodeSequence seq = (NodeSequence)n.f1.node;
        seq.elementAt(1).accept(this);

				//        TypeNode type = (TypeNode)seq.elementAt(1);
        TypeNode type = new TypeNode((Type)seq.elementAt(1));

        tmp = tmp.binExpr(tmp.INSTANCE_OF, type);

        nodeStack.push(tmp);
      }
   }

   //
   // f0 -> ShiftExpression()
   // f1 -> ( ( "<" | ">" | "<=" | ">=" ) ShiftExpression() )*
   //
   public void visit(RelationalExpression n) {
      n.f0.accept(this);

      if(n.f1.present()){
        AbstractExpression tmp = (AbstractExpression)nodeStack.pop();
        for(int i=n.f1.size()-1; i>=0;i--){
          n.f1.elementAt(i).accept(this);
        }

        for(int i=0; i<n.f1.size(); i++){

          NodeSequence seq = (NodeSequence)n.f1.elementAt(i);

          if(((NodeChoice)seq.elementAt(0)).which == 0){
            tmp = tmp.binExpr(tmp.LESS, (AbstractExpression)nodeStack.pop());
          }
          else if(((NodeChoice)seq.elementAt(0)).which == 1){
            tmp = tmp.binExpr(tmp.GREATER, (AbstractExpression)nodeStack.pop());
          }
          else if(((NodeChoice)seq.elementAt(0)).which == 2){
            tmp = tmp.binExpr(tmp.LESS_THAN_EQ, (AbstractExpression)nodeStack.pop());
          }
          else{
            tmp = tmp.binExpr(tmp.GREATER_THAN_EQ, (AbstractExpression)nodeStack.pop());
          }
        }

        nodeStack.push(tmp);
      }

   }

   //
   // f0 -> AdditiveExpression()
   // f1 -> ( ( "<<" | ">>" | ">>>" ) AdditiveExpression() )*
   //
   public void visit(ShiftExpression n) {
      n.f0.accept(this);

      if(n.f1.present()){
        AbstractExpression tmp = (AbstractExpression)nodeStack.pop();
        for(int i=n.f1.size()-1; i>=0;i--){
          n.f1.elementAt(i).accept(this);
        }

        for(int i=0; i<n.f1.size(); i++){

          NodeSequence seq = (NodeSequence)n.f1.elementAt(i);

          if(((NodeChoice)seq.elementAt(0)).which == 0){
            tmp = tmp.binExpr(tmp.LEFT_SHIFT, (AbstractExpression)nodeStack.pop());
          }
          else if(((NodeChoice)seq.elementAt(0)).which == 1){
            tmp = tmp.binExpr(tmp.RIGHT_SHIFT, (AbstractExpression)nodeStack.pop());
          }
          else{
            tmp = tmp.binExpr(tmp.EXTENSION_RIGHT_SHIFT, (AbstractExpression)nodeStack.pop());
          }
        }

        nodeStack.push(tmp);
      }

   }

   //
   // f0 -> MultiplicativeExpression()
   // f1 -> ( ( "+" | "-" ) MultiplicativeExpression() )*
   //
   public void visit(AdditiveExpression n) {
      n.f0.accept(this);

      if(n.f1.present()){
        AbstractExpression tmp = (AbstractExpression)nodeStack.pop();
        for(int i=n.f1.size()-1; i>=0;i--){
          n.f1.elementAt(i).accept(this);
        }

        for(int i=0; i<n.f1.size(); i++){

          NodeSequence seq = (NodeSequence)n.f1.elementAt(i);

          if(((NodeChoice)seq.elementAt(0)).which == 0){
            tmp = tmp.binExpr(tmp.ADD, (AbstractExpression)nodeStack.pop());
          }
          else{
            tmp = tmp.binExpr(tmp.SUBTRACT, (AbstractExpression)nodeStack.pop());
          }
        }

        nodeStack.push(tmp);
      }
   }

   //
   // f0 -> UnaryExpression()
   // f1 -> ( ( "*" | "/" | "%" ) UnaryExpression() )*
   //
   public void visit(MultiplicativeExpression n) {
      n.f0.accept(this);

      if(n.f1.present()){
        AbstractExpression tmp = (AbstractExpression)nodeStack.pop();
        for(int i=n.f1.size()-1; i>=0;i--){
          n.f1.elementAt(i).accept(this);
        }

        for(int i=0; i<n.f1.size(); i++){

          NodeSequence seq = (NodeSequence)n.f1.elementAt(i);

          if(((NodeChoice)seq.elementAt(0)).which == 0){
            tmp = tmp.binExpr(tmp.MULTIPLE, (AbstractExpression)nodeStack.pop());
          }
          else if(((NodeChoice)seq.elementAt(0)).which == 1){
            tmp = tmp.binExpr(tmp.DIVIDE, (AbstractExpression)nodeStack.pop());
          }
          else{
            tmp = tmp.binExpr(tmp.MODULE, (AbstractExpression)nodeStack.pop());
          }
        }

        nodeStack.push(tmp);
      }

   }

   //
   // f0 -> ( "+" | "-" ) UnaryExpression()
   //       | PreIncrementExpression()
   //       | PreDecrementExpression()
   //       | UnaryExpressionNotPlusMinus()
   //
   public void visit(UnaryExpression n) {
      if(n.f0.which != 0){
        n.f0.accept(this);
      }
      else{
        NodeSequence seq = (NodeSequence)n.f0.choice;
        seq.elementAt(1).accept(this);
        AbstractExpression tmp = (AbstractExpression)nodeStack.pop();

        if(((NodeChoice)seq.elementAt(0)).which == 0){
          tmp = tmp.unaryExpr(tmp.PLUS);
        }
        else{
          tmp = tmp.unaryExpr(tmp.MINUS);
        }

        nodeStack.push(tmp);
      }

   }

   //
   // f0 -> "++"
   // f1 -> PrimaryExpression()
   //
   public void visit(PreIncrementExpression n) {
      n.f0.accept(this);
      n.f1.accept(this);
      AbstractExpression tmp = (AbstractExpression)nodeStack.pop();

      tmp = tmp.unaryExpr(tmp.PRE_INCREMENT);
      nodeStack.push(tmp);
   }

   //
   // f0 -> "--"
   // f1 -> PrimaryExpression()
   //
   public void visit(PreDecrementExpression n) {
      n.f0.accept(this);
      n.f1.accept(this);
      AbstractExpression tmp = (AbstractExpression)nodeStack.pop();

      tmp = tmp.unaryExpr(tmp.PRE_DECREMENT);
      nodeStack.push(tmp);
   }

   //
   // f0 -> ( "~" | "!" ) UnaryExpression()
   //       | CastExpression()
   //       | PostfixExpression()
   //
   public void visit(UnaryExpressionNotPlusMinus n) {
      if(n.f0.which != 0){
        n.f0.accept(this);
      }
      else{
        NodeSequence seq = (NodeSequence)n.f0.choice;
        seq.elementAt(1).accept(this);
        AbstractExpression tmp = (AbstractExpression)nodeStack.pop();

        if(((NodeChoice)seq.elementAt(0)).which == 0){
          tmp = tmp.unaryExpr(tmp.COMPLEMENT);
        }
        else{
          tmp = tmp.unaryExpr(tmp.LOGICAL_NOT);
        }

        nodeStack.push(tmp);
      }

   }

   //
   // f0 -> PrimaryExpression()
   // f1 -> [ "++" | "--" | "`"]
   //
   public void visit(PostfixExpression n) {
      n.f0.accept(this);
      AbstractExpression tmp = (AbstractExpression)nodeStack.pop();

      if(n.f1.present()){
        n.f1.accept(this);
        NodeChoice choice = (NodeChoice)n.f1.node;

        if(choice.which == 0)
          tmp = tmp.unaryExpr(tmp.POSTFIX_ADDITION);
        else if(choice.which == 1)
          tmp = tmp.unaryExpr(tmp.POSTFIX_SUBTRACTION);
        else
          tmp = tmp.unaryExpr(tmp.POSTFIX_INDEX_LOCATION);
      }

      nodeStack.push(tmp);
   }

   //
   // f0 -> "(" Type() ")" UnaryExpression()
   //       | "(" Type() ")" UnaryExpressionNotPlusMinus()
   //
   public void visit(CastExpression n) {
      NodeSequence seq = (NodeSequence)n.f0.choice;
      seq.elementAt(3).accept(this);
      AbstractExpression tmp = (AbstractExpression)nodeStack.pop();

//      TypeNode type = (TypeNode)seq.elementAt(1);
      TypeNode type = new TypeNode((Type)seq.elementAt(1));

      tmp = tmp.cast(type);
      
      nodeStack.push(tmp);
   }

  /*
   //
   // f0 -> PrimaryPrefix()
   // f1 -> ( PrimarySuffix() )*
   //
   public void visit(PrimaryExpression n) {
      PrimaryExpressionNode pl = new PrimaryExpressionNode(n);
      AbstractExpression rtn = pl;

      nodeStack.push(rtn);

      //      n.f0.accept(this);
      //      n.f1.accept(this);
   }
  */

   //
   // f0 -> PrimaryExpression()
   // f1 -> ( "/" PrimaryExpression() )*
   //
   public void visit(GroupExpression n) {
      n.f0.accept(this);
      AbstractExpression tmp = (AbstractExpression)nodeStack.pop();

      if(n.f1.present()){

        for(int i=n.f1.size()-1; i>=0;i--){
          n.f1.elementAt(i).accept(this);
        }

        for(int i=0; i<n.f1.size(); i++){
          //tmp = tmp.group((AbstractExpression)nodeStack.pop());
          tmp = tmp.binExpr(tmp.GROUP, (AbstractExpression)nodeStack.pop());
        }

        nodeStack.push(tmp);
      }

      //     nodeStack.push(tmp);
     refHashtable.put(n, tmp);
   }


   //
   // f0 -> PreIncrementExpression()
   //       | PreDecrementExpression()
   //       | PrimaryExpression() [ "++" | "--" | AssignmentOperator() Expression() ]
   //
   public void visit(StatementExpression n) {

      if(n.f0.which != 2){
        n.f0.accept(this);
        AbstractExpression tmp = (AbstractExpression)nodeStack.pop();
        nodeStack.push(tmp);

        refHashtable.put(n, tmp);
        tmp.setStatementExpression(n);
      }
      else{
        NodeSequence seq = (NodeSequence)n.f0.choice;
        seq.elementAt(0).accept(this);
        AbstractExpression tmp = (AbstractExpression)nodeStack.pop();

        if(((NodeOptional)seq.elementAt(1)).present()){
          NodeOptional opt = (NodeOptional)seq.elementAt(1);

          if(((NodeChoice)opt.node).which == 0){
            //tmp = tmp.postfixAddition();
            tmp = tmp.unaryExpr(tmp.POSTFIX_ADDITION);
            opt.accept(this);
          }
          else if(((NodeChoice)opt.node).which == 1){
            //tmp = tmp.postfixSubtraction();
            tmp = tmp.unaryExpr(tmp.POSTFIX_SUBTRACTION);
            opt.accept(this);
          }
          else{
            status.push(RECURSIVE);
            opt.accept(this);
            status.pop();

            switch( op ){
              case 0:
                tmp = tmp.assignment((AbstractExpression)nodeStack.pop());
                break;
              case 1:
                tmp = tmp.multiplicationAssignment((AbstractExpression)nodeStack.pop());
                break;
              case 2:
                tmp = tmp.divisionAssignment((AbstractExpression)nodeStack.pop());
                break;
              case 3:
                tmp = tmp.modulusAssignment((AbstractExpression)nodeStack.pop());
                break;
              case 4:
                tmp = tmp.additionAssignment((AbstractExpression)nodeStack.pop());
                break;
              case 5:
                tmp = tmp.subtractionAssignment((AbstractExpression)nodeStack.pop());
                break;
              case 6:
                tmp = tmp.leftShiftAssignment((AbstractExpression)nodeStack.pop());
                break;
              case 7:
                tmp = tmp.rightShiftAssignment((AbstractExpression)nodeStack.pop());
                break;
              case 8:
                tmp = tmp.extensionRightShiftAssignment((AbstractExpression)nodeStack.pop());
                break;
              case 9:
                tmp = tmp.andAssignment((AbstractExpression)nodeStack.pop());
                break;
              case 10:
                tmp = tmp.exclusiveOrAssignment((AbstractExpression)nodeStack.pop());
                break;
              case 11:
                tmp = tmp.inclusiveOrAssignment((AbstractExpression)nodeStack.pop());
                break;
              default:
                break;
            }

          }

        }

        if (status.peek()==DEFAULT) { 
          refHashtable.put(n, tmp); 
          tmp.setStatementExpression(n);
        } 
        else if (status.peek()==RECURSIVE) { 
          nodeStack.push(tmp); 
        } 
        else { 
          throw new Error("Invalid status."); 
        } 
 
      } 
 
   }


  // From here, PrimaryExpression is converted.

   private IntStack suffixStack = new IntStack();

   final int THIS = 0;
   final int ALLOCATION_EXPRESSION = 1;
   final int DIST_ARRAY_SEC_REF_DIM_LIST = 2;
   final int TRIPLET_OR_EXPRESSION = 3;
   final int IDENTIFIER = 4;
   final int ARGUMENT = 5;

   //
   // f0 -> PrimaryPrefix()
   // f1 -> ( PrimarySuffix() )*
   //
  public void visit(PrimaryExpression n) {

    n.f0.accept(this);

    if (n.f1.present()) {
      AbstractExpression tmp = (AbstractExpression)nodeStack.pop();

      //if(nodeStack.empty())
      //   System.out.println("IT IS EMPTY...");

      //System.out.println("Name of tmp: " + tmp.getClass().getName());

      for(int i=n.f1.size()-1; i>=0;i--){
        n.f1.elementAt(i).accept(this);
      }

      //System.out.println("AbstractExpression I in nodeStack: " + ((AbstractExpression)nodeStack.pop()).getClass().getName());
      //System.out.println("AbstractExpression II in nodeStack: " + ((AbstractExpression)nodeStack.pop()).getClass().getName());
      

      //System.out.println("The size of PrimarySuffix: " + n.f1.size());
  
      //for(int i=0; i<n.f1.size(); i++){
      //for(int i=0; i<=n.f1.size(); i++){
      while(!suffixStack.empty()){
      //while(!nodeStack.empty()){

        switch(suffixStack.pop()){
        case THIS:
          break;
        case ALLOCATION_EXPRESSION:
          tmp = tmp.inner_allocationExpr((AllocationExpressionNode)nodeStack.pop());
          if(!nodeStack.empty()){
            AbstractExpression aexp1 = (AbstractExpression)nodeStack.pop();  
            if(status.peek() == DEFAULT)
              nodeStack.push(aexp1);
          }
          break;
        case DIST_ARRAY_SEC_REF_DIM_LIST:
          tmp = tmp.arraySect((DistArraySecRefDimListNode)nodeStack.pop());
          if(!nodeStack.empty()){
            AbstractExpression aexp1 = (AbstractExpression)nodeStack.pop();  
            if(status.peek() == DEFAULT)
              nodeStack.push(aexp1);
          }
          break;
        case TRIPLET_OR_EXPRESSION:	  
          //if(((AbstractExpression)nodeStack.peek()).getClass().getName().equals
          //   ("hpjava.compiler.abstractexp.IdentifierNode")){
          //  IdentifierNode id = (IdentifierNode)nodeStack.peek();
          //  System.out.println("Identifier in PrimaryExpression: " + id.leaf.tokenImage);
          //}
          //System.out.println("AbstractExpression: " + ((AbstractExpression)nodeStack.peek()).getClass().getName());
          //System.out.println("BEFORE");
          tmp = tmp.arrayElemRef((TripletOrExpressionListNode)nodeStack.pop());

          if(!nodeStack.empty()){
            
            AbstractExpression aexp1 = (AbstractExpression)nodeStack.pop();
            if(status.peek() == DEFAULT)
              nodeStack.push(aexp1);
          }
          //System.out.println("AFTER");	  
          break;
        case IDENTIFIER:
          //AbstractExpression aexp = (AbstractExpression)nodeStack.peek();
          //System.out.println("Name of AbstractExpression: " + aexp.getClass().getName());
					//IdentifierNode id = (IdentifierNode)nodeStack.peek();
          //System.out.println("Identifier in PrimaryExpression: " + id.leaf.tokenImage);

          if(!suffixStack.empty()){
            if(suffixStack.peek() == ARGUMENT){
              tmp = tmp.methodCall((IdentifierNode)nodeStack.pop(),(AbstractExpression)nodeStack.pop());
              suffixStack.pop();
            }
            //else if(suffixStack.peek() == IDENTIFIER){
            else{
              tmp = tmp.classField((IdentifierNode)nodeStack.pop());
            }
          }
          else{
            tmp = tmp.classField((IdentifierNode)nodeStack.pop());
          }
          break;
        case ARGUMENT:
          //AbstractExpression aexp2 = (AbstractExpression)nodeStack.peek();
          //System.out.println("Name of AbstractExpression1: " + tmp.getClass().getName());
          //System.out.println("Name of AbstractExpression2: " + aexp2.getClass().getName());

          //tmp = new MethodCallNode(new ThisNode(),
          //                         (IdentifierNode)tmp,
          //                         (AbstractExpression)nodeStack.pop());

          //ThisNode tmp_This = new ThisNode();
          NullNode tmp_Null = new NullNode();
          tmp = tmp_Null.methodCall((IdentifierNode)tmp,(AbstractExpression)nodeStack.pop());

          break;
        default:
          throw new Error("Unexpected case.");
        }

      }
      nodeStack.push(tmp);
    }
     
    AbstractExpression tmp = (AbstractExpression) nodeStack.peek();
    refHashtable.put(n, tmp);
    tmp.setPrimaryExpression(n);

    //System.out.println("Name of tmp at the end: " + tmp.getClass().getName());

  }


   //
   // f0 -> Literal()
   //       | ResultType() "." "class"
   //       | AllocationExpression()
   //       | "this"
   //       | "super"
   //       | "(" Expression() ")"
   //       | <IDENTIFIER>
   //
   public void visit(PrimaryPrefix n) {
      Node repNode;
      AbstractExpression tmp;
      
      if(n.f0.which != 5 && status.peek() == RECURSIVE)
        n.f0.accept(this);

      switch (n.f0.which) {
      case 0:
        tmp = new LiteralNode((Literal)n.f0.choice);        
        break;
      case 1:
        repNode = ((NodeSequence) n.f0.choice).elementAt(0);
        tmp = new ClassLiteralNode((ResultType) repNode);
        break;
      case 2:
        tmp = new AllocationExpressionNode((AllocationExpression)n.f0.choice);
        break;
      case 3:
        //tmp = new ThisNode();
        tmp = new ThisNode(n.f0.choice);
        break;
      case 4:
        tmp = new SuperNode();
        break;
      case 5:
        repNode = ((NodeSequence) n.f0.choice).elementAt(1);
        tmp = new ExpressionNode((Expression) repNode);
        break;
      case 6:
        tmp = new IdentifierNode((NodeToken)n.f0.choice);

        //System.out.println("IdentifierNode in PrimaryPrefix: " + n.f0.choice);

        break;
        default:
        throw new Error("Unexpected case.");
      }

      //n.f0.accept(this);
      
      nodeStack.push(tmp);

      //AbstractExpression aexp = (AbstractExpression)nodeStack.peek();
      //System.out.println("Name of AbstractExpression in PrimaryPrefix: " + aexp.getClass().getName());

      //if(nodeStack.empty())
      //  System.out.println("nodeStack in PrimaryPrefix is EMPTY...");
      //else
      //  System.out.println("nodeStack in PrimaryPrefix is NOT EMPTY...");

   }

   //
   // f0 -> "." "this"
   //       | "." AllocationExpression()
   //       | "[" "[" DistArraySecRefDimList() "]" "]"
   //       | "[" TripletOrExpressionList() "]"
   //       | "." <IDENTIFIER>
   //       | Arguments()
   //
   public void visit(PrimarySuffix n) {
      AbstractExpression suffix = null;
      Node repNode = null;

      n.f0.accept(this);

      switch (n.f0.which) {
      case 0:
        repNode = ((NodeSequence) n.f0.choice).elementAt(1);
        suffix = new ThisNode();
        suffixStack.push(THIS);
        break;
      case 1:
        repNode = ((NodeSequence) n.f0.choice).elementAt(1);
        suffix = new AllocationExpressionNode((AllocationExpression) repNode);
        suffixStack.push(ALLOCATION_EXPRESSION);
        break;
      case 2:
        repNode = ((NodeSequence) n.f0.choice).elementAt(2);
        suffix = new DistArraySecRefDimListNode((DistArraySecRefDimList) repNode);
        suffixStack.push(DIST_ARRAY_SEC_REF_DIM_LIST);
        break;
      case 3:
        repNode = ((NodeSequence) n.f0.choice).elementAt(1);
        suffix = new TripletOrExpressionListNode((TripletOrExpressionList) repNode);
        suffixStack.push(TRIPLET_OR_EXPRESSION);
        break;
      case 4:
        repNode = ((NodeSequence) n.f0.choice).elementAt(1);
        suffix = new IdentifierNode((NodeToken) repNode);
        suffixStack.push(IDENTIFIER);

        //System.out.println("IdentifierNode in PrimarySuffix: " + repNode);

        break;
      case 5:
        repNode = (Arguments) n.f0.choice;
        suffix = new ArgumentsNode((Arguments) repNode);
        suffixStack.push(ARGUMENT);
        break;
      default:
        throw new Error("Unexpected case.");
      }

      //n.f0.accept(this);
      nodeStack.push(suffix);

      //System.out.println("Name of suffix in PrimarySuffix: " + suffix.getClass().getName());
   }  

}
