/*
 * File         : Generator.java
 * Author       : Guansong Zhang
 * Created      : Fri Jun 19 11:44:00 1998
 * Revision     : $Revision: 1.14 $
 * Updated      : $Date: 1999/03/16 20:25:36 $
 */

package hpjava.compiler.gar;

import java.io.*;
import java.util.*;

import hpjava.compiler.util.*;

import hpjava.compiler.parser.*;
import hpjava.compiler.unparser.*;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

import hpjava.compiler.reflection.*;

public class Generator 
{  
  static final int DIM = 4;

  static public File srcDir;
  static public File repositoryDir;


  static void init() {
    Properties props = new Properties(System.getProperties());

    String home = props.getProperty("hpjava.home");

    String adlib = props.getProperty("adlib.home");

    String repository = props.getProperty("hpjava.repository");

    if (home != null) {
      srcDir = new File(adlib, "src");
    }

    if (repository != null)
      repositoryDir = new File (repository);

  }

  static {
    init();
  }

  static public void generate(ArrayReflectorDescriptor ard) {
    //    throws IOException {
    if (!ard.isDistArray())
      throw new Error();
    throw new Error("not finished.");
  }

  static void addheader(PrintWriter writer) {
    writer.println("/*");
    writer.println(" * Revision     : $"+"Revision"+"$");
    writer.println(" * Updated      : $"+"Date"+"$");
    writer.println(" * Note         : Do not edit this generalized file.");
    writer.println(" */");
    writer.println();
  }

  static String basename (ScalarReflectorDescriptor srd) {
    if (srd.isPrimitive()) 
      return srd.signature();
    else {
      throw new Error();
    }      
  }

  static void initRange(PrintWriter writer, ArrayReflectorDescriptor ard) {
    int dim = ard.dimension();

    if (dim !=0) {
      writer.println("    //init range");

      for (int i =0; i< dim; i++) {
        if (!ard.isCollapsedDim(i)) {
          writer.println("    this.range["+i+"] = r"+i+";");
        }
        else {
          writer.println("    this.range["+i+"] = new CollapsedRange(i"+i+");");
        }
      }
      
      writer.println();
    }
  }

  static void initStride(PrintWriter writer, ArrayReflectorDescriptor ard) {
    int dim = ard.dimension();

    if (dim !=0 ) {
      writer.println("    //init stride");
      
      for (int i =0; i< dim; i++) {
        if (i ==0)
          writer.println("    this.stride["+i+"] = this.range["+i+"].stride(size);");
        else
          writer.println("    this.stride["+i+"] = this.range["+i+"].stride(this.stride["+(i-1)+"].size);");
      }
      writer.println();

      writer.println("    size = stride["+(dim-1)+"].size;");
      writer.println();
    }

  }

  static void allocateConstructor(PrintWriter writer, ArrayReflectorDescriptor ard) {
    String attribute = ard.distAttributeSignature();
    String base = basename((ScalarReflectorDescriptor)ard.getComponentType());
    int dim = ard.dimension();
    String classname = "Array" + dim + attribute+base;

    String typename = ((ScalarReflectorDescriptor)ard.getComponentType()).name();

    writer.print("  public "+classname+" (Group group");    
    for (int i = 0; i<dim; i++) {
      writer.print(",");

      if (ard.isCollapsedDim(i))
        writer.print("int i"+i);
      else
        writer.print("Range r"+i);
    }
    writer.println(") {");

    writer.println("    this.group = group;");
    writer.println();

    if (dim != 0) {
      writer.println("    this.range = new Range["+dim+"];");
      writer.println("    this.stride = new Stride["+dim+"];");
      writer.println();
    }

    initRange(writer, ard);

    writer.println("    int size = 1;");
    writer.println();

    initStride(writer, ard);

    if (dim != 0 ) {
      writer.println("    this.dad = new DAD (group, range, stride);");
    }
    else {
      writer.println("    this.dad = new DAD (group);");
    };

    writer.println();
    writer.println("    this.data = new "+typename+"[size];");
    writer.println("    this.offset = 0;");
    writer.println();
    writer.println("  }");
    writer.println();

  }

  static void reshapConstructor(PrintWriter writer, ArrayReflectorDescriptor ard) {
    String attribute = ard.distAttributeSignature();
    String base = basename((ScalarReflectorDescriptor)ard.getComponentType());

    String classname = "Array" + ard.dimension() + attribute+base;

    String typename = ((ScalarReflectorDescriptor)ard.getComponentType()).name();

    writer.print("  public "+classname+" (Group group,"); 
    for ( int i = 0; i<ard.dimension(); i++) {
      writer.print("Range r"+i+",Stride s"+i+",");
    }     
    writer.println(typename+"[] data,int offset) {");    
   
    writer.println("    this.group = group;");
    writer.println();
    
    if (ard.dimension()!=0) {
      writer.println("    this.range = new Range["+ard.dimension()+"];");
      writer.println("    this.stride = new Stride["+ard.dimension()+"];");
      writer.println();

      for ( int i = 0; i<ard.dimension(); i++) {
        writer.println("    this.range ["+i+"] = r"+i+";");
        writer.println("    this.stride["+i+"] = s"+i+";");
      }
      writer.println();
    }

    if (ard.dimension()!=0) {
      writer.println("    this.dad = new DAD (group, range, stride);");
    }
    else {
      writer.println("    this.dad = new DAD (group);");
    }
    writer.println();

    writer.println("    this.data = data;");
    writer.println("    this.offset = offset;");
    writer.println();
    writer.println("  }");
    writer.println();
  }

  static void implementingSection(PrintWriter writer, ArrayReflectorDescriptor ard) {
    writer.println("  public DAD dad() {");
    writer.println("    return dad;");
    writer.println("  }");
    writer.println();
    writer.println("  public int off() {");
    writer.println("    return offset;");
    writer.println("  }");
    writer.println();
    writer.println("  public Object ref() {");
    writer.println("    return dat();");
    writer.println("  }");
    writer.println();
  }

  static void implementingSectionWithType(PrintWriter writer, ArrayReflectorDescriptor ard) {
    String typename = ((ScalarReflectorDescriptor)ard.getComponentType()).name();

    writer.println("  public "+typename+"[] dat() {");
    writer.println("    return this.data;");
    writer.println("  }");
    writer.println();
  }

  static void implementingSectionWithTypeDistAttr(PrintWriter writer, ArrayReflectorDescriptor ard) {
    writer.print("  public int pos(");
    for (int dim = 0; dim<ard.dimension(); dim++) {
      if (ard.isCollapsedDim(dim))
        writer.print("int i"+dim);
      else
        writer.print("Location l"+dim);
      if (dim!=ard.dimension()-1)
        writer.print(",");
    }
    writer.println(") {");

    writer.print("    return this.offset");
    for (int dim = 0; dim<ard.dimension(); dim++) {
      writer.print("+");

      if (ard.isCollapsedDim(dim))
        writer.print("this.stride["+dim+"].offset(this.range["+dim+"].location(i"+dim+"))");
      else
        writer.print("this.stride["+dim+"].offset(l"+dim+")");
    }
    writer.println(";");

    writer.println("  }");

  }

  static void writeArray(FileOutputStream out, ArrayReflectorDescriptor ard) {
    PrintWriter writer = new PrintWriter(out, true);

    String attribute = ard.distAttributeSignature();
    String base = basename((ScalarReflectorDescriptor)ard.getComponentType());

    String classname = "Array" + ard.dimension() + attribute+base;
    String interfacename = "Section" + ard.dimension() + attribute+base;

    String name = ((ScalarReflectorDescriptor)ard.getComponentType()).name();

    addheader (writer);

    writer.println("package hpjava.lang;");
    writer.println();
    writer.println("public class "+classname+" extends Array implements "+interfacename+" {");
    writer.println("  private "+name+" [] data;");
    writer.println();

    allocateConstructor(writer, ard);

    reshapConstructor(writer, ard);
    
    writer.println("  //implementing Section");
    writer.println();
    implementingSection(writer, ard);

    writer.println("  //implementing Section with type");
    writer.println();
    implementingSectionWithType(writer, ard);

    writer.println("  //implementing Section with type and distribute attributes");
    writer.println();
    implementingSectionWithTypeDistAttr(writer, ard);

    ArrayReflectorDescriptor ards[] = impSectionList(ard);
    if (ards != null) {
      writer.println();
      writer.println("  //implementing inherited Section with type and distribute attributes");
      writer.println();
    
      for (int i = 0; i< ards.length; i++) 
        implementingSectionWithTypeDistAttr(writer, ards[i]);

    }
    writer.println("}");
    writer.println();

    writer.close();
  }

  static void writeSection(FileOutputStream out, ArrayReflectorDescriptor ard) {
    PrintWriter writer = new PrintWriter(out, true);

    String attribute = ard.distAttributeSignature();
    String base = basename((ScalarReflectorDescriptor)ard.getComponentType());

    String interfacename = "Section" + ard.dimension() + attribute + base;

    String name = ((ScalarReflectorDescriptor)ard.getComponentType()).name();

    addheader (writer);

    writer.println("package hpjava.lang;");
    writer.println();

    writer.print("public interface "+interfacename+" extends ");

    if (ard.noCollapsedDim())
      writer.print("Section"+base);
    else {
      // this imp list includes all the section with dist attribute signature.
      // without using indirect inheritance.
      ArrayReflectorDescriptor[] ards = impSectionList(ard);

      for (int i = 0 ; i<ards.length; i++) {
        writer.print("Section"+ard.dimension()+ards[i].distAttributeSignature()+base);
        if (i!=ards.length-1)
          writer.print(",");
      }
    }

    writer.println(" {");

    writer.print("  public int pos(");
    for (int i=0; i<ard.dimension(); i++) {
      if (ard.isCollapsedDim(i)) 
        writer.print("int i"+i);
      else
        writer.print("Location l"+i);      
      if (i != ard.dimension()-1)
        writer.print(",");
    }
    writer.println(");");

    writer.println("}");

    writer.close();
  }

  static ArrayReflectorDescriptor[] impSectionList(ArrayReflectorDescriptor ard) {
    int dim = ard.dimension();

    if (ard.noCollapsedDim())
      return new ArrayReflectorDescriptor[0];
    else {
      int cdim =0;
      for (int i = 0; i<dim; i++)
        if (ard.isCollapsedDim(i))
          cdim++;

      String cmb[] = combine(cdim);
      
      // we don't want it to implent itself
      ArrayReflectorDescriptor[] results = new ArrayReflectorDescriptor[cmb.length-1];

      int counter =0;
      for (int i = 0; i<cmb.length; i++) {

        StringBuffer c = new StringBuffer();
        for (int j = 0 ; j < dim; j++)
          if (ard.isCollapsedDim(j))
            c.append('c');

        if (cmb[i].equals(c.toString())){
          continue;
        }                

        ClassReflectorDescriptor crd = ard.getComponentType();

        StringBuffer newattr = new StringBuffer();

        int index = 0;
        for (int j = 0 ; j<dim; j++) {
          if (!ard.isCollapsedDim(j))
            newattr.append('d');
          else {
            newattr.append(cmb[i].charAt(index));
            index++;
          }
        }

        results[counter] = new ArrayReflectorDescriptor(crd, "]"+dim+newattr);
        counter++;
      }
      return results;
    }

  }

  static void generateArray(File srcDir, ArrayReflectorDescriptor ard) throws IOException {
    if (!ard.isDistArray())
      throw new Error();

    File dest = new File (srcDir, "hpjava");
    dest = new File(dest, "lang");

    String attribute = ard.distAttributeSignature();
    String base = basename((ScalarReflectorDescriptor)ard.getComponentType());

    String classfilename = "Array" + ard.dimension() + attribute + base +".java";
    String interfacefilename = "Section" + ard.dimension() + attribute + base +".java";

    File array = new File(dest, classfilename);
    File section = new File(dest, interfacefilename);

    //    if ((!array.exists())&&(!section.exists())) {
    if (true) {// not generated befor 

      FileOutputStream out0 = new FileOutputStream(array);
      FileOutputStream out1 = new FileOutputStream(section);

      writeArray(out0, ard);
      writeSection(out1, ard);

      out0.close();
      out1.close();

    }
  }
  
  static void generateSection(File srcDir, ScalarReflectorDescriptor srd) throws IOException {
    File dest = new File (srcDir, "hpjava");
    dest = new File(dest, "lang");

    String signature = srd.signature();

    String filename = "Section"+basename(srd)+".java";

    File section = new File(dest, filename);

    //if (!section.exists()) {// not generated befor 
    if (true) {  

      FileOutputStream out = new FileOutputStream(section);
      PrintWriter writer = new PrintWriter(out, true);

      addheader (writer);

      writer.println("package hpjava.lang;");
      writer.println();
      writer.println("public interface Section"+ basename(srd) +" extends Section {");
      writer.println("  public "+srd.name()+"[] dat();");
      writer.println("}");

      out.close();
    }

  }

  static String[] combine(int dim) {
    String [] results = null;
    
    if (dim ==0)
      throw new Error();

    if (dim == 1) {
      results = new String [2];
      results[0]="d";
      results[1]="c";
    }
    else {
      String [] recursive = combine(dim-1);
      results = new String [2*recursive.length];
      
      for (int i =0; i<recursive.length; i++) {
        results[i*2]   =recursive[i]+"d";
        results[i*2+1] =recursive[i]+"c"; 
      }

    }
    
    return results;
  }

  static void generateSectionType (ScalarReflectorDescriptor srd) throws IOException {
    if (srd.isPrimitive()) {
      generateSection(Generator.srcDir, srd);
    }
    else {
      throw new Error();
    }
  }

  static void generateUpto(int dim, PrimitiveTypeDescriptor ptd) throws IOException {
    generateSectionType(ptd);
    
    if (dim == 0) {
      StringBuffer sig = new StringBuffer("]");
      sig.append(String.valueOf(0));
      
      generateArray(Generator.srcDir, new ArrayReflectorDescriptor(ptd, sig.toString()));
    }
    else {
      generateUpto(dim-1, ptd);

      StringBuffer sig = new StringBuffer("]");
      sig.append(String.valueOf(dim));

      String [] str = combine(dim);

      for (int j=0; j < str.length; j++) {
	ArrayReflectorDescriptor ard = new ArrayReflectorDescriptor(ptd, sig.toString()+str[j]);
	generateArray(Generator.srcDir, ard);
      }
    }
  }

  static public void main(String[] argv) throws IOException {
    Properties props = new Properties(System.getProperties());

    //    String os = props.getProperty("os.name");
    //    if (os!=null)
    //      System.out.println(os);


    String images[] = {"boolean", "char", "byte", "short", "int", "long", "float", "double"};

    //String images[] = {"int"};                                                       
    
    PrimitiveTypeDescriptor pdts[] = new PrimitiveTypeDescriptor [images.length];

    for (int i = 0; i<pdts.length; i++) {
      pdts[i] = new PrimitiveTypeDescriptor (images[i]);

      generateUpto (DIM, pdts[i]);
    }
    
  }

}
