/*
 * File         : ArrayReflectorDescriptor.java
 * Author       : Guansong Zhang
 * Created      : Tue May 19 13:07:52 1998
 * Revision     : $Revision: 1.16 $
 * Updated      : $Date: 1999/10/14 09:03:48 $
 */

package hpjava.compiler.reflection;

import java.util.*;
import java.lang.reflect.*;

import hpjava.compiler.util.*;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

public class ArrayReflectorDescriptor extends ClassReflectorDescriptor {
  private ScalarReflectorDescriptor scalarReflector;
  private ArraySignaturesImage arraySignature;

  private ClassReflectorDescriptor crd;
  private boolean isDistArray = false;
  private int dim = 1;

  private char [] distCode;

  public boolean isDistArray() {
    return isDistArray;
  }

  public ClassReflectorDescriptor getComponentType() {
    return crd;
  }

  public int dimension() {
    return dim;
  }

  public boolean isCollapsedDim(int i) {
    if (!this.isDistArray())
      throw new Error();

    if (distCode[i] == 'c')
      return true;
    else
      return false;
  }

  public String distAttributeSignature() {
    if (!this.isDistArray())
      throw new Error();
    
    StringBuffer attribute = new StringBuffer();

    //attribute.append(String.valueOf(this.dimension()));
    
    for (int i = 0; i<this.dimension(); i++) {
      if (this.isCollapsedDim(i)) 
	attribute.append('c');
      else
	attribute.append('d');
    }

    return attribute.toString();   
  }  

  public boolean noCollapsedDim() {
    for (int i = 0 ; i<dimension(); i++)
      if (isCollapsedDim(i))
        return false;

    return true;
  }

  public boolean allCollapsedDim() {
    for (int i = 0 ; i<dimension(); i++)
      if (!isCollapsedDim(i))
        return false;

    return true;
  }

  String substitute(String input) {
    StringBuffer output = new StringBuffer();
    boolean inClassName = false;

    for (int i=0; i<input.length(); i++) {
      if (input.charAt(i) == '[') {
	output.append("_Array");
	continue;
      }
      if (input.charAt(i) == '.') {
	output.append("_");
	continue;
      }
      if (input.charAt(i) == 'L') {
	if (inClassName)
	  output.append("L");
	else {
	  output.append("_");
	  inClassName = true;
	}
	continue;
      }
      if (input.charAt(i) == ';') {
	inClassName=false;
	continue;
      }
      
      output.append(input.charAt(i));
    }

    //System.out.println(output);
    return output.toString();
  }

  String newSignature (String suffix) {
    StringBuffer newSignature = new StringBuffer();

    if (suffix.startsWith("]")) {
      newSignature.append("L");
      newSignature.append("hpjava.lang.Array");
      //newSignature.append("hpjava.lang.Section");
      newSignature.append(suffix.substring(1));
      newSignature.append(substitute(crd.signature));
      newSignature.append(";");
    }
    else if (suffix.startsWith("[")) {
      newSignature.append("[");
      newSignature.append(crd.signature);
    }
    else {
      //System.out.println(suffix);
      throw new Error();
    }

    //System.out.println("crd.signature: "+crd.signature);
    //System.out.println("suffix: "+suffix);
    //Syetem.out.println("new signature: "+newSignature);
    
    return(newSignature.toString());

  }


  public ArrayReflectorDescriptor(ClassReflectorDescriptor crd, String suffix) {
    this.crd = crd;
    
    if (suffix.startsWith("]")) {
      this.isDistArray = true;
      // System.out.println(suffix +":"+ suffix.substring(1,2)+":"+Integer.valueOf(suffix.substring(1,2)));
      this.dim = Integer.valueOf(suffix.substring(1,2)).intValue(); 

      this.distCode = new char [dim];
      
      for (int i =0; i< dim; i++) {
	distCode[i] = suffix.charAt(2+i);
      }
     
    }

    this.signature = newSignature(suffix);
  }
  
  int largestArrayIndex (String image) {
    int index0 = image.lastIndexOf('[');
    int index1 = image.lastIndexOf(']');
    
    return  (index0>index1? index0: index1);
  }

  public ArrayReflectorDescriptor(ScalarReflectorDescriptor srd, ArraySignaturesImage asi) {  

    if (asi.name.length() == 0)
      throw new Error();

    //String temp = javaClassNameFor(srd.signature, asi.name);  

    this.crd = srd;

    //    System.out.println("srd.signature in ARD.java: " + srd.signature);
    //    System.out.println("srd.name() in ARD.java: " + srd.name());

    String image = asi.name;

    //    System.out.println("asi.name in ArrayRefDesc.java: " + image);

    int index = largestArrayIndex(image);

    while (index != 0) {
      this.crd = new ArrayReflectorDescriptor(crd, image.substring(index));
      image = image.substring(0,index);

      index = largestArrayIndex(image);
    }

    // the following lined repeated with the other constructor
    // change them later.

    String suffix = image;
    
    if (image.startsWith("]")) {
      this.isDistArray = true;

      this.dim = Integer.valueOf(suffix.substring(1,2)).intValue();  
 
      this.distCode = new char [dim]; 
       
      for (int i =0; i< dim; i++) { 
        distCode[i] = suffix.charAt(2+i); 
      } 

    }

    this.signature= newSignature(image);

    //System.out.println("----"+this.signature);
    //System.out.println("++++"+temp);
 
  }

  String javaClassNameFor(String scalarReflectorName, String arraySignature) {
  
    StringBuffer base = new StringBuffer(scalarReflectorName);
    StringBuffer aSignature = new StringBuffer(arraySignature);

    //System.out.println(asd.name+btd.name);
    
    StringBuffer nameBuffer = new StringBuffer();
    
    boolean hpj = false;
    
    for (int i=0; i<aSignature.length(); i++) {
      if (aSignature.charAt(i) == '[') {
        if (hpj)
          nameBuffer.append("a");
        else
          nameBuffer.append("[");
      }    
          
      if (aSignature.charAt(i) == ']' ) {
        if (hpj)
          nameBuffer.append("_hpjava_lang_Array");
        else {
          hpj=true;
          //nameBuffer.append("Lhpjava.lang.Array");   
          nameBuffer.append("Lhpjava.lang.Section");   
        }
        
        i++;
        
        int j=i;
        while ((j<aSignature.length()) && (aSignature.charAt(j)!='[' && aSignature.charAt(j)!=']')){
          nameBuffer.append(aSignature.charAt(j));
          j++;
        }
        //System.out.println(j);
        i=j-1;                  
      }
    }
    
    if (hpj) {
      if (base.charAt(0)=='Z') {
        nameBuffer.append("Boolean");
      }
      if (base.charAt(0)=='C') {
        nameBuffer.append("Char");
      }
      if (base.charAt(0)=='B') {
        nameBuffer.append("Byte");
      }
      if (base.charAt(0)=='S') {
        nameBuffer.append("Short");
      }
      if (base.charAt(0)=='I') {
        nameBuffer.append("Int");
      }
      if (base.charAt(0)=='J') {
        nameBuffer.append("Long");
      }
      if (base.charAt(0)=='F') {
        nameBuffer.append("Float");
      }
      if (base.charAt(0)=='D') {
        nameBuffer.append("Double");
      }
      if (base.charAt(0)=='L') {
        nameBuffer.append("_");
        for (int i =1; i< base.length()-1; i++) {
          if (base.charAt(i) != '.')
            nameBuffer.append(base.charAt(i));
          else
            nameBuffer.append('_');        
        }
      }
      nameBuffer.append(';');            
    }
    else {
      nameBuffer.append(base);    
    }
    
    return nameBuffer.toString();
  }

}
