/*
 * File         : ArraySignaturesImage.java
 * Author       : Guansong Zhang
 * Created      : Tue May 19 13:07:52 1998
 * Revision     : $Revision: 1.6 $
 * Updated      : $Date: 1999/03/16 20:27:28 $
 */

package hpjava.compiler.reflection;

import java.util.*;
import java.lang.reflect.*;

import hpjava.compiler.util.*;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

import hpjava.compiler.abstractexp.*;

import hpjava.compiler.unparser.*;

public class ArraySignaturesImage {
  public String name;

  public ArraySignaturesImage(String name) {
    this.name = name;
  }
  
  // this is for ArrySignature list only. Change this later.
  public ArraySignaturesImage(NodeListOptional n) {
    ArraySignaturesResolver resolver = new ArraySignaturesResolver(this);

    resolver.status.push(ArraySignaturesResolver.FORARRAYSIGNATURE);

    n.accept(resolver);  

    resolver.status.pop();

    this.name = resolver.nameBuffer.toString();

  } 

  public ArraySignaturesImage(ArraySignature n) {
    ArraySignaturesResolver resolver = new ArraySignaturesResolver(this);

    n.accept(resolver);  

    this.name = resolver.nameBuffer.toString();
  } 

  /*
  // this one is used for allocation constructor of DAD.
  public ArraySignaturesImage(DistArrayShapesAndGroup n) {
    ArraySignaturesResolver resolver = new ArraySignaturesResolver(this);

    n.accept(resolver);  

    this.name = resolver.nameBuffer.toString();
  }   
  */

  public ArraySignaturesImage(ArraySignaturesImage image0,
                              ArraySignaturesImage image1) {

    name=image0.name+image1.name;
  } 
  
  public boolean isEmpty() {
    if (name.length()==0)
      return true;
    else
      return false;  
  }

  public String toString() {
    return name;
  }
  
}

/* 
  The resolver starting from NodeListOptional 
  for multidimention array
*/

/*
 * File         : ArraySignaturesResolver.java
 * Author       : Guansong Zhang
 * Created      : Tue May 19 13:07:52 1998
 * Revision     : $Revision: 1.6 $
 * Updated      : $Date: 1999/03/16 20:27:28 $
 */


// used only for class name in source code 

class ArraySignaturesResolver extends DepthFirstVisitor {
  
  ArraySignaturesImage image;

  public StringBuffer nameBuffer = new StringBuffer();  

  protected IntStack status = new IntStack();
  
  protected ClassReflector cr = null;

  static final int DEFAULT  = 0;
  static final int FORARRAYSIGNATURE =1;
  static final int INDISTARRAYDIMS = 2;
  
  ArraySignaturesResolver(ArraySignaturesImage image) {
    this.image = image;

    status.push(DEFAULT);
  }
  
  public void visit(NodeListOptional n) {
    if(status.peek() == FORARRAYSIGNATURE) {
      if ( n.present() ) {
        for ( Enumeration e = n.elements(); e.hasMoreElements(); ) {
          Node node = (Node) e.nextElement();
        
          if (node instanceof ArraySignature) {
            visit((ArraySignature)node);
          }
          else {
            throw new Error("Not ArraySignature.");
          }
        }
      }
    }
    else {
      // original one;
      
      super.visit(n);
      
    }
    
  }
  

   //
   // f0 -> "[" "]"
   //       | "[" "[" DistArrayDimAttList() "]" "]"
   //       | "#"
   //
   public void visit(ArraySignature n) {
      switch(n.f0.which) {
        case 0:
          nameBuffer.append("[");
          break;
        case 1:
          nameBuffer.append("]");
          n.f0.accept(this);
          break;
        case 2: 
          nameBuffer.append("]0");   
          break;
      }   
   }
  
   //
   // f0 -> [ "*" ]
   // f1 -> ( "," [ "*" ] )*
   //
   public void visit(DistArrayDimAttList n) {
      int counter=1;
      if (n.f1.present()) {    
        counter+=n.f1.size();
      }  
      
      nameBuffer.append(counter);
      
      if(n.f0.present()) { 
        nameBuffer.append("c");
      }
      else{
        nameBuffer.append("d");
      }

      if (n.f1.present()) {    
        for (int i = 0; i<n.f1.size(); i++) {
          NodeSequence node = (NodeSequence) n.f1.elementAt(i);
          if (( (NodeOptional)node.elementAt(1) ).present()) {
            nameBuffer.append("c");
          }
          else {
            nameBuffer.append("d");
          }
        }        
      }
   }
    
  /*
   //
   // f0 -> ( ( "[" "[" Expression() ( "," Expression() )* "]" "]" )+ | "#" )
   // f1 -> ( ArraySignature() )*
   // f2 -> [ "on" GroupExpression() ]
   //
  public void visit (DistArrayShapesAndGroup n) {
    if (n.f0.which ==1) {
      nameBuffer.append("]0");
    }
    else {
      nameBuffer.append("]");

      status.push(INDISTARRAYDIMS);
      n.f0.accept(this);
      status.pop();
    }

    if (n.f1.present()) {
      ArraySignaturesImage image = new ArraySignaturesImage(n.f1);
      nameBuffer.append(image.name);
    }

  }
  */

  // "[" "[" Expression() ( "," Expression() )* "]" "]"
  public void visit (NodeSequence n) {
    if (status.peek()==INDISTARRAYDIMS) {
      int counter =1;
      if (((NodeListOptional)n.elementAt(3)).present()) {
        counter += ( (NodeListOptional)n.elementAt(3) ).size();
      }

      nameBuffer.append(counter);
      
      AbstractExpression aexp = 
        ExpressionManager.lookup((Expression)n.elementAt(2));
      
      ClassReflector reflector = aexp.getClassReflector();

      if (reflector == null) { throw new Error("Not finished.");}

      if (reflector.getName().equals("int")) { 
        nameBuffer.append('c');
      }
      else if (reflector.getName().equals("hpjava.lang.Range")){
        nameBuffer.append('d');
      }
      else {
        throw new Error("Type checking should ensure this doesnot happend:" 
                        +reflector.getName());
      }

      if ( ( (NodeListOptional)n.elementAt(3) ).present()) {
        NodeListOptional listOptional = 
          (NodeListOptional)n.elementAt(3); 

        for (int i = 0; i<listOptional.size(); i++) {
          NodeSequence sequence = (NodeSequence) listOptional.elementAt(i);

          aexp = ExpressionManager.lookup((Expression)sequence.elementAt(1));
          reflector = aexp.getClassReflector();

          if (reflector == null) {throw new Error("Not finished.");}

          if (reflector.getName().equals("int")) { 
            nameBuffer.append('c');
          }
          else if (reflector.getName().equals("hpjava.lang.Range")){
            nameBuffer.append('d');
          }
          else {
            throw new Error("Type checking should ensure this doesnot happend:" 
                            +reflector.getName());
          }

        }
      }
    }
    else {

      // original visit;

      super.visit(n);
    }

  }

}
