/*
 * File         : ClassReflector.java
 * Author       : Guansong Zhang, Han-Ku Lee
 * Created      : Mon May 18 17:20:06 1998
 * Revision     : $Revision: 1.11 $
 * Updated      : $Date: 1999/12/01 19:06:47 $
 */

package hpjava.compiler.reflection;

import java.io.*;
import java.util.*;
import java.lang.reflect.*;

abstract public class ClassReflector {
  
  //  private ClassReflectorDescriptor crd;
  private static ClassReflector cr = null;

  //abstract public ClassReflector getDeclaringClass();


  public static ClassReflector forName(String className)
    throws ClassNotFoundException {

    cr = ClassReflectorManager.lookup(className);

    /*
    // cr can not be null ...
    if(cr == null){
      try{
        Class cls = Class.forName(className);
        cr = new ClassReflectorByByteCode(cls);
      }
      catch(ClassNotFoundException e){
        throw new Error(className + " Class can not be found");
        //System.out.println(className + " Class can not be found");
      }
    }
    */
    return cr;
  }

  abstract public ClassReflector getComponentType();

  abstract public ConstructorReflector getConstructor(ClassReflector [] paratypeName)
    throws NoSuchMethodException, SecurityException ;

  abstract public ConstructorReflector[] getConstructors()
    throws SecurityException ;

  abstract public FieldReflector getDeclaredField(String name)
    throws NoSuchFieldException, SecurityException ;

  abstract public FieldReflector[] getDeclaredFields()
    throws SecurityException ;

  abstract public MethodReflector getDeclaredMethod(String name, ClassReflector [] paratypeName) 
    throws NoSuchMethodException, SecurityException ;

  abstract public MethodReflector[] getDeclaredMethods()
    throws SecurityException ;

  //abstract public MethodReflector getMethodOfInterface(String name, ClassReflector [] paratypeName) 
  //  throws NoSuchMethodException, SecurityException ;

  //abstract public String getSuperclassName();
  //abstract public String getSuperclass();
  abstract public ClassReflector getSuperclass();

  //abstract public String[] getInterfaceNames();
  abstract public ClassReflector[] getInterfaces();

  abstract public ModifierReflector getModifier();

  abstract public String getName();

  abstract public boolean isArray();

  abstract public boolean isDistArray();
  
  abstract public boolean isInterface();

  abstract public boolean isPrimitive();

  protected String BOOLEAN = "boolean";
  protected String BYTE = "byte";
  protected String CHAR = "char";
  protected String SHORT = "short";
  protected String INT = "int";
  protected String LONG = "long";
  protected String FLOAT = "float";
  protected String DOUBLE = "double";
  protected String VOID = "void";

  public MethodReflector getMethodOfInterface(String name, ClassReflector [] paratypeName) throws NoSuchMethodException {
    MethodReflector m = null;
    
    ClassReflector[] tmp = this.getInterfaces();

    //for(int i=0; i<tmp.length; i++){
    //  System.out.println("tmp[i].getName(): " + tmp[i].getName());
    //}

    if(tmp == null)
      throw new NoSuchMethodException("Method " + name + " is NOT found in super Interfaces ...");
    else{
      
      while(tmp != null){
        if(tmp.length == 0)
          break;

        for(int i=0; i<tmp.length; i++){
          try{
            m = tmp[i].getDeclaredMethod(name, paratypeName);
            //System.out.println("Method " + name + " is FOUND  in " + tmp[i].getName() + " in getMethodOfInterface in CRBS ...");
            tmp = null;
            break;
          }
          catch(NoSuchMethodException nsm){
            System.out.println("Method " + name + " is NOT FOUND in " + tmp[i].getName() + " in getMethodOfInterface in CRBS ...");
          }

        }

        if(tmp != null){
          //System.out.println("tmp.length = " + tmp.length);
          tmp = tmp[0].getInterfaces();
        }
      }

    }

    if(m == null){
      //System.out.println("m is NULL ...");
      throw new NoSuchMethodException("Method " + name + " is NOT found in super Interfaces ...");
    }
    //System.out.println("m is " + m.getDeclaringClass().getName());
    return m;
  }

  public boolean isSuperclassOf(ClassReflector sub_class){
    boolean tmp = false;
    ClassReflector super_class = null;

    if(sub_class.isPrimitive()){
      if(this.isPrimitive()){
        //tmp = true;    // Don't forget about Primitive conversion !!!

				//System.out.println("############ this.getName(): " + this.getName());
        //System.out.println("############ sub_class.getName(): " + sub_class.getName());

        if(sub_class.getName().equals(BYTE)){
          if(this.getName().equals(SHORT) || this.getName().equals(INT)
             || this.getName().equals(LONG) || this.getName().equals(FLOAT)
             || this.getName().equals(DOUBLE) || this.getName().equals(BYTE)){
            tmp = true;
          }
        }
        else if(sub_class.getName().equals(SHORT) || sub_class.getName().equals(CHAR)){
          if(this.getName().equals(INT) || this.getName().equals(LONG)
             || this.getName().equals(FLOAT) || this.getName().equals(DOUBLE)
             || this.getName().equals(SHORT) || this.getName().equals(CHAR)){
            tmp = true;
          }
        }
        else if(sub_class.getName().equals(INT)){
          if(this.getName().equals(LONG) || this.getName().equals(FLOAT) 
             || this.getName().equals(DOUBLE) || this.getName().equals(INT)){
            tmp = true;
          }
        }
        else if(sub_class.getName().equals(LONG)){
          if(this.getName().equals(FLOAT) || this.getName().equals(DOUBLE)
             || this.getName().equals(LONG)){
            tmp = true;
          }
        }
        else if(sub_class.getName().equals(FLOAT)){
          if(this.getName().equals(DOUBLE) || this.getName().equals(FLOAT)){
            tmp = true;
          }
        }
      }
    }
    else{
      if(this.getName().equals("java.lang.Object"))
        tmp = true;
      else if(sub_class.getName().equals("hpjava.compiler.abstractexp.NullNode")){
        tmp = true;
      }
      else if(this.getName().equals(sub_class.getName())){
        tmp = true;
      }
      else{
        ClassReflector tmp_super = this;

        //System.out.println("############ this.getName():" + this.getName());
        /*
        if(this.isInterface())
          System.out.println("############ this is an interface ...");

        System.out.println("############ tmp_super:" + tmp_super);

        ClassReflector[] str1 = tmp_super.getInterfaces();
        if(str1 == null)
          System.out.println("############ str1 is NULL");
        else
          for(int i=0; i<str1.length; i++)
            System.out.println("############ str1: " + str1[i].getName());
        */

        if(tmp_super.getName().endsWith(";")  && sub_class.getName().endsWith(";")){
          try{
            String tmp_this = tmp_super.getName().substring(tmp_super.getName().lastIndexOf("[")+2, 
                                                     tmp_super.getName().lastIndexOf(";"));
            String tmp_sub = sub_class.getName().substring(sub_class.getName().lastIndexOf("[")+2,
                                                         sub_class.getName().lastIndexOf(";"));

            //System.out.println("############ tmp_this: " + tmp_this);
            //System.out.println("############ tmp_sub: " + tmp_sub);

            tmp_super = ClassReflector.forName(tmp_this);
            sub_class = ClassReflector.forName(tmp_sub);
          }
          catch(ClassNotFoundException cnf){
            System.out.println("ClassNotFoundException occurred in ClassReflector ...");
          }
        }

        while(sub_class != null){
          //System.out.println("sub_class.getName() in isSuperclassOf(): " + sub_class.getName());
          if(sub_class.getSuperclass() == null){
            //System.out.println("... sub_class.getSuperclass() is null ...");
            break;
          }
          else{
            try{
              //Class cls = Class.forName(sub_class.getSuperclass().getName());
              if(sub_class.getSuperclass() != null)
                super_class = ClassReflector.forName(sub_class.getSuperclass().getName());
              //super_class = new ClassReflectorByByteCode(cls);

              //System.out.println("&&& sub_class.getName(): " + sub_class.getName());
              //System.out.println("&&& tmp_super.getName(): " + tmp_super.getName());
              //System.out.println("&&& super_class.getName(): " + super_class.getName());
              if(tmp_super.getName() == super_class.getName()){
                sub_class = null;
                tmp = true;
                //System.out.println("They are EQUAL ...");
              }
              else{
                sub_class = super_class;
                tmp = false;
                //System.out.println("They are NOT EQUAL ...");
              }

            }
            catch(ClassNotFoundException cnf){
              System.out.println("Super class Not Found in isSuperclassOf ...");
            }

          }
        }
      }
    }

    //if(tmp)
    //  System.out.println("tmp is TRUE in isSuperclassOf ...");
    //else
    //  System.out.println("tmp is FALSE in isSuperclassOf ...");

    return tmp;
  }

  public boolean isSubclassOf(ClassReflector super_class){
    return super_class.isSuperclassOf(this);
  }

  public boolean isInterfaceOf(ClassReflector sub){
    boolean flag = false;

    Vector tmp_interface = new Vector();

    ClassReflector[] tmp = sub.getInterfaces();
    ClassReflector tmp_supclass = sub.getSuperclass();

    //System.out.println("######### this.getName() in isInterfaceOf: " + this.getName());
    //System.out.println("######### sub.getName() in isInterfaceOf: " + sub.getName());

    if(tmp != null)
      for(int i=0; i<tmp.length; i++)
        tmp_interface.addElement(tmp[i]);
      
    while(tmp_interface.size() != 0 || tmp_supclass != null){

      while(tmp_interface.size() != 0){
        for(int i=0; i<tmp_interface.size(); i++){
          //System.out.println("tmp_interface.elementAt(i).getName() in isInterfaceOf: " 
          //                   + ((ClassReflector)tmp_interface.elementAt(i)).getName());
          if(this.getName().equals(((ClassReflector)tmp_interface.elementAt(i)).getName())){
            //System.out.println("******* this.getName() is equal to tmp_interface.elementAt(i).getName()");
            //n.cr = n.right.cr;
            flag = true;
            tmp_interface.removeAllElements();
            break;
          }

        }

        Vector new_interface = new Vector();

        for(int i=0; i<tmp_interface.size(); i++){
          ClassReflector[] _tmp = ((ClassReflector)tmp_interface.elementAt(i)).getInterfaces();

          if(_tmp != null)
            for(int j=0; j<_tmp.length; j++){
              new_interface.addElement(_tmp[j]);
            }
        }

        tmp_interface.removeAllElements();
        tmp_interface = new_interface;

      }

      if(flag) break;

      if(tmp_supclass != null){
        //System.out.println("tmp_supclass.getName(): " + tmp_supclass.getName());

        if(this.getName().equals(tmp_supclass.getName())){
          //System.out.println("******* n.left.cr.getName() is equal to tmp_supclass.getName()");
          //n.cr = n.right.cr;
          flag = true;
          break;
        }
        else{
          Vector new_interface = new Vector();
          ClassReflector[] _tmp = tmp_supclass.getInterfaces();

          for(int j=0; j<_tmp.length; j++){
            new_interface.addElement(_tmp[j]);
          }
            
          tmp_interface = new_interface;
          tmp_supclass = tmp_supclass.getSuperclass();

        }

      }

    }

    return flag;
  }

  abstract protected void showConstructors(String space, PrintStream out);
  abstract protected void showDeclaredFields(String space, PrintStream out);
  abstract protected void showDeclaredMethods(String space, PrintStream out);

  public void show(PrintStream out) {
    out.println("CLASS:");    
    
    if (isInterface())
      out.println("  NAME(I):\t"+ getName());
    else
      out.println("  NAME(C):\t"+ getName());

    ModifierReflector mr=getModifier();
    if(mr!=null)
      mr.show("  ", out);
      
    if (getSuperclass()!=null)
      out.println("  SUPERCLASS:\t"    + getSuperclass());

    ClassReflector [] InterfaceNames = getInterfaces();
    if(InterfaceNames!=null) {
      out.print("  INTERFACES:\t") ;
      for (int i=0; i<InterfaceNames.length; i++) 
        out.print(InterfaceNames[i]+" ");
      out.println();  
    }          

    showConstructors("  ", out);
    showDeclaredFields("  ",out);
    showDeclaredMethods("  ", out);
  } 

}
