/*
 * File         : ClassReflectorByByteCode.java
 * Author       : Guansong Zhang, Han-Ku Lee
 * Created      : Fri May 22 11:23:17 1998
 * Revision     : $Revision: 1.6 $
 * Updated      : $Date: 1999/12/01 19:06:48 $
 */

package hpjava.compiler.reflection;

import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import hpjava.compiler.abstractexp.*;

public class ClassReflectorByByteCode extends ClassReflector {

  protected Class cls;

  ExpressionConversion_reflection EC = new ExpressionConversion_reflection();
  // a public method for creating reflector from Java reflection.

  public ClassReflectorByByteCode (Class cls) {
    this.cls=cls;
  }

  public ClassReflector getComponentType() {
    return new ClassReflectorByByteCode(cls.getComponentType());
  }


  public ConstructorReflector getConstructor(ClassReflector[] paratypeName)
    throws NoSuchMethodException, SecurityException {

    Class [] paratype = null;

    if (paratypeName != null) {
      paratype = new Class [paratypeName.length];

      try{
        for(int i = 0; i< paratypeName.length; i++){
          paratype[i] = Clazz.forName(paratypeName[i].getName());

          //System.out.println("paratype[i].getName(): " + paratype[i].getName());
        }
      }
      catch (ClassNotFoundException e) {
        // Make sure the class for the name already found? performance
        System.err.println(e.getMessage());
      }
    }
    else
      paratype = null;

    return ( new ConstructorReflectorByteCode (cls.getConstructor(paratype)) );
  }

  public ConstructorReflector[] getConstructors()
    throws SecurityException {
    Constructor [] cList = cls.getConstructors();
    if (cList == null)
      return null;
    else {
      ConstructorReflector [] crList = new ConstructorReflector[cList.length];
      for (int i=0; i<crList.length; i++)
        crList[i]= new ConstructorReflectorByteCode(cList[i]);

      return crList;
    }
  }

  public FieldReflector getDeclaredField(String name)
    throws NoSuchFieldException, SecurityException  {
    return (new FieldReflectorByteCode(cls.getDeclaredField(name)));
  }

  public FieldReflector[] getDeclaredFields() throws SecurityException {
    Field [] fList = cls.getDeclaredFields();
    if (fList == null)
      return null;
    else {
      FieldReflector [] frList = new FieldReflector[fList.length];
      for (int i=0; i<frList.length; i++)
        frList[i]= new FieldReflectorByteCode(fList[i]);

      return frList;
    }
  }

  public MethodReflector getDeclaredMethod(String name, ClassReflector [] paratypeName)
    throws NoSuchMethodException, SecurityException {

    ExpressionConversion_reflection EC = new ExpressionConversion_reflection();

    //System.out.println("$$$$$$$$$ In ClassReflectorByByteCode.java $$$$$$$$$");

    Vector signature = new Vector();

    if (paratypeName!=null)
      for (int i=0; i<paratypeName.length; i++){
          signature.addElement(paratypeName[i].getName());
      }

    //System.out.println("cls.getName(): " + cls.getName());
    Method[] mList = cls.getDeclaredMethods();
    Vector vec = new Vector();    
    Vector applicable_vec = new Vector();    

    //System.out.println("&&&&&&&&&&&&&& mList.length: " + mList.length);

    for(int i=0; i<mList.length; i++){
      //System.out.println("&&&&&&&&&&&&&& mList[i]'s name: " + mList[i].getName());
      Class[] cl = mList[i].getParameterTypes();
      
      //System.out.println("&&&&&&&&&&&&&& name: " + name);
      //System.out.println("&&&&&&&&&&&&&& cl.length: " + cl.length);
      //System.out.println();
      //System.out.println("&&&&&&&&&&&&&& paratypeName.length: " + paratypeName.length);
      //System.out.println();

      if(mList[i].getName().equals(name) && cl.length == paratypeName.length){
        vec.addElement(mList[i]);
        //System.out.println("&&&&&&&&&&&&&& vec: " + mList[i].getName());
      }
    }

    //System.out.println("&&&&&&&&&&&&&& vec.size(): " + vec.size());

    Method tmp = null;
    int num_of_maximal_method = 0;

    boolean isApplicable = false;

    for(int i=0; i<vec.size(); i++){
      Method tmp_method = (Method)vec.elementAt(i);
      Class[] cl = tmp_method.getParameterTypes();
      //System.out.println();
      //System.out.println("******** cl.length: " + cl.length);

      if(cl.length == 0){
        isApplicable = true;
      }
      else{
        try{
          for(int j=0; j<tmp_method.getParameterTypes().length; j++){
            //System.out.println("******** cl[j].getName(): " + cl[j].getName());
            //System.out.println("******** paratypeName[j].getName(): " + paratypeName[j].getName());
            if(cl[j].getName().equals(paratypeName[j].getName())){
              isApplicable = true;
              //System.out.println("isApplicable is TRUE");
            }
            else if(ClassReflector.forName(cl[j].getName()).isSuperclassOf(ClassReflector.forName(paratypeName[j].getName()))){
              isApplicable = true;
              //System.out.println("isApplicable is TRUE");
            }
            else if(ClassReflector.forName(cl[j].getName()).isInterfaceOf(ClassReflector.forName(paratypeName[j].getName()))){
              isApplicable = true;
              //System.out.println("isApplicable is TRUE");
            }
            else if((ClassReflector.forName(cl[j].getName()).isInterface()
                     || ClassReflector.forName(cl[j].getName()).isArray()
                     || !ClassReflector.forName(cl[j].getName()).isInterface()) 
                    && ClassReflector.forName(paratypeName[j].getName()).equals("hpjava.compiler.abstractexp.NullNode")){
              isApplicable = true;
              //System.out.println("isApplicable is TRUE");
            }
            else if(ClassReflector.forName(cl[j].getName()).equals("java.lang.Object") 
                    && ClassReflector.forName(paratypeName[j].getName()).isInterface()){
              isApplicable = true;
              //System.out.println("isApplicable is TRUE");
            }
            else if(ClassReflector.forName(cl[j].getName()).equals("java.lang.Object") 
                    && ClassReflector.forName(paratypeName[j].getName()).isArray()){
              isApplicable = true;
              //System.out.println("isApplicable is TRUE");
            }
            else if(ClassReflector.forName(cl[j].getName()).equals("java.lang.Cloneable") 
                    && ClassReflector.forName(paratypeName[j].getName()).isArray()){
              isApplicable = true;
              //System.out.println("isApplicable is TRUE");
            }
            else if(ClassReflector.forName(cl[j].getName()).isArray() && ClassReflector.forName(paratypeName[j].getName()).isArray()){

              if(ClassReflector.forName(cl[j].getName()).getComponentType().getName().equals(ClassReflector.forName(paratypeName[j].getName()).getComponentType().getName())){
                isApplicable = true;
                //System.out.println("isApplicable is TRUE");
              }
              else if(ClassReflector.forName(cl[j].getName()).getComponentType().isSuperclassOf(ClassReflector.forName(paratypeName[j].getName()).getComponentType())){
                isApplicable = true;
                //System.out.println("isApplicable is TRUE");
              }
              else if(ClassReflector.forName(cl[j].getName()).getComponentType().isInterfaceOf(ClassReflector.forName(paratypeName[j].getName()).getComponentType())){
                isApplicable = true;
                //System.out.println("isApplicable is TRUE");
              }
              else if(ClassReflector.forName(cl[j].getName()).getComponentType().isInterface()
                      || ClassReflector.forName(cl[j].getName()).getComponentType().isArray()
                      || !ClassReflector.forName(cl[j].getName()).getComponentType().isInterface()){
                if(ClassReflector.forName(paratypeName[j].getName()).getComponentType().getName().equals("hpjava.compiler.abstractexp.NullNode")){
                  isApplicable = true;
                  //System.out.println("isApplicable is TRUE");
                }
                
              }
              else if(ClassReflector.forName(cl[j].getName()).getComponentType().getName().equals("java.lang.Object") 
                      && ClassReflector.forName(paratypeName[j].getName()).getComponentType().isInterface()){
                isApplicable = true;
                //System.out.println("isApplicable is TRUE");
              }
              else if(ClassReflector.forName(cl[j].getName()).getComponentType().getName().equals("java.lang.Object") 
                      && ClassReflector.forName(paratypeName[j].getName()).getComponentType().isArray()){
                isApplicable = true;
                //System.out.println("isApplicable is TRUE");
              }
              else if(ClassReflector.forName(cl[j].getName()).getComponentType().getName().equals("java.lang.Cloneable") 
                      && ClassReflector.forName(paratypeName[j].getName()).getComponentType().isInterface()){
                isApplicable = true;
                //System.out.println("isApplicable is TRUE");
              }

            }
            else{
              isApplicable = false;
              //System.out.println("isApplicable is FALSE");
              break;
            } 

          }
        }
        catch(ClassNotFoundException cnf){
          System.out.println("ClassNotFoundException occurred ...");
        }
      }
      if(isApplicable)
        applicable_vec.addElement(tmp_method);
    }

    //System.out.println("*********** applicable_vec.size(): " + applicable_vec.size());

    int size2 = 0;

    if(applicable_vec != null)
      size2 = applicable_vec.size();

    //for(int i=0; i<applicable_vec.size(); i++){
    for(int i=0; i<size2; i++){
      //System.out.println("index: " + i);
      if(EC.isMaximallySpecific(applicable_vec, i)){
        //System.out.println("Source is Maximally Specific ...");
        num_of_maximal_method++;
        tmp = (Method)applicable_vec.elementAt(i);
      }
      //System.out.println("num_of_maximal_method: " + num_of_maximal_method);
    }

    //if(tmp == null)
    //  System.out.println("&&&& tmp is NULL ...");
    //else
    //  System.out.println("&&&& tmp is NOT NULL ...");

    //System.out.println("@@@@@@@ num_of_maximal_method: " + num_of_maximal_method);

    if(num_of_maximal_method == 0)
      throw new NoSuchMethodException(name + signature.toString());
    else if(num_of_maximal_method != 1)
      throw new Error("There are more than one maximally specific methods of " + name);

    //System.out.println("******* Name of the method: " + tmp.getName());
    //System.out.println("******* Name of the DeclaringClass: " + tmp.getDeclaringClass().getName());
    //for(int i=0; i<tmp.getParameterTypes().length; i++)
    //  System.out.println("******* " + i + "th arg: " + tmp.getParameterTypes()[i].getName());

    return (new MethodReflectorByteCode(tmp));

  }

  public MethodReflector[] getDeclaredMethods() throws SecurityException {
    Method [] mList = cls.getDeclaredMethods();
    if (mList == null)
      return null;
    else {
      MethodReflector [] mrList = new MethodReflector[mList.length];
      for (int i=0; i<mrList.length; i++)
        mrList[i]= new MethodReflectorByteCode(mList[i]);

      return mrList;
    }
  }

  //public String getSuperclass() {
  public ClassReflector getSuperclass() {
    //return cls.getSuperclass().getName();

    //System.out.println("cls in CLSRBB: " + cls);

    if(cls.getSuperclass() == null)
      return null;
    else
      return new ClassReflectorByByteCode(cls.getSuperclass());
  }

    //public String[] getInterfaceNames() {
  public ClassReflector[] getInterfaces() {
    Class [] interfaces = cls.getInterfaces();

    //String [] names;
    ClassReflector[] names;

    if (interfaces == null)
      names = null;
    else {
      names = new ClassReflector[interfaces.length];
      for (int i=0; i<interfaces.length; i++){
          //names[i]=interfaces[i].getName();
        try{
          names[i] = ClassReflector.forName(interfaces[i].getName());
        }
        catch(ClassNotFoundException cnf){
          System.out.println("Interface Not Found ...");
        }
      }
    }

    return names;
  }

  public ModifierReflector getModifier() {
    return new ModifierReflector(this.cls.getModifiers());
  }

  public String getName() {
    return this.cls.getName();
  }

  public boolean isArray() {
    return cls.isArray();
  }

  public boolean isDistArray() {
    if (this.cls.getName().startsWith("hpjava.lang.Array"))
      return true;

    if (this.cls.getName().startsWith("hpjava.lang.Section"))
      return true;

    return false;
  }

  public boolean isInterface() {
    return cls.isInterface();
  }

  public boolean isPrimitive() {
    return cls.isPrimitive();
  }

  protected void showConstructors(String space, PrintStream out) {
    ConstructorReflector [] crList = getConstructors();
    if (crList!=null) {
      for (int i=0; i<crList.length; i++)
        crList[i].show(space, out);
    }

  }

  protected void showDeclaredFields(String space, PrintStream out) {

  }

  protected void showDeclaredMethods(String space, PrintStream out) {

  }
}
