/*
 * File         : ClassReflectorByDescriptor.java
 * Author       : Guansong Zhang, Han-Ku Lee
 * Created      : Wed Oct  7 15:05:38 1998
 * Revision     : $Revision: 1.11 $
 * Updated      : $Date: 1999/12/01 19:06:49 $
 */

package hpjava.compiler.reflection;

import java.io.*;
import java.lang.reflect.*;

public class ClassReflectorByDescriptor extends ClassReflector {

  private ClassReflectorDescriptor crd;

  private ClassReflector cr = null;
  private ClassReflector component = null;

  private boolean isArray = false;
  private boolean isDistArray = false;

  public ClassReflectorByDescriptor(ClassReflectorDescriptor crd) {
    this.crd=crd;

    if (crd.isArray()) {
      //System.out.println("##### crd in CRBD: " + crd);

      this.isArray = true;

      component = new ClassReflectorByDescriptor(crd.getComponentType());

      if (crd.isDistArray()) {
        this.isDistArray = true;

        // a problem here, [[]] of user defined type including local classes.
        // hpjava.compiler.gar.Generator.generate((ArrayReflectorDescriptor)crd);
      }
    }
    else {
      this.cr = ClassReflectorManager.lookup((ScalarReflectorDescriptor)crd);
    }
  }

  public ClassReflector getComponentType() {
    return component;
  }

  public ConstructorReflector getConstructor(ClassReflector[] paratypeName)
    throws NoSuchMethodException, SecurityException {

    if (cr!=null)
      return cr.getConstructor(paratypeName);
    else
      return null;
  }

  public ConstructorReflector[] getConstructors()
    throws SecurityException {

    if (cr!=null)
      return cr.getConstructors();
    else
      return null;
  }

  public FieldReflector getDeclaredField(String name)
    throws NoSuchFieldException, SecurityException {

    if (cr!=null)
      return cr.getDeclaredField(name);
    else
      return null;
  }

  public FieldReflector[] getDeclaredFields()
    throws SecurityException {

    if (cr!=null)
      return cr.getDeclaredFields();
    else
      return null;
  }

  public MethodReflector getDeclaredMethod(String name, ClassReflector[] paratypeName)
    throws NoSuchMethodException, SecurityException {

    if (cr!=null)
      return cr.getDeclaredMethod(name, paratypeName);
    else
      return null;
  }

  public MethodReflector[] getDeclaredMethods()
    throws SecurityException {

    if (cr!=null)
      return cr.getDeclaredMethods();
    else
      return null;
  }

  //public String getSuperclass() {
  public ClassReflector getSuperclass() {
    if (cr!=null)
      return cr.getSuperclass();
    else
      return null;
  }

    //public String[] getInterfaceNames(){
  public ClassReflector[] getInterfaces(){
    if (cr!=null)
      return cr.getInterfaces();
    else
      return null;
  }

  public ModifierReflector getModifier() {
    if (cr!=null)
      return cr.getModifier();
    else
      return null;
  }

  public String getName() {
    if (cr!=null)
      return cr.getName();
    else 
      return crd.getName();
  }

  public boolean isArray() {
    if (cr!=null)
      return cr.isArray();
    else
      return this.isArray;
  }

  public boolean isDistArray() {
    if (cr!=null)
      return cr.isDistArray();
    else
      return this.isDistArray;
  }

  public boolean isInterface() {
    if (cr!=null)
      return cr.isInterface();
    else
      return false;
  }

  public boolean isPrimitive() {
    if (cr!=null)
      return cr.isPrimitive();
    else
      return false;
  }

  public ClassReflectorDescriptor descriptor() {
    return crd;
  }

  protected void showConstructors(String space, PrintStream out) {
    if (cr!=null)
      cr.showConstructors(space, out);
    else
      out.println(space+"This is an Array.");
  }

  protected void showDeclaredFields(String space, PrintStream out) {
    if (cr!=null)
      cr.showDeclaredFields(space, out);
    else
      out.println(space+"This is an Array.");
  }

  protected void showDeclaredMethods(String space, PrintStream out) {
    if (cr!=null)
      cr.showDeclaredMethods(space, out);
    else
      out.println(space+"This is an Array.");
  }

}

