/*
 * File         : ClassTypeDescriptor.java
 * Author       : Guansong Zhang
 * Created      : Tue May 19 13:07:52 1998
 * Revision     : $Revision: 1.8 $
 * Updated      : $Date: 1999/12/01 19:06:52 $
 */

package hpjava.compiler.reflection;

import java.util.*;
import java.lang.reflect.*;

import hpjava.compiler.util.*;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

public class ClassTypeDescriptor extends ScalarReflectorDescriptor {

  public ClassTypeDescriptor(String pname, Vector inames, NodeToken n) {
    ClassTypeResolver resolver = new ClassTypeResolver(pname, inames, this);
    n.accept(resolver);
    
    this.signature = "L"+this.name+";";
  }

  public ClassTypeDescriptor(String pname, Vector inames, Name n) {
    ClassTypeResolver resolver = new ClassTypeResolver(pname, inames, this);
    n.accept(resolver);

    this.signature = "L"+this.name+";";
  }

  public ClassTypeDescriptor(Class cls) {
    if (cls.isArray())
      throw new Error();

    this.name = cls.getName();

    this.signature = "L"+this.name+";";
  }
  

  Node scope = null;
  
  public boolean isNotLocal() {
    if (scope == null)
      return true;
    else
      return false;
  }
  
  public String name() {
    return name;
  }

  
}

/*
 * File         : ClassTypeResolver.java
 * Author       : Guansong Zhang
 * Created      : Tue May 19 13:07:52 1998
 * Revision     : $Revision: 1.8 $
 * Updated      : $Date: 1999/12/01 19:06:52 $
 */

// used only for class name in source code 

class ClassTypeResolver extends DepthFirstVisitor {
  ClassTypeDescriptor cnd;

  private String pname;
  private Vector inames;

  private StringBuffer nameBuffer = new StringBuffer();  

  private IntStack status = new IntStack();
  
  final int DEFAULT  = 0;
  final int NEEDNAME = 1;

  public ClassTypeResolver(String pname, Vector inames, ClassTypeDescriptor cnd) {
    this.cnd = cnd;
    this.pname = pname;
    this.inames = inames;
    
    status.push(DEFAULT);
  }

   //
   // f0 -> < IDENTIFIER >
   // f1 -> ( "." < IDENTIFIER >  )*
   //
  
  public void visit (Name n) {
    String name;
    
    status.push(NEEDNAME);
    
    n.f0.accept(this);
    if (n.f1.present()) {
      // a full name is found here.
      visit (n.f1);
          
      name = nameBuffer.toString();
          
      if (!ExportedClassNameManager.knows(name)) {
				//System.out.println("Class name I: " + name);
        String message = "class "+name+" is not defined or can not be found.";
        throw new CompilingError(message, n.f0);            
      }     
    }
    else {
      // a short name is found here
      name = nameBuffer.toString();

      String fullName;
    
      fullName = LocalClassNameManager.resolve(n.f0);

      if ( fullName == null )      
        fullName = ExportedClassNameManager.resolve (pname, inames, n.f0.tokenImage);
          
      if ( fullName == null) {
				//System.out.println("Class name II: " + fullName);
        String message = "class "+name+" is not defined or can not be found.";
        throw new CompilingError(message, n.f0);            
      }
      else
        name = fullName;                                   
    }
  
    cnd.name = name;

    status.pop();
  }
  
  // NodeToken to create a complete name;
  
  public void visit (NodeToken n) {
    switch (status.peek()) {
    case NEEDNAME:    
      nameBuffer.append(n.tokenImage);
      break;

    case DEFAULT:
    default:
      String name = n.tokenImage;
      
      String fullName;

      fullName = LocalClassNameManager.resolve(n);

      if ( fullName == null )      
        fullName = ExportedClassNameManager.resolve (pname, inames, n.tokenImage);
          
      if ( fullName == null) {
				//System.out.println("Class name III: " + fullName);
        String message = "class "+name+" is not defined or can not be found.";
        throw new CompilingError(message, n);            
      }
      else
        name = fullName;                                   
  
      cnd.name = name;
      
      break;  
    }

  }

}
