/*
 * File         : ConstructorReflectorSourceCode.java
 * Author       : Guansong Zhang
 * Created      : Fri May 22 16:29:59 1998
 * Revision     : $Revision: 1.5 $
 * Updated      : $Date: 1999/03/16 20:27:33 $
 */

package hpjava.compiler.reflection;

import java.util.*;
import java.lang.reflect.*;

import hpjava.compiler.util.*; 

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

public class ConstructorReflectorSourceCode extends ConstructorReflector {

  protected ClassReflector declaringClassReflector;

  public ClassReflector getDeclaringClass() {
    return (this.declaringClassReflector);
  }

  protected String name;

  public String getName() {
    return (name);
  }

  protected ModifierReflector modifier;

  public ModifierReflector getModifier() {
    return modifier;
  }

  // protected field can be accessed by Class in the same Package!
  
  ConstructorReflectorSourceCode (String pname, Vector inames, Node n) {
    ConstructorReflectorComposer composer = new ConstructorReflectorComposer(pname, inames, this);
    n.accept(composer);    
  }

  /*
  public void setName(String name) {
    this.name=name;
  }

  public void setDeclaringClassReflector (ClassReflector cr) {
    this.declaringClassReflector=cr;
  }

  public void setModifierReflector (ModifierReflector mr) {
    this.modifier=mr;
  }
  */
}

/*
 * File         : ConstructorReflectorComposer.java
 * Author       : Guansong Zhang
 * Created      : Thu May 14 12:03:02 1998
 * Revision     : $Revision: 1.5 $
 * Updated      : $Date: 1999/03/16 20:27:33 $
 */

class ConstructorReflectorComposer extends DepthFirstVisitor {

  boolean debug=false;

  final int DEFAULT    = 0;
  final int SUPERCLASS = 1;
  final int INTERFACE  = 2; 

  private IntStack status=new IntStack();

  private String pname;
  private Vector inames;
  
  private ConstructorReflectorSourceCode cr;   

  public ConstructorReflectorComposer (String pname, Vector inames, ConstructorReflectorSourceCode cr) {
    this.pname=pname;
    this.inames=inames;
    this.cr=cr;
   
    status.push(DEFAULT);  
  }
  
   //
   // f0 -> [ "public" | "protected" | "private" ]
   // f1 -> <IDENTIFIER>
   // f2 -> FormalParameters()
   // f3 -> [ "throws" NameList() ]
   // f4 -> "{"
   // f5 -> [ ExplicitConstructorInvocation() ]
   // f6 -> ( BlockStatement() )*
   // f7 -> "}"
   //
   public void visit(ConstructorDeclaration n) {
      if (n.f0.present()) {
        ModifierReflector mr = new ModifierReflector(n.f0);
        cr.modifier=mr;      
      }
   
      cr.name=n.f1.tokenImage;
      
      //System.out.println(cr.name);    
   }
}


