/*
 * File         : ExportedClassNameManager.java
 * Author       : Guansong Zhang
 * Created      : Thu May 14 12:03:02 1998
 * Revision     : $Revision: 1.8 $
 * Updated      : $Date: 1999/09/03 18:17:08 $
 */

package hpjava.compiler.reflection;

import java.io.*;
import java.util.*;

import hpjava.compiler.util.*;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

public class ExportedClassNameManager {

    static Vector nameList = new Vector ();
  
    static public boolean knows (String name) {
      if (nameList.contains(name))
        return true;
      else {
        try {
          Class cls=Class.forName(name);
          return true;
        }
        catch (ClassNotFoundException e) {
          return false;  
        }
      }  
    }
 

    static public String resolve (String pname, Vector inames, String idname) {
      String[] defaultPackage = 
        new String [] { "java.lang.*", "hpjava.lang.*", "hpjava.spmd.*"};
    
      for (int i=0; i<defaultPackage.length; i++)
        if (!inames.contains(defaultPackage[i]))
          inames.addElement(defaultPackage[i]);
    
      Enumeration enum = inames.elements();
    
      LOOP: while (enum.hasMoreElements()) {
        String lookingName = null;
        String iname=(String) enum.nextElement();

        //System.out.println("iname in ExportedClassNameManager: " + iname);
        //System.out.println("idname in ExportedClassNameManager: " + idname);
           
        if (iname.endsWith("*")) {
          String leadingName = iname.substring(0,iname.length()-1);
          lookingName = leadingName + idname;
        } 
        else if (iname.endsWith("." + idname)) {       
          lookingName = iname;
        }

        //System.out.println("^^^^^^^ lookingName is " + lookingName);
      
        if (lookingName != null) {
          if (ExportedClassNameManager.knows(lookingName)){
            return lookingName;        
          }
        }
        //else
          //System.out.println("lookingName is NULL");

      }
        
      String lookingName;
      if (pname.length()==0)
				lookingName = idname;
      else
				lookingName = pname + "."+idname;

      if (ExportedClassNameManager.knows(lookingName))
        return lookingName;        
      else 
        return null;
    }

    public  ExportedClassNameManager (Project prj) {

      File srcPath,srcDirPath; 
      String srcFilePathName, srcFileName;
      Enumeration enum = prj.rootFileTable.keys();
   
      LOOP: while (enum.hasMoreElements()) {

        Node root = (Node) enum.nextElement();      

				
        srcFilePathName = (String) prj.rootFileTable.get(root);
        srcPath = new File ( (new File (srcFilePathName)).getAbsolutePath() );      
        srcFileName = srcPath.getName();

        StringBuffer pnameBuffer = new StringBuffer();
      
        PackageNameCatcher packageName = new PackageNameCatcher(pnameBuffer);
        root.accept(packageName);
        
        if(prj.isVerbose()) {

				if(pnameBuffer.length()==0)
					System.err.println("Package name not found for " + srcFileName + ".");
				else
					System.err.println("Package name: " + pnameBuffer.toString() + " found for " + srcFileName + ".");

				System.err.println();         
          
        }    
				

        NameSpaceDescriptor nsd = NameSpaceManager.nameSpaceOf(root); 
            
        ExportedClassNameFinder finder = new ExportedClassNameFinder (nsd.pname(), nameList);
        root.accept (finder);

      }  

      if (prj.isVerbose()) {
        System.err.println("Exported class names: " + nameList + " created.");        
        System.err.println();
      }     
        
    }

}

/*
 * File         : ExportedClassNameFinder.java
 * Author       : Guansong Zhang
 * Created      : Thu May 14 12:03:02 1998
 * Revision     : $Revision: 1.8 $
 * Updated      : $Date: 1999/09/03 18:17:08 $
 */

class ExportedClassNameFinder extends DepthFirstVisitor {
  boolean debug=false;

  final int DEFAULT = 0;
  final int NESTED = 1;
  final int LOCAL = 2;

  private IntStack status=new IntStack();

  protected StringBuffer header;

  protected Vector nameList;

  public ExportedClassNameFinder(String pname) {
    this.header=new StringBuffer(pname);
    nameList = new Vector();

    status.push(DEFAULT);

    debug=true;
  }

  public ExportedClassNameFinder(String pname, Vector nameList) {
    this.header=new StringBuffer(pname);
    this.nameList = nameList;

    status.push(DEFAULT);
  }

   //
   // f0 -> [ PackageDeclaration() ]
   // f1 -> ( ImportDeclaration() )*
   // f2 -> ( TypeDeclaration() )*
   // f3 -> < EOF >
   //
   public void visit(CompilationUnit n) {   
     if ( n.f2.present() ) {
       n.f2.accept(this);
     }
     if (debug) 
      System.out.println(nameList);
   }


   //
   // f0 -> "class"
   // f1 -> < IDENTIFIER >
   // f2 -> [ "extends" Name() ]
   // f3 -> [ "implements" NameList() ]
   // f4 -> ClassBody()
   //
   // continued from ClassDeclaration
   //

   public void visit(UnmodifiedClassDeclaration n) {
     String newName;
     switch (status.peek()) {
     case NESTED:
       // it seems there is no need to put "$" here ?
       if (header.length() == 0)
	 newName=n.f1.tokenImage;
       else
	 newName=header.toString()+"."+n.f1.tokenImage;

       if (nameList.contains(newName))
	 System.err.println("Repeated class declaration: "+newName);

       nameList.addElement(newName);
       break;

     case LOCAL:
       break;

     case DEFAULT:
     default:
       if (header.length() == 0)
	 newName=n.f1.tokenImage;
       else
	 newName=header.toString()+"."+n.f1.tokenImage;

       if (nameList.contains(newName))
	 System.err.println("Repeated class declaration: "+newName);

       nameList.addElement(newName);
       break;
     }       
     
     int orgLength=header.length();

     switch (status.peek()) {
     case NESTED:
       // it seems there is no need to put "$" here ?
       header.append("."+n.f1.tokenImage);
       break;
     case DEFAULT:
     default:
       header.append("."+n.f1.tokenImage);
       break;
     }       

     status.push(NESTED);
     n.f4.accept(this);
     status.pop();
     header.setLength(orgLength);

   }

   //
   // f0 -> "interface"
   // f1 -> < IDENTIFIER >
   // f2 -> [ "extends" NameList() ]
   // f3 -> "{"
   // f4 -> ( InterfaceMemberDeclaration() )*
   // f5 -> "}"
   //

   public void visit(UnmodifiedInterfaceDeclaration n) {
     String newName;
     switch (status.peek()) {
     case NESTED:
       // it seems there is no need to put "$" here ?
       if (header.length() == 0)
	 newName=n.f1.tokenImage;
       else
	 newName=header.toString()+"."+n.f1.tokenImage;

       if (nameList.contains(newName))
      	 System.err.println("Repeated class declaration: "+newName);

       nameList.addElement(newName);
       break;

     case LOCAL:
       break;

     case DEFAULT:
     default:
       if (header.length() == 0)
	 newName=n.f1.tokenImage;
       else
	 newName=header.toString()+"."+n.f1.tokenImage;
       
       if (nameList.contains(newName))
      	 System.err.println("Repeated class declaration: "+newName);

       nameList.addElement(newName);
       break;
     }       
       
     if (n.f4.present()) {
       int orgLength=header.length();

       switch (status.peek()) {
       case NESTED:
	 // it seems there is no need to put "$" here ?
	 if (header.length() == 0)
	   header.append("."+n.f1.tokenImage);
	 else
	   header = new StringBuffer(n.f1.tokenImage);
	 break;
       case DEFAULT:
       default:
	 if (header.length() == 0)
	   header = new StringBuffer(n.f1.tokenImage);
	 else
	   header.append("."+n.f1.tokenImage);
	 break;
       }       

       status.push(NESTED);
       n.f4.accept(this);
       status.pop();
       header.setLength(orgLength);
     }
   }

  // in the following cut of the possibility to visit local class

   //
   // f0 -> [ "static" ]
   // f1 -> Block()
   //
   public void visit(Initializer n) {

   }
   

   //
   // f0 -> [ "public" | "protected" | "private" ]
   // f1 -> <IDENTIFIER>
   // f2 -> FormalParameters()
   // f3 -> [ "throws" NameList() ]
   // f4 -> "{"
   // f5 -> [ ExplicitConstructorInvocation() ]
   // f6 -> ( BlockStatement() )*
   // f7 -> "}"
   //
   public void visit(ConstructorDeclaration n) {

   }

   //
   // f0 -> ( "public" | "protected" | "private" | "static" | "abstract" | "final" | "native" | "synchronized" )*
   // f1 -> ResultType()
   // f2 -> MethodDeclarator()
   // f3 -> [ "throws" NameList() ]
   // f4 -> ( Block() | ";" )
   //
   public void visit(MethodDeclaration n) {

   }

   //
   // f0 -> ( "public" | "protected" | "private" | "static" | "final" | "transient" | "volatile" )*
   // f1 -> Type()
   // f2 -> VariableDeclarator()
   // f3 -> ( "," VariableDeclarator() )*
   // f4 -> ";"
   //
   public void visit(FieldDeclaration n) {

   }


}
