/*
 * File         : ExpressionConversion.java
 * Author       : Han-Ku Lee
 * Created      : Sat Aug 14 16:07:58 1999
 * Revision     : $Revision: 1.4 $
 * Updated      : $Date: 1999/12/01 19:06:54 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1999
 */

package hpjava.compiler.reflection;

import java.io.*;
import java.util.*;
import java.lang.reflect.*;

import hpjava.compiler.util.*; 

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;
import hpjava.compiler.abstractexp.*;

public class ExpressionConversion_reflection {
  protected String BOOLEAN = "boolean";
  protected String BYTE = "byte";
  protected String CHAR = "char";
  protected String SHORT = "short";
  protected String INT = "int";
  protected String LONG = "long";
  protected String FLOAT = "float";
  protected String DOUBLE = "double";
  protected String VOID = "void";

  public boolean widening_primitive_conversion(ClassReflector left, ClassReflector right){
    boolean flag = false;

    if(left.getName().equals(BYTE)){
      if(right.getName().equals(SHORT)
         || right.getName().equals(INT)
         || right.getName().equals(LONG)
         || right.getName().equals(FLOAT)
         || right.getName().equals(DOUBLE))
        flag = true;
    }
    else if(left.getName().equals(SHORT) ||left.getName().equals(CHAR)){
      if(right.getName().equals(INT)
         || right.getName().equals(LONG)
         || right.getName().equals(FLOAT)
         || right.getName().equals(DOUBLE))
        flag = true;
    }
    else if(left.getName().equals(INT)){
      if(right.getName().equals(LONG)
         || right.getName().equals(FLOAT)
         || right.getName().equals(DOUBLE))
        flag = true;
    }
    else if(left.getName().equals(LONG)){
      if(right.getName().equals(FLOAT)
         || right.getName().equals(DOUBLE))
        flag = true;
    }
    else if(left.getName().equals(FLOAT)){
      if(right.getName().equals(DOUBLE))
        flag = true;
    }

    return flag;
  }
  
  public boolean isMoreSpecific(MethodReflector source, MethodReflector target){
    boolean flag = false;

    //System.out.println("source.getName(): " + source.getName());
    //System.out.println("target.getName(): " + target.getName());

    ClassReflector[] src = source.getParameterTypes();
    ClassReflector[] tar = target.getParameterTypes();

    for(int i=0; i<src.length; i++){
			flag = false;
      //System.out.println("i = " + i);
			//System.out.println("source's " + i + "th argument: " + src[i].getName());
			//System.out.println("target's " + i + "th argument: " + tar[i].getName());

      if(src[i].getName().equals(tar[i].getName())){
        flag = true;
      }
      else if(src[i].isSubclassOf(tar[i])){
        //System.out.println("******** SUPER ********");
        flag = true;
      }
      else if(tar[i].isInterfaceOf(src[i])){
        //System.out.println("******** INTERFACE ********");
        flag = true;
      }
      else if((tar[i].isInterface() || tar[i].isArray() || !tar[i].isInterface())
              && src[i].getName().equals("hpjava.compiler.abstractexp.NullNode")){
        flag = true;
				//System.out.println("******** 1");
      }
      else if(tar[i].equals("java.lang.Object") && src[i].isInterface()){
        flag = true;
				//System.out.println("******** 2");
      }
      else if(tar[i].equals("java.lang.Object") && src[i].isArray()){
        flag = true;
				//System.out.println("******** 3");
      }
      else if(tar[i].equals("java.lang.Cloneable") && src[i].isInterface()){
        flag = true;
				//System.out.println("******** 4");
      }
      else if(tar[i].isArray() && src[i].isArray()){
				//System.out.println("******** 5");
        if(tar[i].getComponentType().getName().equals(src[i].getComponentType().getName())){
          flag = true;
        }
        else if(tar[i].getComponentType().isSuperclassOf(src[i].getComponentType())){
          flag = true;
        }
        else if(tar[i].getComponentType().isInterfaceOf(src[i].getComponentType())){
          flag = true;
        }
        else if((tar[i].getComponentType().isInterface() 
                 || tar[i].getComponentType().isArray() 
                 || !tar[i].getComponentType().isInterface())
                && src[i].getComponentType().getName().equals("hpjava.compiler.abstractexp.NullNode")){
            flag = true;
        }
				else if(tar[i].getComponentType().getName().equals("java.lang.Object")
								&& src[i].getComponentType().isInterface()){
					flag = true;
				}
				else if(tar[i].getComponentType().getName().equals("java.lang.Object") 
								&& src[i].getComponentType().isArray()){
					flag = true;
				}
				else if(tar[i].getComponentType().getName().equals("java.lang.Cloneable") 
								&& src[i].getComponentType().isArray()){
					flag = true;
				}
				else{
					flag = false;
					break;
				}
			}
		}

    //if(flag)
    //  System.out.println("flag = true");
    //else
    //  System.out.println("flag = false");

    return flag;
	}	

  public boolean isMoreSpecific(Method source, Method target){
    boolean flag = false;

    //System.out.println("source.getName(): " + source.getName());
    //System.out.println("target.getName(): " + target.getName());

    Class[] src = source.getParameterTypes();
    Class[] tar = target.getParameterTypes();

    try{
      for(int i=0; i<src.length; i++){
				flag = false;
        //System.out.println("i = " + i);
        //System.out.println("source's " + i + "th argument: " + src[i].getName());
        //System.out.println("target's " + i + "th argument: " + tar[i].getName());

        if(src[i].getName().equals(tar[i].getName())){
          flag = true;
        }
        else if(ClassReflector.forName(src[i].getName()).isSubclassOf(ClassReflector.forName(tar[i].getName()))){
          //System.out.println("******** SUPER ********");
          flag = true;
        }
        else if(ClassReflector.forName(tar[i].getName()).isInterfaceOf(ClassReflector.forName(src[i].getName()))){
          //System.out.println("******** INTERFACE ********");
          flag = true;
        }
        else if((ClassReflector.forName(tar[i].getName()).isInterface() 
            || ClassReflector.forName(tar[i].getName()).isArray()
            || !ClassReflector.forName(tar[i].getName()).isInterface())
           && ClassReflector.forName(src[i].getName()).getName().equals("hpjava.compiler.abstractexp.NullNode")){
          flag = true;
        }
        else if(ClassReflector.forName(tar[i].getName()).getName().equals("java.lang.Object")
                && ClassReflector.forName(src[i].getName()).isInterface()){
          flag = true;
        }
        else if(ClassReflector.forName(tar[i].getName()).getName().equals("java.lang.Object")
                && ClassReflector.forName(src[i].getName()).isArray()){
          flag = true;
        }
        else if(ClassReflector.forName(tar[i].getName()).getName().equals("java.lang.Cloneable")
                && ClassReflector.forName(src[i].getName()).isInterface()){
          flag = true;
        }
        else if(ClassReflector.forName(tar[i].getName()).isArray()
                && ClassReflector.forName(src[i].getName()).isArray()){
          if(ClassReflector.forName(tar[i].getName()).getComponentType().getName().equals(ClassReflector.forName(src[i].getName()).getName())){
            flag = true;
          }
          else if(ClassReflector.forName(tar[i].getName()).getComponentType().isSuperclassOf(ClassReflector.forName(src[i].getName()).getComponentType())){
            flag = true;
          }
          else if(ClassReflector.forName(tar[i].getName()).getComponentType().isInterfaceOf(ClassReflector.forName(src[i].getName()).getComponentType())){
            flag = true;
          }
          else if((ClassReflector.forName(tar[i].getName()).getComponentType().isInterface()
                   || ClassReflector.forName(tar[i].getName()).getComponentType().isArray()
                   || !ClassReflector.forName(tar[i].getName()).getComponentType().isInterface())
                  && ClassReflector.forName(src[i].getName()).getComponentType().getName().equals("hpjava.compiler.abstractexp.NullNode")){
            flag = true;
          }
          else if(ClassReflector.forName(tar[i].getName()).getComponentType().getName().equals("java.lang.Object")
                  && ClassReflector.forName(src[i].getName()).getComponentType().isInterface()){
            flag = true;
          }
          else if(ClassReflector.forName(tar[i].getName()).getComponentType().getName().equals("java.lang.Object")
                  && ClassReflector.forName(src[i].getName()).getComponentType().isArray()){
            flag = true;
          }
          else if(ClassReflector.forName(tar[i].getName()).getComponentType().getName().equals("java.lang.Cloneable")
                  && ClassReflector.forName(src[i].getName()).getComponentType().isInterface()){
            flag = true;
          }
        }
        else{
          flag = false;
          break;
        }
      }
    }
    catch(ClassNotFoundException cnf){
      System.out.println("ClassNotFoundException ...");
    }

    //if(flag)
    //  System.out.println("flag = true");
    //else
    //  System.out.println("flag = false");

    return flag;
  }
  
  public boolean isMaximallySpecific(Vector sources, int index){
    boolean flag = true;

    for(int i=0; i<sources.size(); i++){
      if(i != index){
        if(sources.elementAt(i).getClass().getName().equals("java.lang.reflect.Method"))
          if(isMoreSpecific((Method)sources.elementAt(i), 
                            (Method)sources.elementAt(index))){
            flag = false;
            //flag = true;
            //System.err.println("******* Source is NOT more specific than target");
            //System.out.println("******* Source is more specific than target");
          }
          else{
            //flag = true;
            //flag = false;
            //System.err.println("******* Source is more specific than target");
						//System.out.println("******* Source is NOT more specific than target");
          }
        else
          if(isMoreSpecific((MethodReflector)sources.elementAt(i), 
                            (MethodReflector)sources.elementAt(index))){
            flag = false;
            //flag = true;
            //System.err.println("******* Source is NOT more specific than target");
            //System.out.println("******* Source is more specific than target");
          }
          else{
            //flag = true;
            //flag = false;
            //System.err.println("******* Source is more specific than target");
						//System.out.println("******* Source is NOT more specific than target");
          }
      }
    }

    return flag;
  }


}
