/*
 * File         : FieldReflectorSourceCode.java
 * Author       : Guansong Zhang
 * Created      : Tue May 19 11:44:10 1998
 * Revision     : $Revision: 1.9 $
 * Updated      : $Date: 1999/05/20 18:31:28 $
 */

package hpjava.compiler.reflection;


import java.io.*;
import java.util.*;

import java.lang.reflect.*;

import hpjava.compiler.util.*; 

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

public class FieldReflectorSourceCode extends FieldReflector {

  protected ClassReflector declaringClassReflector; 
 
  public ClassReflector getDeclaringClass() {
    return (this.declaringClassReflector);
  }

  protected String name;

  public String getName() {
    return (this.name);
  }

  protected String typeName;

  public String getTypeName() {
    return (this.typeName);
  }

  protected ModifierReflector modifier;

  public ModifierReflector getModifier() {
    return this.modifier;
  }

  // protected field can be accessed by Class in the same Package!

  public FieldReflectorSourceCode (ScalarReflectorDescriptor srd,
				   ArraySignaturesImage prefix, 
				   Node n) {
    FieldReflectorComposer composer = new FieldReflectorComposer(srd, prefix, this);
    n.accept(composer); 
  }  

  /*
  public void setName(String name) {
    this.name=name;
  }

  public void setTypeName(String name) {
    this.typeName=name;
  }

  public void setDeclaringClassReflector (ClassReflector cr) {
    this.declaringClassReflector=cr;
  }

  public void setModifierReflector (ModifierReflector mr) {
    this.modifier=mr;
  }
  */

}  

class FieldReflectorComposer extends DepthFirstVisitor {

  boolean debug=false;

  final int DEFAULT    = 0;
  final int SUPERCLASS = 1;
  final int INTERFACE  = 2; 

  private IntStack status=new IntStack();

  private ScalarReflectorDescriptor srd;
  private ArraySignaturesImage prefix;
  
  private FieldReflectorSourceCode fr;   

  public FieldReflectorComposer (ScalarReflectorDescriptor srd,
				 ArraySignaturesImage prefix, 
				 FieldReflectorSourceCode fr) {
    this.srd=srd;
    this.prefix=prefix;
    this.fr=fr;
   
    status.push(DEFAULT);  
  }

   //
   // f0 -> VariableDeclaratorId()
   // f1 -> [ "=" VariableInitializer() ]
   //
   public void visit(VariableDeclarator n) {
      n.f0.accept(this);
   }
  
   //
   // f0 -> <IDENTIFIER>
   // f1 -> ( ArraySignature() )*
   //
   public void visit(VariableDeclaratorId n) {
     this.fr.name = n.f0.tokenImage;
     
     //     System.out.println("");
     //     System.out.println("<IDENTIFIER> " + n.f0.tokenImage);
     //     System.out.println("********* this.prefix: " + this.prefix);      
     
     ArraySignaturesImage asd = this.prefix;

     if (n.f1.present()) {
        ArraySignaturesImage postfix =
          new ArraySignaturesImage((NodeListOptional)(n.f1));
        
        if (asd !=null){
          asd = new ArraySignaturesImage(postfix, this.prefix);
        }
        else{
          asd = postfix;
        }  
     }

     if (asd == null) {       
       //fr.typeName = srd.signature;
       fr.typeName = srd.name();

       /*
       if(srd.signature.startsWith("L"))
         fr.typeName = srd.name();
       else
         fr.typeName = srd.signature;
       */
       //System.out.println("********* srd.signature: " + srd.signature);
       //System.out.println("********* srd.name(): " + srd.name());
     }
     else{ 
       ArrayReflectorDescriptor atd = new ArrayReflectorDescriptor(this.srd, asd);     
       fr.typeName = atd.signature;
        
       //System.out.println("********* asd.name: " + asd.name);
       //System.out.println("********* srd.signature: " + srd.signature);
       //System.out.println("********* atd.signature: " + atd.signature);
     }

   }

}
