/*
 * File         : ImportNamesCatcher.java
 * Author       : Guansong Zhang
 * Created      : Thu May 14 12:03:02 1998
 * Revision     : $Revision: 1.6 $
 * Updated      : $Date: 1999/09/03 20:05:54 $
 */

package hpjava.compiler.reflection;

import java.io.*;
import java.util.*;

import hpjava.compiler.util.*;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

public class ImportNamesCatcher extends DepthFirstVisitor {
  final int DEFAULT=0;
  final int IMPORTDECLARATION=1;

  boolean debug=false;

  private IntStack status = new IntStack();
  private String name="";

  protected Vector names; 

  public ImportNamesCatcher() {
    names=new Vector();
    debug=true;
    status.push(DEFAULT);
  }

  public ImportNamesCatcher (Vector names) {
    this.names=names;
    status.push(DEFAULT);
  }
  
/*
  public ImportNamesCatcher (ImportNamesDescriptor ind) {
    names = new Vector();
    this.ind = ind;
    ind.names = names;  
  }  
*/

  // No need to use public any more;

   //
   // f0 -> [ PackageDeclaration() ]
   // f1 -> ( ImportDeclaration() )*
   // f2 -> ( TypeDeclaration() )*
   // f3 -> < EOF >
   //
   public void visit(CompilationUnit n) {   
     if ( n.f1.present() ) {
       status.push(IMPORTDECLARATION);
       //names.addElement("java.lang.*");
       n.f1.accept(this);
       status.pop();
     }
     if (debug) 
      System.out.println(names);
   }
   
   public void visit(NodeListOptional n) {

     switch(status.peek()) {
     case IMPORTDECLARATION:
       for ( Enumeration e = n.elements(); e.hasMoreElements(); ) {
         name="";

         // without knowing the concrete Node type, you can not use visit.?
        
         ((Node)e.nextElement()).accept(this);

         names.addElement(name);
       }
       break;

     case DEFAULT:
     default:
       for ( Enumeration e = n.elements(); e.hasMoreElements(); ) 
         ((Node)e.nextElement()).accept(this);
       break;
     }
   }

   //
   // f0 -> "import"
   // f1 -> Name()
   // f2 -> [ "." "*" ]
   // f3 -> ";"
   //
   // uses the toString() method of the NodeTokens to print manually
   // (rather than visiting the tokens).
   //
   public void visit(ImportDeclaration n) {   
      n.f1.accept(this);
      n.f2.accept(this);
   }

   //
   // f0 -> < IDENTIFIER >
   // f1 -> ( "." < IDENTIFIER >  )*
   //
   public void visit(Name n) {
     n.f0.accept(this);
     if (n.f1.present()) {
       status.push(DEFAULT);
       n.f1.accept(this);
       status.pop();
     }  
   }
   

  // NodeToken
  public void visit (NodeToken n) {
    name+=n.tokenImage;
  }
 
  public String toString() {
     return names.toString();
  }

}
