/*
 * File         : LocalClassNameManager.java
 * Author       : Guansong Zhang
 * Created      : Thu May 14 12:03:02 1998
 * Revision     : $Revision: 1.5 $
 * Updated      : $Date: 1999/03/16 20:27:37 $
 */

package hpjava.compiler.reflection;

import java.io.*;
import java.util.*;

import hpjava.compiler.util.*;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;


public class LocalClassNameManager {

    static Vector nameList = new Vector ();
  
    static public String resolve (NodeToken id) {
      ScopeDescriptor scope = ScopeManager.scopeOf(id);
             
      while (scope != ScopeManager.globalScope) {
      
        String name = id.tokenImage+"."+ scope.toString();

        if (nameList.contains(name) )
          return name;
          
        scope=ScopeManager.scopeOf(scope.node());    
      }
        
      return null;  
    }

    public  LocalClassNameManager (Project prj) {

      File srcPath,srcDirPath; 
      String srcFilePathName, srcFileName;
      Enumeration enum = prj.rootFileTable.keys();
   
      LOOP: while (enum.hasMoreElements()) {

        Node root = (Node) enum.nextElement();      

        srcFilePathName = (String) prj.rootFileTable.get(root);
        srcPath = new File ( (new File (srcFilePathName)).getAbsolutePath() );      
        srcFileName = srcPath.getName();
    
        LocalClassNameFinder finder = new LocalClassNameFinder (nameList);
        root.accept (finder);

      }  

      if (prj.isVerbose()) {
        System.err.println("Local class names: "+nameList+" created.");        
        System.err.println();
      }     
        
    }
  
}

/*
 * File         : LocalClassNameFinder.java
 * Author       : Guansong Zhang
 * Created      : Thu May 14 12:03:02 1998
 * Revision     : $Revision: 1.5 $
 * Updated      : $Date: 1999/03/16 20:27:37 $
 */

class LocalClassNameFinder extends DepthFirstVisitor {
  boolean debug=false;

  final int DEFAULT = 0;
  final int NESTED = 1;
  final int LOCAL = 2;

  private IntStack status=new IntStack();

  protected Vector nameList;

  public LocalClassNameFinder() {
    nameList = new Vector();

    status.push(DEFAULT);

    debug=true;
  }

  public LocalClassNameFinder(Vector nameList) {
    this.nameList = nameList;

    status.push(DEFAULT);
  }

   //
   // f0 -> [ PackageDeclaration() ]
   // f1 -> ( ImportDeclaration() )*
   // f2 -> ( TypeDeclaration() )*
   // f3 -> < EOF >
   //
   public void visit(CompilationUnit n) {   
     if ( n.f2.present() ) {
       n.f2.accept(this);
     }
     if (debug) 
      System.out.println(nameList);
   }


   //
   // f0 -> "class"
   // f1 -> < IDENTIFIER >
   // f2 -> [ "extends" Name() ]
   // f3 -> [ "implements" NameList() ]
   // f4 -> ClassBody()
   //
   // continued from ClassDeclaration
   //

   public void visit(UnmodifiedClassDeclaration n) {
     String newName;
     StringBuffer newNameBuffer = new StringBuffer();

     switch (status.peek()) {
     case LOCAL:
       ScopeDescriptor scope = ScopeManager.scopeOf(n.f1);

       newNameBuffer=newNameBuffer.append(n.f1.tokenImage);
       newNameBuffer=newNameBuffer.append(".");
       newNameBuffer=newNameBuffer.append(scope.toString());

       newName=newNameBuffer.toString();

       if (nameList.contains(newName))
	 System.err.println("Repeated local class declaration: "+n.f1.tokenImage);
       nameList.addElement(newName);
       break;

     case NESTED:
     case DEFAULT:
     default:
       break;
     }       

     switch (status.peek()) {
     case LOCAL:
       status.push(LOCAL);
       break;
     case NESTED:
       status.push(NESTED);
       break;
     case DEFAULT:
     default:
       status.push(DEFAULT);
       break;
     }
     n.f4.accept(this);
     status.pop();

   }

   //
   // f0 -> "interface"
   // f1 -> < IDENTIFIER >
   // f2 -> [ "extends" NameList() ]
   // f3 -> "{"
   // f4 -> ( InterfaceMemberDeclaration() )*
   // f5 -> "}"
   //

   public void visit(UnmodifiedInterfaceDeclaration n) {

   }

   //
   // f0 -> [ "static" ]
   // f1 -> Block()
   //
   public void visit(Initializer n) {
       status.push(LOCAL);
       n.f1.accept(this);
       status.pop();
   }
   

   //
   // f0 -> [ "public" | "protected" | "private" ]
   // f1 -> <IDENTIFIER>
   // f2 -> FormalParameters()
   // f3 -> [ "throws" NameList() ]
   // f4 -> "{"
   // f5 -> [ ExplicitConstructorInvocation() ]
   // f6 -> ( BlockStatement() )*
   // f7 -> "}"
   //
   public void visit(ConstructorDeclaration n) {
       status.push(LOCAL);
       n.f6.accept(this);
       status.pop();
   }

   //
   // f0 -> ( "public" | "protected" | "private" | "static" | "abstract" | "final" | "native" | "synchronized" )*
   // f1 -> ResultType()
   // f2 -> MethodDeclarator()
   // f3 -> [ "throws" NameList() ]
   // f4 -> ( Block() | ";" )
   //
   public void visit(MethodDeclaration n) {
       status.push(LOCAL);
       n.f4.accept(this);
       status.pop();
   }

  // cut field off.

   //
   // f0 -> ( "public" | "protected" | "private" | "static" | "final" | "transient" | "volatile" )*
   // f1 -> Type()
   // f2 -> VariableDeclarator()
   // f3 -> ( "," VariableDeclarator() )*
   // f4 -> ";"
   //
   public void visit(FieldDeclaration n) {

   }

}
