/*
 * File         : MethodReflectorSourceCode.java
 * Author       : Guansong Zhang
 * Created      : Fri May 22 12:10:05 1998
 * Revision     : $Revision: 1.12 $
 * Updated      : $Date: 1999/12/01 19:06:56 $
 * Reviser      : Han-Ku Lee
 */

package hpjava.compiler.reflection;

import java.util.*;
import java.lang.reflect.*;

import hpjava.compiler.util.*; 

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

public class MethodReflectorSourceCode extends MethodReflector {
  
  protected ClassReflector declaringClassReflector;

  public ClassReflector getDeclaringClass() {
    return (this.declaringClassReflector);
  }

  protected String name;

  public String getName() {
    return (name);
  }

  protected ClassReflector[] parametertypes = null;

  public ClassReflector[] getParameterTypes(){
    return parametertypes;
  }

  protected String returnTypeName;

  public String getReturnTypeName() {
    return returnTypeName;
  }

  protected ModifierReflector modifier;

  public ModifierReflector getModifier() {
    return modifier;
  }

  // protected field can be accessed by Class in the same Package!

  public MethodReflectorSourceCode(String pname, Vector inames, Node n) {
    MethodReflectorComposer composer = new MethodReflectorComposer(pname, inames, this);
    n.accept(composer); 
  }  


  /*
  public void setName(String name) {
    this.name=name;
  }

  public void setReturnTypeName(String name) {
    this.returnTypeName=name;
  }

  public void setDeclaringClassReflector (ClassReflector cr) {
    this.declaringClassReflector=cr;
  }

  public void setModifierReflector (ModifierReflector mr) {
    this.modifier=mr;
  }
  */
}


/*
 * File         : MethodReflectorComposer.java
 * Author       : Guansong Zhang
 * Created      : Thu May 14 12:03:02 1998
 * Revision     : $Revision: 1.12 $
 * Updated      : $Date: 1999/12/01 19:06:56 $
 */

class MethodReflectorComposer extends DepthFirstVisitor {

  boolean debug=false;

  final int DEFAULT    = 0;
  final int SUPERCLASS = 1;
  final int INTERFACE  = 2; 

  private IntStack status=new IntStack();

  private String pname;
  private Vector inames;
  
  private MethodReflectorSourceCode mr;
  private MemberSignatureImage msi;

  public MethodReflectorComposer (String pname, Vector inames, MethodReflectorSourceCode mr) {
    this.pname=pname;
    this.inames=inames;
    this.mr=mr;
   
    status.push(DEFAULT);  
  }
  
   //
   // f0 -> ( "public" | "protected" | "private" | "static" | "abstract" | "final" | "native" | "synchronized" )*
   // f1 -> ResultType()
   // f2 -> MethodDeclarator()
   // f3 -> [ "throws" NameList() ]
   // f4 -> ( Block() | ";" )
   //
   public void visit(MethodDeclaration n) {
      if (n.f0.present()) {
        ModifierReflector mod = new ModifierReflector(n.f0);
        mr.modifier=mod;      
      }

      msi = new MemberSignatureImage(pname, inames, n);

      //mr.name=n.f2.f0.tokenImage;
      mr.name = msi.name;
      ClassReflector[] tmp = new ClassReflector[msi.number_of_arguments];
      
      //System.out.println("msi.number_of_arguments: " + msi.number_of_arguments);
      try{
        for(int i=0; i<msi.number_of_arguments; i++){
            //System.out.println("***** Before (String)msi.argumentsList.elementAt(i): " + (String)msi.argumentsList.elementAt(i));

          if(((String)msi.argumentsList.elementAt(i)).startsWith("L")
             && ((String)msi.argumentsList.elementAt(i)).endsWith(";")){
              //System.out.println("msi.argumentsList.lastIndexOf: " + ((String)msi.argumentsList.elementAt(i)).lastIndexOf(";"));
            msi.argumentsList.setElementAt(((String)msi.argumentsList.elementAt(i)).substring(1, ((String)msi.argumentsList.elementAt(i)).lastIndexOf(";")), i);
          }

          //System.out.println("***** After (String)msi.argumentsList.elementAt(i): " + (String)msi.argumentsList.elementAt(i));

          tmp[i] = ClassReflector.forName((String)msi.argumentsList.elementAt(i));
        }
      }
      catch(ClassNotFoundException cnf){
        System.out.println("ClassNotFoundException occurred ...");
      }

      mr.parametertypes = tmp;

      if (n.f1.f0.which==0)
      	mr.returnTypeName="void";
      else {
        Type typeNode = (Type)(n.f1.f0.choice);

      	ArraySignaturesImage asd0 =
    	  new ArraySignaturesImage((NodeListOptional)n.f2.f2);

      	ArraySignaturesImage asd1 =
      	  new ArraySignaturesImage((NodeListOptional)typeNode.f1);
	
      	ArraySignaturesImage asd =
      	  new ArraySignaturesImage(asd0, asd1);

        ScalarReflectorDescriptor srd;

        switch (typeNode.f0.which) { 
        case 0:
          srd = new PrimitiveTypeDescriptor((PrimitiveType)typeNode.f0.choice);
          break;
        case 1:
          srd = new ClassTypeDescriptor(pname, inames, (Name)typeNode.f0.choice); 
          break;  
        default:
          throw new Error();
        }

        if(asd.isEmpty()) {
          //mr.returnTypeName = srd.signature;
          mr.returnTypeName = srd.name();

          //System.out.println("@@@@@@@@@@@@@@@@@@ srd.signature: " + srd.signature);
          //System.out.println("@@@@@@@@@@@@@@@@@@ srd.name(): " + srd.name());
        }
        else {
      	  ArrayReflectorDescriptor ard = new ArrayReflectorDescriptor(srd, asd);

      	  mr.returnTypeName=ard.signature;
          //System.out.println("@@@@@@@@@@@@@@@@@@ asd.name: " + asd.name);
          //System.out.println("@@@@@@@@@@@@@@@@@@ srd.signature: " + srd.signature);
          //System.out.println("@@@@@@@@@@@@@@@@@@ ard.srd.signature" + ard.signature);
      	}
      }
   }

   //
   // f0 -> "void"
   //       | Type()
   //
   public void visit(ResultType n) {

   }

   //
   // f0 -> ( PrimitiveType() | Name() )
   // f1 -> ( ArraySignature() )*
   //
   public void visit(Type n) {

   }

   //
   // f0 -> <IDENTIFIER>
   // f1 -> FormalParameters()
   // f2 -> ( ArraySignature() )*
   //
   public void visit(MethodDeclarator n) {

   }
  
}

