/*
 * File         : ModifierReflector.java
 * Author       : Guansong Zhang
 * Created      : Tue May 19 13:07:52 1998
 * Revision     : $Revision: 1.7 $
 * Updated      : $Date: 1999/03/16 20:27:43 $
 */

package hpjava.compiler.reflection;

import java.util.*;
import java.io.*;
import java.lang.reflect.*;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

public class ModifierReflector {

  protected boolean _abstract=false;
  protected boolean _final=false;
  protected boolean _interface=false;
  protected boolean _native=false;
  protected boolean _private=false;
  protected boolean _protected=false;
  protected boolean _public=false;
  protected boolean _static=false;
  protected boolean _synchronized=false;
  protected boolean _transient=false;
  protected boolean _volatile=false;

  public ModifierReflector (Node n) {
    ModifierReflectorComposer composer = new ModifierReflectorComposer(this);
    n.accept(composer);
  }

  // a public method for creating reflector from Java reflection.

  public ModifierReflector (int mod) {
    if (Modifier.isAbstract(mod)) 
      _abstract=true;
    if (Modifier.isFinal(mod))
      _final=true;
    if (Modifier.isInterface(mod))
      _interface=true;
    if (Modifier.isNative(mod))
      _native=true;
    if (Modifier.isPrivate(mod))
      _private=true;
    if (Modifier.isProtected(mod))
      _protected=true;
    if (Modifier.isPublic(mod))
      _public=true;
    if (Modifier.isStatic(mod))
      _static=true;
    if (Modifier.isSynchronized(mod))
      _synchronized=true;
    if (Modifier.isTransient(mod))
      _transient=true;
    if (Modifier.isVolatile(mod))
      _volatile=true;

  }

  public boolean isAbstract() {
    return (_abstract);
  }

  public boolean isFinal() {
    return (_final);
  }
  
  public boolean isInterface() {
    return (_interface);
  }

  public boolean isNative() {
    return (_native);
  }
  
  public boolean isPrivate() {
    return (_private);
  }

  public boolean isProtected() {
    return (_protected);
  }

  public boolean isPublic() {
    return (_public);
  }
  
  public boolean isStatic() {
    return (_static);
  }

  public boolean isSynchronized() {
    return (_synchronized);
  }

  public boolean isTransient() {
    return (_transient);
  }
 
  public boolean isVolatile() {
    return (_volatile);
  }

  public void show(String space, PrintStream out) {
    out.print(space+"MODIFIER:\t");
    if (isAbstract())
      out.print("abstract ");
    if (isFinal())
      out.print("final ");
    if (isInterface())
      out.print("interface ");
    if (isNative())
      out.print("native ");
    if (isPrivate())
      out.print("private ");
    if (isProtected())
      out.print("protected ");
    if (isPublic())
      out.print("public ");
    if (isStatic())
      out.print("static ");
    if (isSynchronized())
      out.print("synchronized ");
    if (isTransient())
      out.print("transient ");
    if (isVolatile()) 
      out.print("volatile ");
    out.println(); 
  }

}

/*
 * File         : ModifierReflectorComposer.java
 * Author       : Guansong Zhang
 * Created      : Thu May 14 12:03:02 1998
 * Revision     : $Revision: 1.7 $
 * Updated      : $Date: 1999/03/16 20:27:43 $
 */


class ModifierReflectorComposer extends DepthFirstVisitor {
  private ModifierReflector mr;

  public ModifierReflectorComposer(ModifierReflector mr) {
    this.mr = mr;
  }

  public void visit(NodeToken n) {
    if (n.tokenImage.equals("abstract"))
      mr._abstract=true;
    if (n.tokenImage.equals("final"))
      mr._final=true;
    if (n.tokenImage.equals("interface"))
      mr._interface=true;
    if (n.tokenImage.equals("native"))
      mr._native=true;
    if (n.tokenImage.equals("private"))    
      mr._private=true;
    if (n.tokenImage.equals("protected"))    
      mr._protected=true;
    if (n.tokenImage.equals("public"))    
      mr._public=true;
    if (n.tokenImage.equals("static"))
      mr._static=true;
    if (n.tokenImage.equals("synchronized"))
      mr._synchronized=true;
    if (n.tokenImage.equals("transient"))
      mr._transient=true;
    if (n.tokenImage.equals("volatile"))
      mr._volatile=true;      
  }

}
