/*
 * File         : NameSpaceVisitoor.java
 * Author       : Han-Ku Lee
 * Created      : Thu March 4 1999
 *
 * Update       : Wed March 17 1999
 */

package hpjava.compiler.reflection;

import java.io.*;
import java.util.*;
import java.lang.*;

import hpjava.compiler.util.*;

import hpjava.compiler.parser.*;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

public class NameSpaceFinder extends DepthFirstVisitor {

  protected NameSpaceDescriptor nsd;
  protected NameSpaceDescriptor newnsd;
  protected Stack nsStack = new Stack();
  protected Hashtable rootSpaceTable = new Hashtable();

  public NameSpaceFinder(Hashtable table, Node root) {
      nsd = new NameSpaceDescriptor(root);
      rootSpaceTable = table;

      // this is added for temparory work in reflection.
      rootSpaceTable.put(root, nsd.clone());
  }

   public void visit(NodeToken n){
      if(n.kind == HPJavaParserConstants.IDENTIFIER){
					/* 
        System.out.println(n.tokenImage);
        System.out.println(nsd.toString());
        System.out.println();
					*/
      }
   }

   /**
    * f0 -> [ PackageDeclaration() ]
    * f1 -> ( ImportDeclaration() )*
    * f2 -> ( TypeDeclaration() )*
    * f3 -> <EOF>
    */
   public void visit(CompilationUnit n) {
      n.f2.accept(this);
      n.f3.accept(this);
   }

   /**
    * f0 -> "class"
    * f1 -> <IDENTIFIER>
    * f2 -> [ "extends" Name() ]
    * f3 -> [ "implements" NameList() ]
    * f4 -> ClassBody()
    */
   public void visit(UnmodifiedClassDeclaration n) {
      n.f0.accept(this);
      n.f1.accept(this);
      newnsd = (NameSpaceDescriptor)nsd.clone();
      rootSpaceTable.put(n.f1, newnsd);
      if(nsStack.empty()){
        nsStack.push(n.f1.tokenImage);
      }
      else{
        nsStack.push(nsStack.peek() + "." + n.f1.tokenImage);
      }
      n.f2.accept(this);
      n.f3.accept(this);
      n.f4.accept(this);

      // In case of inner class, is this line a bug?
      // Or, what about the other cases???
      nsd.cname = new String();  

      nsStack.pop();
   }

   /**
    * f0 -> "{"
    * f1 -> ( ClassBodyDeclaration() )*
    * f2 -> "}"
    */
   public void visit(ClassBody n) {
      if(n.f1.present()){
        for(int i=0; i<n.f1.size(); i++){
          //if(nsd.inames.isEmpty() && nsd.pnameBuffer.length() == 0){
          if(nsd.pnameBuffer.length() == 0){
            nsd.inames.addElement(nsd.pnameBuffer.toString() + nsStack.peek());  
						//System.out.println("nsd.inames is EMPTY... " + i);
						//System.out.println("nsd.pnameBuffer.toString(): " + nsd.pnameBuffer.toString());
						//System.out.println("nsStack.peek(): " + nsStack.peek());
						//System.out.println("nsd.inames: " + nsd.inames);
						//System.out.println();
          }
          else{
						//System.out.println("nsd.inames: " + nsd.inames);
						//System.out.println("nsd.pnameBuffer.length(): " + nsd.pnameBuffer.length());
            nsd.inames.addElement(nsd.pnameBuffer.toString() + "." + nsStack.peek());
						//System.out.println("nsd.inames is  NOT EMPTY... " + i);
						//System.out.println("nsd.pnameBuffer.toString(): " + nsd.pnameBuffer.toString());
						//System.out.println("nsStack.peek(): " + nsStack.peek());
						//System.out.println("nsd.inames: " + nsd.inames);
						//System.out.println();
          }

          nsd.cname = nsd.inames.lastElement().toString();

          n.f1.elementAt(i).accept(this);
          nsd.inames.removeElement(nsd.inames.lastElement());      
        }
      }
   }
   /**
    * f0 -> ( "public" | "protected" | "private" | "static" | "abstract" | "final" | "native" | "synchronized" )*
    * f1 -> ResultType()
    * f2 -> <IDENTIFIER>
    * f3 -> "("
    */
   public void visit(MethodDeclarationLookahead n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
      newnsd = (NameSpaceDescriptor)nsd.clone();
      rootSpaceTable.put(n.f2, newnsd);
      n.f3.accept(this);
   }

   /**
    * f0 -> "interface"
    * f1 -> <IDENTIFIER>
    * f2 -> [ "extends" NameList() ]
    * f3 -> "{"
    * f4 -> ( InterfaceMemberDeclaration() )*
    * f5 -> "}"
    */
   public void visit(UnmodifiedInterfaceDeclaration n) {
      n.f0.accept(this);
      n.f1.accept(this);
      newnsd = (NameSpaceDescriptor)nsd.clone();
      rootSpaceTable.put(n.f1, newnsd);
      if(nsStack.empty()){
        nsStack.push(n.f1.tokenImage);
      }
      else{
        nsStack.push(nsStack.peek() + "." + n.f1.tokenImage);
      }
      n.f2.accept(this);

      if(n.f4.present()){
        for(int i=0; i<n.f4.size(); i++){
          //if(nsd.inames.isEmpty() && nsd.pnameBuffer.length() == 0){
          if(nsd.pnameBuffer.length() == 0){
            nsd.inames.addElement(nsd.pnameBuffer.toString() + nsStack.peek());  
          }
          else{
            nsd.inames.addElement(nsd.pnameBuffer.toString() + "." + nsStack.peek());
          }

          nsd.cname = nsd.inames.lastElement().toString();

          n.f4.elementAt(i).accept(this);
          nsd.inames.removeElement(nsd.inames.lastElement());
          
          // In case of inner class, is this line a bug?
          // Or, what about the other cases???
          nsd.cname = new String();  
        }
      }

      nsStack.pop();

   }

   /**
    * f0 -> <IDENTIFIER>
    * f1 -> ( ArraySignature() )*
    */
   public void visit(VariableDeclaratorId n) {
      n.f0.accept(this);
      newnsd = (NameSpaceDescriptor)nsd.clone();
      rootSpaceTable.put(n.f0, newnsd);
      n.f1.accept(this);
   }

   /**
    * f0 -> <IDENTIFIER>
    * f1 -> FormalParameters()
    * f2 -> ( ArraySignature() )*
    */
   public void visit(MethodDeclarator n) {
      n.f0.accept(this);
      newnsd = (NameSpaceDescriptor)nsd.clone();
      rootSpaceTable.put(n.f0, newnsd);
      n.f1.accept(this);
      n.f2.accept(this);
   }

   /**
    * f0 -> [ "public" | "protected" | "private" ]
    * f1 -> <IDENTIFIER>
    * f2 -> FormalParameters()
    * f3 -> [ "throws" NameList() ]
    * f4 -> "{"
    * f5 -> [ ExplicitConstructorInvocation() ]
    * f6 -> ( BlockStatement() )*
    * f7 -> "}"
    */
   public void visit(ConstructorDeclaration n) {
      n.f0.accept(this);
      n.f1.accept(this);
      newnsd = (NameSpaceDescriptor)nsd.clone();
      rootSpaceTable.put(n.f1, newnsd);
      n.f2.accept(this);
      n.f3.accept(this);
      n.f5.accept(this);
      n.f6.accept(this);
   }

   /**
    * f0 -> <IDENTIFIER>
    * f1 -> ( "." <IDENTIFIER> )*
    */
   public void visit(Name n) {
      n.f0.accept(this);
      newnsd = (NameSpaceDescriptor)nsd.clone();
      rootSpaceTable.put(n.f0, newnsd);

      for(int i=0; i<n.f1.size(); i++){
        NodeSequence seq = (NodeSequence)n.f1.elementAt(i);
        seq.elementAt(1).accept(this);
        newnsd = (NameSpaceDescriptor)nsd.clone();
        rootSpaceTable.put(seq.elementAt(1), newnsd);
      }

   }

   /**
    * f0 -> "(" PrimitiveType()
    *       | "(" Name() ArraySignature()
    *       | "(" Name() ")" ( "~" | "!" | "(" | <IDENTIFIER> | "this" | "super" | "new" | Literal() )
    */
   public void visit(CastLookahead n) {
      n.f0.accept(this);
      if(n.f0.which == 2){
        NodeSequence seq = (NodeSequence)n.f0.choice;
        NodeChoice nc = (NodeChoice)seq.elementAt(3);
        if(nc.which == 3){
          newnsd = (NameSpaceDescriptor)nsd.clone();
          rootSpaceTable.put(nc.choice, newnsd);
        }
      }
   }

   /**
    * f0 -> Literal()
    *       | ResultType() "." "class"
    *       | AllocationExpression()
    *       | "this"
    *       | "super"
    *       | "(" Expression() ")"
    *       | <IDENTIFIER>
    */
   public void visit(PrimaryPrefix n) {
      n.f0.accept(this);

      //if(n.f0.which == 6){
      if(n.f0.which == 3 || n.f0.which == 6){
        newnsd = (NameSpaceDescriptor)nsd.clone();
        rootSpaceTable.put(n.f0.choice, newnsd);
      }
   }

   /**
    * f0 -> "." "this"
    *       | "." AllocationExpression()
    *       | "[" "[" DistArraySecRefDimList() "]" "]"
    *       | "[" TripletOrExpressionList() "]"
    *       | "." <IDENTIFIER>
    *       | Arguments()
    */
   public void visit(PrimarySuffix n) {
      n.f0.accept(this);

      if(n.f0.which == 4){
        NodeSequence seq = (NodeSequence)n.f0.choice;
        seq.elementAt(1).accept(this);
        newnsd = (NameSpaceDescriptor)nsd.clone();
        rootSpaceTable.put(seq.elementAt(1), newnsd);
      }
   }

   /**
    * f0 -> <IDENTIFIER>
    * f1 -> ":"
    * f2 -> Statement()
    */
   public void visit(LabeledStatement n) {
      n.f0.accept(this);
      newnsd = (NameSpaceDescriptor)nsd.clone();
      rootSpaceTable.put(n.f0, newnsd);
      n.f1.accept(this);
      n.f2.accept(this);
   }

   /**
    * f0 -> "break"
    * f1 -> [ <IDENTIFIER> ]
    * f2 -> ";"
    */
   public void visit(BreakStatement n) {
      n.f0.accept(this);
      if(n.f1.present()){
        n.f1.accept(this);
        newnsd = (NameSpaceDescriptor)nsd.clone();
        rootSpaceTable.put(n.f1, newnsd);
      }
   }

   /**
    * f0 -> "continue"
    * f1 -> [ <IDENTIFIER> ]
    * f2 -> ";"
    */
   public void visit(ContinueStatement n) {
      n.f0.accept(this);
      if(n.f1.present()){
        n.f1.accept(this);
        newnsd = (NameSpaceDescriptor)nsd.clone();
        rootSpaceTable.put(n.f1, newnsd);
      }
   }

   /**
    * f0 -> <IDENTIFIER>
    * f1 -> "="
    * f2 -> PrimaryExpression()
    */
   public void visit(NamedLocation n) {
      n.f0.accept(this);
      newnsd = (NameSpaceDescriptor)nsd.clone();
      rootSpaceTable.put(n.f0, newnsd);
      n.f1.accept(this);
      n.f2.accept(this);
   }

   /**
    * f0 -> <IDENTIFIER>
    * f1 -> "="
    * f2 -> PrimaryExpression()
    * f3 -> "for"
    * f4 -> Triplet()
    */
   public void visit(NamedLocationRange n) {
      n.f0.accept(this);
      newnsd = (NameSpaceDescriptor)nsd.clone();
      rootSpaceTable.put(n.f0, newnsd);
      n.f1.accept(this);
      n.f2.accept(this);
   }

}
