/*
 * File         : PackageNameCatcher.java
 * Author       : Guansong Zhang
 * Created      : Thu May 14 12:03:02 1998
 * Revision     : $Revision: 1.5 $
 * Updated      : $Date: 1999/03/16 20:27:46 $
 */

package hpjava.compiler.reflection;

import java.io.*;
import java.util.*;

import hpjava.compiler.util.*;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

public class PackageNameCatcher extends DepthFirstVisitor {
  boolean debug=false;
  
  private StringBuffer nameBuffer;
  
  public PackageNameCatcher() {
    nameBuffer=new StringBuffer();
    debug=true;
  }    

  public PackageNameCatcher ( StringBuffer nameBuffer ) {
    this.nameBuffer=nameBuffer;
  }
  
   //
   // f0 -> [ PackageDeclaration() ]
   // f1 -> ( ImportDeclaration() )*
   // f2 -> ( TypeDeclaration() )*
   // f3 -> < EOF >
   //
   public void visit(CompilationUnit n) {   
      if ( n.f0.present() ) {
        n.f0.accept(this);
      }
      
      if (debug)
        System.out.println(nameBuffer);
   }
   
   //
   // f0 -> "package"
   // f1 -> Name()
   // f2 -> ";"
   //
   public void visit(PackageDeclaration n) {
      n.f1.accept(this);
   }

   //
   // f0 -> < IDENTIFIER >
   // f1 -> ( "." < IDENTIFIER >  )*
   //
   public void visit(Name n) {
     n.f0.accept(this);
     if (n.f1.present()) {
       n.f1.accept(this);
     }  
   }

  // NodeToken
  public void visit (NodeToken n) {
    nameBuffer.append(n.tokenImage);
  }

  public String toString() {
     return nameBuffer.toString();
  }
}
