/*
 * File         : PrimitiveTypeDescriptor.java
 * Author       : Guansong Zhang, Han-Ku Lee
 * Created      : Tue May 19 13:07:52 1998
 * Revision     : $Revision: 1.10 $
 * Updated      : $Date: 1999/12/01 19:06:59 $
 */

package hpjava.compiler.reflection;

import java.util.*;
import java.lang.reflect.*;

import hpjava.compiler.util.*;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

public class PrimitiveTypeDescriptor extends ScalarReflectorDescriptor {  

  public PrimitiveTypeDescriptor(PrimitiveType n) {
    PrimitiveTypeResolver resolver = new PrimitiveTypeResolver(this);
    n.accept(resolver);    
  }

  // the following is used to create PTD directly from their image;

  public PrimitiveTypeDescriptor(String image) {
    if (image.equals("boolean")) {
      this.signature="Z";
      this.name="boolean";
      return; 
    }  
    if (image.equals("char")) {
      this.signature="C";
      this.name="char";
      return;
    }  
    if (image.equals("byte")) {
      this.signature="B";
      this.name="byte";
      return;
    }  
    if (image.equals("short")) {
      this.signature="S";
      this.name="short";
      return;
    }  
    if (image.equals("int")) {
      this.signature="I";
      this.name="int";
      return;
    }  
    if (image.equals("long")) {
      this.signature="J";
      this.name="long";
      return;
    }  
    if (image.equals("float")) {
      this.signature="F";
      this.name="float";
      return;
    }  
    if (image.equals("double")) {
      this.signature="D";
      this.name="double";
      return;
    }            
    
  }

  public boolean isBoolean() {
    return this.signature.equals("Z");
  }

  public boolean isChar() {
    return this.signature.equals("C");
  }

  public boolean isByte() {
    return this.signature.equals("B");
  }

  public boolean isShort() {
    return this.signature.equals("S");
  }

  public boolean isInt() {
    return this.signature.equals("I");
  }

  public boolean isLong() {
    return this.signature.equals("J");
  }

  public boolean isFloat() {
    return this.signature.equals("F");
  }

  public boolean isDouble() {
    return this.signature.equals("D");
  }


}

class PrimitiveTypeResolver extends DepthFirstVisitor {
  PrimitiveTypeDescriptor ptd;

  public PrimitiveTypeResolver(PrimitiveTypeDescriptor ptd) {
    this.ptd = ptd;
  }

   //
   // f0 -> "boolean"
   //       | "char"
   //       | "byte"
   //       | "short"
   //       | "int"
   //       | "long"
   //       | "float"
   //       | "double"
   //
   public void visit(PrimitiveType n) {
      if (((NodeToken)n.f0.choice).tokenImage.equals("boolean")) {
        ptd.signature="Z";
	ptd.name="boolean";
        return; 
      }  
      if (((NodeToken)n.f0.choice).tokenImage.equals("char")) {
        ptd.signature="C";
	ptd.name="char";
        return;
      }  
      if (((NodeToken)n.f0.choice).tokenImage.equals("byte")) {
        ptd.signature="B";
	ptd.name="byte";
        return;
      }  
      if (((NodeToken)n.f0.choice).tokenImage.equals("short")) {
        ptd.signature="S";
	ptd.name="short";
        return;
      }  
      if (((NodeToken)n.f0.choice).tokenImage.equals("int")) {
        ptd.signature="I";
	ptd.name="int";
        return;
      }  
      if (((NodeToken)n.f0.choice).tokenImage.equals("long")) {
        ptd.signature="J";
	ptd.name="long";
        return;
      }  
      if (((NodeToken)n.f0.choice).tokenImage.equals("float")) {
        ptd.signature="F";
	ptd.name="float";
        return;
      }  
      if (((NodeToken)n.f0.choice).tokenImage.equals("double")) {
        ptd.signature="D";
	ptd.name="double";
        return;
      }        
   }

}
