/*
 * File         : ScopeManager.java
 * Author       : Guansong Zhang
 * Created      : Thu May 14 12:03:02 1998
 * Revision     : $Revision: 1.17 $
 * Updated      : $Date: 1999/10/26 21:36:47 $
 */

package hpjava.compiler.reflection;

import java.io.*;
import java.util.*;

import hpjava.compiler.util.*;

import hpjava.compiler.parser.*;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

// not all nodes registered.
// only IDENTIFIER and the node can be scope themself
// since the problem inside JTB, all tokens are registered now.

public class ScopeManager {

  static final String treeName = "hpjava.compiler.syntaxtree.";
  
  static final int NODENUM = 1024;
    
  static Hashtable nodeScopeTable  = new Hashtable (NODENUM);

  static ScopeDescriptor globalScope = new ScopeDescriptor();

  public ScopeManager(Node root){
    ScopeFinder sf = new ScopeFinder(nodeScopeTable);

    root.accept(sf);
  }

  public ScopeManager (Project prj) {
    ScopeFinder sf = new ScopeFinder(nodeScopeTable);
    
    Enumeration enum = prj.rootFileTable.keys();
   
    LOOP: while (enum.hasMoreElements()) {
      
      Node root = (Node) enum.nextElement();      

      root.accept(sf);
    
    }
    
    if (prj.isVerbose()) {
      //System.err.println("Node scope: "+nodeScopeTable+" created."); 
      System.err.println("Node scope created."); 
      System.err.println();
    }     

  }

  static public void add(NodeToken n){
    nodeScopeTable.put(n, ScopeFinder.scopeOfMain());
  }

  static public ScopeDescriptor scopeOf (Node n) {
  
    /*
    if (n instanceof CompilationUnit)
      return n;
    */
    
    ScopeDescriptor scope = (ScopeDescriptor) nodeScopeTable.get(n);

    if (scope == null) 
      throw new Error("Scope of "+n.toString()+" not found.");
    else
      return scope;
  }
  
/*
  static public Node rootOf (Node n) {
  
    while (!(n instanceof CompilationUnit))
      n = ScopeManager.scopeOf(n).node();
      
    return n;

  }  
*/
  
/* 
  static public String nameOf (Node n) {
    // change this to a better one in Java 1.2 ?
    // something like n.reference.toString()
    // toString() may be overrided by the real Node object later.
    // here the one for Object is used. 
    // Is there a way in Java to do objectOfSubC.C::fun() ?
    
    // or to create a class as ID descriptor, 
    // with name, scope and an id number as fields
      
    //  String fullName= n.super.toString(); // should work? super can only work inside a Class.

    String fullName= n.toString();
    
    if (fullName.startsWith(treeName))
      return (fullName.substring(treeName.length()));
    else
      throw new Error("Invalid scope: "+fullName);         
  }
*/

}

/*
 * File         : ScopeFinder.java
 * Author       : Guansong Zhang
 * Created      : Thu May 14 12:03:02 1998
 * Revision     : $Revision: 1.17 $
 * Updated      : $Date: 1999/10/26 21:36:47 $
 */

class ScopeFinder extends DepthFirstVisitor {

  boolean debug=false;

  final int DEFAULT = 0;
  final int TYPEDEC = 1;

  private IntStack status=new IntStack();
  
  private Stack scopeStack = new Stack();

  protected Hashtable nodeScopeTable;

  public ScopeFinder() {
    nodeScopeTable = new Hashtable ();
    status.push(DEFAULT);

    debug=true;
  }

  public ScopeFinder(Hashtable nodeScopeTable) {
    this.nodeScopeTable = nodeScopeTable;
    status.push(DEFAULT);
  }

  private String id = "";

  static private ScopeDescriptor scp = null;

  static public ScopeDescriptor scopeOfMain(){
    return scp;
  }

   //
   // User-generated visitor methods below
   //
   //
   // f0 -> [ PackageDeclaration() ]
   // f1 -> ( ImportDeclaration() )*
   // f2 -> ( TypeDeclaration() )*
   // f3 -> < EOF >
   //
   public void visit(CompilationUnit n) {
     nodeScopeTable.put(n, ScopeManager.globalScope);
        
     scopeStack.push(new ScopeDescriptor(n)); 
      
     if ( n.f2.present() ) {
       n.f2.accept(this);
     }
     if (debug) 
      System.out.println(nodeScopeTable);
   }

   //
   // f0 -> ( "abstract" | "final" | "public" )*
   // f1 -> UnmodifiedClassDeclaration()
   //
   public void visit(ClassDeclaration n) {
      nodeScopeTable.put(n, scopeStack.peek());
      n.f1.f1.accept(this);
      n.f1.f2.accept(this);
      n.f1.f3.accept(this);
      scopeStack.push(new ScopeDescriptor(n));
      n.f1.f4.accept(this);
      scopeStack.pop();
   }

   //
   // f0 -> ( "abstract" | "public" )*
   // f1 -> UnmodifiedInterfaceDeclaration()
   //
   public void visit(InterfaceDeclaration n) {
      nodeScopeTable.put(n, scopeStack.peek());
      n.f1.f1.accept(this);
      n.f1.f2.accept(this);
      n.f1.f3.accept(this);
      scopeStack.push(new ScopeDescriptor(n));
      n.f1.f4.accept(this);
      scopeStack.pop();
   }
   
   //
   // f0 -> ( "static" | "abstract" | "final" | "public" | "protected" | "private" )*
   // f1 -> UnmodifiedClassDeclaration()
   //
   public void visit(NestedClassDeclaration n) {
      nodeScopeTable.put(n, scopeStack.peek());
      n.f1.f1.accept(this);
      n.f1.f2.accept(this);
      n.f1.f3.accept(this);
      scopeStack.push(new ScopeDescriptor(n));
      n.f1.f4.accept(this);
      scopeStack.pop();
   }
 
   //
   // f0 -> ( "static" | "abstract" | "final" | "public" | "protected" | "private" )*
   // f1 -> UnmodifiedInterfaceDeclaration()
   //
   public void visit(NestedInterfaceDeclaration n) {
      nodeScopeTable.put(n, scopeStack.peek());
      n.f1.f1.accept(this);
      n.f1.f2.accept(this);
      n.f1.f3.accept(this);
      scopeStack.push(new ScopeDescriptor(n));
      n.f1.f4.accept(this);
      scopeStack.pop();
   }

   //
   // f0 -> "class"
   // f1 -> <IDENTIFIER>
   // f2 -> [ "extends" Name() ]
   // f3 -> [ "implements" NameList() ]
   // f4 -> ClassBody()
   //
   public void visit(UnmodifiedClassDeclaration n) {
      nodeScopeTable.put(n, scopeStack.peek());
      n.f1.accept(this);
      n.f2.accept(this);
      n.f3.accept(this);
      scopeStack.push(new ScopeDescriptor(n));
      n.f4.accept(this);
      scopeStack.pop();
   }


   //
   // f0 -> [ "public" | "protected" | "private" ]
   // f1 -> <IDENTIFIER>
   // f2 -> FormalParameters()
   // f3 -> [ "throws" NameList() ]
   // f4 -> "{"
   // f5 -> [ ExplicitConstructorInvocation() ]
   // f6 -> ( BlockStatement() )*
   // f7 -> "}"
   //
   public void visit(ConstructorDeclaration n) {
      nodeScopeTable.put(n, scopeStack.peek());
      scopeStack.push(new ScopeDescriptor(n));
      n.f1.accept(this);
      n.f2.accept(this);
      n.f5.accept(this);
      n.f6.accept(this);
      scopeStack.pop();
   }

   //
   // f0 -> ( "public" | "protected" | "private" | "static" | "abstract" | "final" | "native" | "synchronized" )*
   // f1 -> ResultType()
   // f2 -> MethodDeclarator()
   // f3 -> [ "throws" NameList() ]
   // f4 -> ( Block() | ";" )
   //
   public void visit(MethodDeclaration n) {
      nodeScopeTable.put(n, scopeStack.peek());
      n.f1.accept(this);
      scopeStack.push(new ScopeDescriptor(n));

      if(n.f2.f0.tokenImage.equals("main")){
        id = "main";
        //System.out.println("n.f2.f0.tokenImage is MAIN in ScopeManager ...");
      }

      n.f2.accept(this);
      n.f3.accept(this);
      n.f4.accept(this);
      scopeStack.pop();
   }

   //
   // f0 -> ( "public" | "protected" | "private" | "static" | "final" | "transient" | "volatile" )*
   // f1 -> Type()
   // f2 -> VariableDeclarator()
   // f3 -> ( "," VariableDeclarator() )*
   // f4 -> ";"
   //
   public void visit(FieldDeclaration n) {
      nodeScopeTable.put(n, scopeStack.peek());
      n.f1.accept(this);
      n.f2.accept(this);
      n.f3.accept(this);
   }

   //
   // f0 -> "{"
   // f1 -> ( BlockStatement() )*
   // f2 -> "}"
   //
   public void visit(Block n) {
      nodeScopeTable.put(n, scopeStack.peek());
      scopeStack.push(new ScopeDescriptor(n));
      n.f1.accept(this);

      if(id.equals("main")){
        scp = new ScopeDescriptor(n);
        id = "";
      }

      scopeStack.pop();
   }
   
   //
   // f0 -> "for"
   // f1 -> "("
   // f2 -> [ ForInit() ]
   // f3 -> ";"
   // f4 -> [ Expression() ]
   // f5 -> ";"
   // f6 -> [ ForUpdate() ]
   // f7 -> ")"
   // f8 -> Statement()
   //
   public void visit(ForStatement n) {
      nodeScopeTable.put(n, scopeStack.peek());
      scopeStack.push(new ScopeDescriptor(n));

      n.f2.accept(this);
      n.f4.accept(this);
      n.f6.accept(this);

      n.f8.accept(this);

      scopeStack.pop();
   }

   //
   // f0 -> "try"
   // f1 -> Block()
   // f2 -> ( "catch" "(" FormalParameter() ")" Block() )*
   // f3 -> [ "finally" Block() ]
   //
   public void visit(TryStatement n) {

      n.f1.accept(this);

      if (n.f2.present()) {
	for (int i = 0; i<n.f2.size(); i++) {
	  NodeSequence sequence =(NodeSequence) n.f2.elementAt(i);

	  nodeScopeTable.put(sequence, scopeStack.peek());

	  scopeStack.push(new ScopeDescriptor(sequence));

	  sequence.elementAt(2).accept(this);
	  sequence.elementAt(4).accept(this);
	  
	  scopeStack.pop();

	}

      }

      n.f3.accept(this);
   }
         
   // 
   // f0 -> "at" 
   // f1 -> "(" 
   // f2 -> NamedLocationList() 
   // f3 -> ")" 
   // f4 -> Statement() 
   // 
   public void visit(AtStatement n) {
      nodeScopeTable.put(n, scopeStack.peek());
      scopeStack.push(new ScopeDescriptor(n));
      n.f2.accept(this);

      n.f4.accept(this);

      scopeStack.pop();

   }

   // 
   // f0 -> "overall" 
   // f1 -> "(" 
   // f2 -> NamedLocationRangeList() 
   // f3 -> ")" 
   // f4 -> Statement() 
   // 
   public void visit(OverallStatement n) {
      nodeScopeTable.put(n, scopeStack.peek());
      scopeStack.push(new ScopeDescriptor(n));
      n.f2.accept(this);

      n.f4.accept(this);
 
      scopeStack.pop();
   }

  // define every identifier
  
  public void visit (NodeToken n) {
    if (n.kind == HPJavaParserConstants.IDENTIFIER)
      nodeScopeTable.put(n, scopeStack.peek());
  }

}


  


// to do list

/*

Is there a way to automat the procedure?

scope of the statement level.

Do not register all tokens.

*/
