/** 
 * Author       : Guansong Zhang 
 * Revision     : $Revision: 1.2 $ 
 * Updated      : $Date: 1999/03/17 16:36:45 $ 
 * Node         : Generated from JTB 1.1.1 modified by author. 
 * Copyright    : All rights reserved 
 */

package hpjava.compiler.syntaxtree;

import java.util.*;
/**
 * Represents a single token in the grammar.  If the "-tk" option
 * is used, also contains a Vector of preceding special tokens.
 */
public class NodeToken extends SyntaxNode implements Node {
   public NodeToken(String s) {
      this(s, -1, -1, -1, -1, -1);    }

   public NodeToken(String s, int kind, int beginLine, int beginColumn, int endLine, int endColumn) {
      tokenImage = s;
      specialTokens = null;
      this.kind = kind;
      this.beginLine = beginLine;
      this.beginColumn = beginColumn;
      this.endLine = endLine;
      this.endColumn = endColumn;
   }

   public NodeToken getSpecialAt(int i) {
      if ( specialTokens == null )
         throw new NoSuchElementException("No specials in token");
      return (NodeToken)specialTokens.elementAt(i);
   }

   public int numSpecials() {
      if ( specialTokens == null ) return 0;
      return specialTokens.size();
   }

   public void addSpecial(NodeToken s) {
      if ( specialTokens == null ) specialTokens = new Vector();
      specialTokens.addElement(s);
      s.setParent(this);
   }

   public void trimSpecials() {
      if ( specialTokens == null ) return;
      specialTokens.trimToSize();
   }

   public String toString()     { return tokenImage; }

   public String withSpecials() {
      if ( specialTokens == null )
          return tokenImage;

       StringBuffer buf = new StringBuffer();

       for ( Enumeration e = specialTokens.elements(); e.hasMoreElements(); )
          buf.append(e.nextElement().toString());

       buf.append(tokenImage);
       return buf.toString();
   }

   public void accept(hpjava.compiler.visitor.Visitor v) {
      v.visit(this);
   }

   public void setParent(Node n) { parent = n; }
   public Node getParent() { 
      if (parent == null) throw new Error("Parent not set.");
      return parent; 
   }

   public void setChildAt(Node n, int i) { 
      throw new Error("No children for NodeToken.");  
   }
   public Vector getChildren() { 
      throw new Error("No children for NodeToken.");  
   }

   private Node parent;
   public String tokenImage;

   // Stores a list of NodeTokens
   public Vector specialTokens;

   // -1 for these ints means no position info is available.
   public int beginLine, beginColumn, endLine, endColumn;

   // Equal to the JavaCC token "kind" integer.
   // -1 if not available.
   public int kind;
}

