/*
 * File         : SyntaxNode.java
 * Author       : Guansong Zhang
 * Created      : Thu Oct  8 13:00:47 1998
 * Revision     : $Revision: 1.1 $
 * Updated      : $Date: 1999/03/16 20:29:27 $
 */

package hpjava.compiler.syntaxtree;

import java.util.*;

import hpjava.compiler.unparser.*;

abstract public class SyntaxNode implements Node {

  public SyntaxNode getParent(int k) {
    Node parent = null;
    Node child = this;

    if (k <1) 
      throw new Error("Invalid level.");

    for (int i = 0; i< k; i++) {
      parent = child.getParent();
      child = parent;
    }
    return (SyntaxNode) parent;
  }

  protected int indexOfChildAs(Node n) {
    Vector vec = getChildren();
    return vec.indexOf(n);
  }

  public void replaceWith(Node n) {
    SyntaxNode parent = this.getParent(1);    
    int index = parent.indexOfChildAs(this);

    parent.setChildAt(n, index);

    this.setParent(null);
  }

  public void destroy() {
    SyntaxNode parent = this.getParent(1);

    int index = parent.indexOfChildAs(this);

    if ( !(parent instanceof NodeListInterface) ) {
      throw new Error("Parent is not a NodeListInterface."); 
    }

    Vector vec = ((NodeListInterface) parent).getChildren();
    vec.removeElementAt(index);
    this.setParent(null);
  }

  public void insertBefore(Node n) {
    SyntaxNode parent = this.getParent(1);

    int index = parent.indexOfChildAs(this);

    if ( !(parent instanceof NodeListInterface) ) {
      throw new Error("Parent is not a NodeListInterface."); 
    }

    Vector vec = ((NodeListInterface) parent).getChildren();
    vec.insertElementAt(n, index);

    if (n != null) 
      n.setParent(this.getParent());
    else
      throw new Error("Can not insert null.");
  }

  public void insertAfter(Node n) {
    SyntaxNode parent = this.getParent(1);

    int index = parent.indexOfChildAs(this);

    if ( !(parent instanceof NodeListInterface) ) {
      throw new Error("Parent is not a NodeListInterface."); 
    }

    Vector vec = ((NodeListInterface) parent).getChildren();
    vec.insertElementAt(n, index+1);

    if (n != null) 
      n.setParent(this.getParent());
    else
      throw new Error("Can not insert null.");
  }

  public void show() {
    IPrettyPrinter printer = new IPrettyPrinter(System.err);
    this.accept(printer);
    printer.flushWriter();
    System.err.println();
  }

  public void showParent(int level) {
    for (int count = 1; count <= level; count ++) {
      System.err.println("Parent "+ count + ": "+this.getParent(count).toString());
    }
  }

}

