/*
 * File         : AbstractExpressionTranslator.java
 * Author       : Guansong Zhang
 * Created      : Thu Oct  8 12:41:06 1998
 * Revision     : $Revision: 1.7 $
 * Updated      : $Date: 2000/01/05 22:07:30 $
 */

/* Modified by Qiang Zheng
 * after removing the AbstractPrimaryExpression from 
 * the AbstractExpression Hierachy
 */

package hpjava.compiler.translator;

import java.util.*;
import hpjava.compiler.util.*;

import hpjava.compiler.syntaxtree.*;

import hpjava.compiler.unparser.*;

import hpjava.compiler.abstractexp.*;
import hpjava.compiler.reflection.*;
import hpjava.compiler.wrapper.*;


public class AbstractExpressionTranslator {
  AbstractExpression exp;
  LinkedList blockStatementList;
  Prefix pre;
  
  public AbstractExpressionTranslator(AbstractExpression exp, LinkedList blockStatements, Prefix _pre) {
    this.exp = exp;
    blockStatementList = blockStatements;
    pre = _pre;
  }
  
  public void process() {
    AbstractExpressionProcessor processor = 
      new AbstractExpressionProcessor(blockStatementList, pre);

    this.exp.accept(processor);
  }
  
}

class AbstractExpressionProcessor 
  extends AbstractExpressionVisitor 
{

  final int PREFIX = 0;
  final int SUFFIX = 1;

  private ClassReflector cr;
  private AbstractExpression prefix;

  private LinkedList blockStatementList;
  private Prefix pre;
  
  private Hashtable transTable = new Hashtable();
  // a table used to store the original AbstractExpression
  // And translated result of AbstractExpression
  // The "key" of an element in this hashtable is the original
  // AbstractExpression. The "value" is the translated result
  // By pushing and getting from the hashtable, they actually
  // perform the function of replaceWith
  

  private IntStack status=new IntStack();
  int oldStatus = PREFIX;
  private StringBuffer image = new StringBuffer();

  public AbstractExpressionProcessor(LinkedList blockStatements, Prefix _pre) {
    // Added by Qiang Zheng
    // for simple expression with out concatenation, it is a prefix.
    status.push(PREFIX);
    blockStatementList = blockStatements;
    pre = _pre;
  }

  public void visit ( BinExprNode n ) 
  {
    this.prefix = n.left;
    //System.out.println("n.left.getClassReflector().getName() " + n.left.getClassReflector().getName());
    //System.out.println("n.binaryOperator = " + n.binaryOperator);
    //System.out.println("n.right.getClassReflector().getName() " + n.right.getClassReflector().getName());
    
    if(!n.left.isTypeChecked() || !n.right.isTypeChecked()){
      n.left.accept(this);
      n.right.accept(this);
    }
    else{

      if((n.binaryOperator==n.ADD || n.binaryOperator==n.SUBTRACT)
         && ( n.left.getClassReflector().getName().equals ("hpjava.lang.Location"))
         && ( n.right.getClassReflector().getName().equals("int") ) ){
        n.left.accept(this);
        AbstractExpression objName = (AbstractExpression)transTable.get(n.left);
        IdentifierNode mName = new IdentifierNode( new NodeToken("shift") );
        n.right.accept( this );
        ArgumentsNode arg =
            new ArgumentsNode((AbstractExpression)transTable.get(n.right));
        MethodCallNode method = new MethodCallNode(objName, mName, arg);

        if( n.binaryOperator == n.SUBTRACT ) {
          ArgumentsNode newArg
              = new ArgumentsNode( new UnaryExprNode(n.MINUS, n.right) );
          method = new MethodCallNode(objName, mName, newArg);
        }
        // n.replaceWith( method );
        transTable.put(n, method );
      }
      else{
        transTable.put( n,n );
        // Just do not translate any thing
        // for an expression if it is not shift
        n.left.accept(this);
        n.right.accept(this);
      }

    }      

  }

  public void visit ( AssignmentNode n ) 
  {
      // this.prefix = n.left;
    //( (AssignmentNode) n).left.accept( this );
    //( (AssignmentNode) n).right.accept( this );
    n.left.accept( this );
    n.right.accept( this );
  }

  public void visit ( LiteralNode n ) 
  {
    // don't need to translate literal
    transTable.put( n,n );
  }

  public void visit ( IdentifierNode n )
  {
    if(n.isTypeChecked()){
      // Temporary statement ... should be deleted...
      //blockStatementList.addLast(n);  

      if (status.peek()==PREFIX) {
        this.cr = n.getClassReflector();
      }
      else if (status.peek()==SUFFIX) {
        // System.out.println( "method name is: " + n.leaf.toString() );
      }
      else {
        throw new Error();
      }

      if(Flags.isInDistArray()){
        // the class reflectors of location
        if((n.getClassReflector().getName().equals("hpjava.lang.Location"))){
          // Generating a new "Location" type variable
          NodeToken nToken = Mixer.location(n.leaf);
          IdentifierNode niNode= new IdentifierNode(nToken);
          transTable.put(n, niNode);
          // n.replaceWith(niNode);
        } 
        else if((n.getClassReflector().getName().equals("int"))){
          // if not location, it should be an integer
          transTable.put(n, n);
        } 
        else{
          throw new Error();
        }
      } 
      else{
        transTable.put(n, n);
      }
    }
  }
  
  public void visit ( UnaryExprNode n ) 
  {
    if ( n.uOperator == n.POSTFIX_INDEX_LOCATION ) {
      Flags.setIdxFlag();
      n.replaceWith( n.operand );
      n.operand.accept( this );
      Flags.resetIdxFlag();
    } 
    else{
      ((UnaryExprNode) n).operand.accept( this );
    }
  }

  public void visit ( ArraySectNode n ) 
  {
    //if(n.toExpression() == null)
    //  System.out.println("null ...");
    //else
    //  System.out.println("NOT null ...");

    //System.out.println("In ArraySectNode ...");
    this.prefix = n.arrayName;
    ((ArraySectNode)n).arrayName.accept(this);
    dasrdldInSuffix( (ArraySectNode) n);
  }

  public void visit ( ArrayElemRefNode n ) 
  {
    this.prefix = n.arrayName;
    this.cr = n.getClassReflector();

    ((ArrayElemRefNode) n).arrayName.accept(this);
    Flags.setInArrayFlag();
    toeldInSuffix(( ArrayElemRefNode ) n);
    Flags.resetInArrayFlag();
  }

  public void visit ( ClassFieldNode n ) 
  {
    this.prefix = n.objName;

    ( (ClassFieldNode) n ).objName.accept(this);
    IdentifierNode in = ((ClassFieldNode)n).fieldName;

    AbstractExpressionTranslator translator = new AbstractExpressionTranslator((AbstractExpression) in, blockStatementList, pre);
    translator.process();
  }

  public void visit ( MethodCallNode n ) 
  {
    this.prefix = n.objName;
    
    status.push(PREFIX);

    n.objName.accept(this);

    oldStatus = status.pop();

    status.push( SUFFIX );

    IdentifierNode in = n.methodName;

    // System.out.println( "The top of status stack is: " + status.peek());

    in.accept( this );

    // AbstractExpressionTranslator translator0 = new
    // 	AbstractExpressionTranslator((AbstractExpression) in);

    //ArgumentsDescriptor args = (ArgumentsDescriptor) n.suffix;
    ArgumentsNode argsNd = (ArgumentsNode) ((MethodCallNode) n).args;
    ArgumentsDescriptor args = argsNd.descriptor;

    for (int i =0; i< args.numberOfArgs(); i++) {
      AbstractExpression aexp = args.abstractArg(i);
      AbstractExpressionTranslator translator1
        = new AbstractExpressionTranslator(aexp, blockStatementList, pre);

      translator1.process();
    }

    oldStatus = status.pop();

  }


  public void visit(ExpressionNode n) {
    AbstractExpression exp 
      = ExpressionManager.lookup(n.leaf);

    AbstractExpressionTranslator translator = 
      new AbstractExpressionTranslator(exp, blockStatementList, pre);

    translator.process();
    
  }

  // begin for TripletOrExpressionListDescriptor
  private void toeldInSuffix (ArrayElemRefNode n) {
    //System.out.println("this.cr.getName(): " + this.cr.getName());

    Stack prefixStack = new Stack();

    if (this.cr.isDistArray()) {
      Flags.setInDistArrayFlag();

      // TripletOrExpressionListDescriptor suffix =
      //  (TripletOrExpressionListDescriptor)n.suffix;

      TripletOrExpressionListNode suffix = n.eList;

      // AbstractExpression data;
      MethodCallNode data;

      // data = this.prefix.append(new IdentifierDescriptor("dat"));
      // data = this.prefix.append(new IdentifierNode("dat"));
      // data = data.append(new ArgumentsDescriptor());
      // data = data.append(new ArgumentsNode());

	  data = new MethodCallNode(this.prefix, 
		new IdentifierNode("dat"), new ArgumentsNode() );

      // AbstractPrimaryExpression position =
      //  this.prefix.append(new IdentifierDescriptor("pos"));

      // AbstractExpression position =
      //  this.prefix.append(new IdentifierNode("pos"));

      // ArgumentsDescriptor posArg = new ArgumentsDescriptor();
      ArgumentsNode posArg = new ArgumentsNode();

      prefixStack.push( this.prefix );

      for (int i = 0; i < suffix.descriptor.numberOfDims(); i++) {
        //ExpressionDescriptor ed = (ExpressionDescriptor)suffix.dim(i);
        // ExpressionNode eNd
        //    = (ExpressionNode)suffix.descriptor.dim(i);

        // ExpressionNode priExprNd =
        //    (ExpressionNode) ExpressionManager.lookup(eNd.leaf);

        ExpressionNode priExprNd
              = (ExpressionNode)suffix.descriptor.dim(i);

        AbstractExpression iNode =
            (AbstractExpression) ExpressionManager.lookup(priExprNd.leaf);

        //System.out.println("...... " + priExprNd.leaf.getClass().getName());

        iNode.accept(this);
        // in the accept/visit method, the newly translated abstraction
        // result is put in hashtable

        AbstractExpression newNode = (AbstractExpression)transTable.get(iNode);
        Expression empty = new Expression(null, new NodeOptional());
        ExpressionManager.register(empty, newNode);        
        //posArg.descriptor.addArg( newNode.toExpression() );
        posArg.descriptor.addArg(empty);
        /* possible error here */
        // End of the revision of Qiang Zheng
        // posArg.addArg(ed.leaf);
      }

      // System.out.println("$$$$: position:");
      // System.out.println(((IdentifierNode)this.prefix).leaf.tokenImage);

      this.prefix = (AbstractExpression)prefixStack.pop();

      // position = position.append(posArg);
      MethodCallNode position =
		new MethodCallNode(this.prefix, new IdentifierNode("pos"),
		  posArg);

      //data = data.append(new TripletOrExpressionListDescriptor(position));
      // data = data.append(new TripletOrExpressionListNode(position));
      // n.replaceWith(data);

	  ArrayElemRefNode aerNd = new ArrayElemRefNode(data, 
		new TripletOrExpressionListNode(position) );

	  n.replaceWith(aerNd);
      //transTable.put(n, aerNd);     

    } else {
      // if this is a range.
      System.out.println("Not finished in ArrayElemRefNode in AbstractExpressionTranslator ... ");
    }

    Flags.resetInDistArrayFlag();
  // Using the flags set/reset here might introduce
  // potential problem if the distributed Array element is nested
  }
  // end for TripletOrExpressionListDescriptor

  // begin for DistArraySecRefDimListDescriptor
  /*
  private void addSecGroup(ArraySectNode n,
                           ArgumentsDescriptor arg) {

    // DistArraySecRefDimListDescriptor dasrd =
    //  (DistArraySecRefDimListDescriptor)n.suffix;
    DistArraySecRefDimListNode dasrdn = n.dimList;

    DistArraySecRefDimListDescriptor dasrd = dasrdn.descriptor;

    // AbstractPrimaryExpression group =
    //  n.prefix.append(new IdentifierDescriptor("grp"));
    // group=group.append(new ArgumentsDescriptor());

    // AbstractExpression group =
    //  n.arrayName.append(new IdentifierNode("grp"));
    // group=group.append(new ArgumentsNode());

	AbstractExpression group = new MethodCallNode (
		n.arrayName, new IdentifierNode("grp"), new ArgumentsNode() );

    // for (int i =0; i<dasrd.numberOfDims(); i++)
    for (int i =0; i<dasrd.numberOfDims(); i++) {
      // if (dasrd.isExpressionDim(i))
      //  group=group.append(new IdentifierDescriptor("divide"));
      if (dasrd.isExpressionDim(i)) {
	   // group=group.append(new IdentifierNode("divide"));
       // group= group.append
       //  (new ArgumentsDescriptor(((ExpressionDescriptor)dasrd.dim(i)).leaf));
       // group= group.append (new ArgumentsNode(
       //     ((ExpressionNode)dasrd.dim(i)).leaf));
       // group=new BinExprNode( DIVIDE, group,
	   //	new ArgumentsNode( ((ExpressionNode)dasrd.dim(i)).leaf) );
       group=new BinExprNode( 18 , group,
	   	 new ArgumentsNode( ((ExpressionNode)dasrd.dim(i)).leaf) );
      }

    }

    arg.addArg(group);
  }
  */
  /*
  private void addSecDimension(ArraySectNode n,
                               ArgumentsDescriptor arg) {

    DistArraySecRefDimListNode dasrdn = n.dimList;
    DistArraySecRefDimListDescriptor dasrd = dasrdn.descriptor;

    //AbstractPrimaryExpression range =
    //  n.prefix.append(new IdentifierDescriptor("rng"));
    // range = range.append(new ArgumentsDescriptor());

    // AbstractExpression range =
    //   n.arrayName.append(new IdentifierNode("rng"));
    // range = range.append(new ArgumentsNode());

	MethodCallNode range = new MethodCallNode( n.arrayName,
		new IdentifierNode("rng"), new ArgumentsNode() );

    // AbstractPrimaryExpression stride =
    //  n.prefix.append(new IdentifierDescriptor("str"));
    //stride = stride.append(new ArgumentsDescriptor());

    // AbstractExpression stride =
    //  n.arrayName.append(new IdentifierNode("str"));
    // stride = stride.append(new ArgumentsNode());
	MethodCallNode stride = new MethodCallNode( n.arrayName,
		new IdentifierNode("str"), new ArgumentsNode() );

    for (int i =0; i<dasrd.numberOfDims(); i++) {
      if (!dasrd.isExpressionDim(i)) {  // should be not a location expression
        //AbstractPrimaryExpression dim = new LiteralDescriptor(i);
        AbstractExpression dim = new LiteralNode(i);

        //AbstractPrimaryExpression rng
        //  =range.append(new TripletOrExpressionListDescriptor(dim));
        // AbstractExpression rng
        //  =range.append(new TripletOrExpressionListNode(dim));
		ArrayElemRefNode rng = new ArrayElemRefNode (
			range, new TripletOrExpressionListNode(dim) );
        arg.addArg(rng);

        // AbstractPrimaryExpression str
        //  =stride.append(new TripletOrExpressionListDescriptor(dim));
        // AbstractExpression str
        //   =stride.append(new TripletOrExpressionListNode(dim));
		ArrayElemRefNode str = new ArrayElemRefNode (
			stride, new TripletOrExpressionListNode(dim) );

        arg.addArg(str);
      }
    }

  }
  */
  /*
  private void addSecDate(ArraySectNode n,
                          ArgumentsDescriptor arg) {

    DistArraySecRefDimListNode dasrdn = n.dimList;

    // AbstractExpression date = n.arrayName.append(new IdentifierNode("dat"));
    // date = date.append(new ArgumentsNode());

	MethodCallNode date = new MethodCallNode (
		n.arrayName, new IdentifierNode("dat") , new ArgumentsNode() );

    arg.addArg(date);

  }

  private void addSecOffset(ArraySectNode n, ArgumentsDescriptor arg) {
    DistArraySecRefDimListNode dasrdn = n.dimList;
    DistArraySecRefDimListDescriptor dasrd = dasrdn.descriptor;

    // AbstractPrimaryExpression parent =
    //  n.prefix.append(new IdentifierDescriptor("off"));
    // parent = parent.append(new ArgumentsDescriptor());

    // AbstractExpression parent =
    //  n.arrayName.append(new IdentifierNode("off"));
    // parent = parent.append(new ArgumentsNode());
    // AbstractExpression offset = parent.toAbstractExpression();

    AbstractExpression offset = new MethodCallNode(
	  n.arrayName, new IdentifierNode("off"), new ArgumentsNode() );
 
    // IdentifierDescriptor primaryOffset =
    //  new IdentifierDescriptor("offset");
    // AbstractPrimaryExpression stride =
    //  n.prefix.append(new IdentifierDescriptor("str"));

    IdentifierNode primaryOffset = new IdentifierNode("offset");

    // AbstractExpression stride =
    //  n.arrayName.append(new IdentifierNode("str"));

    for (int i =0; i<dasrd.numberOfDims(); i++) {
      if (dasrd.isExpressionDim(i)) {
        // AbstractPrimaryExpression dim = new LiteralDescriptor(i);
        AbstractExpression dim = new LiteralNode(i);

        //AbstractPrimaryExpression str = stride.append(new ArgumentsDescriptor());
        // AbstractExpression str = stride.append(new ArgumentsNode());
		MethodCallNode str = new MethodCallNode(
			n.arrayName, new IdentifierNode("str"), new ArgumentsNode() );

        //str = str.append(new TripletOrExpressionListDescriptor(dim));
        // str = str.append(new TripletOrExpressionListNode(dim));
        // str = str.append( new ArgumentsDescriptor(
        //    ((ExpressionDescriptor)dasrd.dim(i)).leaf ) );

        // str = str.append(primaryOffset);
        // str = str.append( new ArgumentsNode(
        //    ((ExpressionNode)dasrd.dim(i)).leaf ) );
		str = new MethodCallNode ( str, primaryOffset,
			    new ArgumentsNode(((ExpressionNode)dasrd.dim(i)).leaf) );

        if (offset == null) {
          // offset = str.toAbstractExpression();
          offset = str;
        }
        else {
          // offset = offset.add(str.toAbstractExpression());
          offset = new BinExprNode(15,offset,str);
        }

      }
    }
    arg.addArg(offset);
  }
  */
  private void dasrdldInSuffix( ArraySectNode n ) {
    //System.out.println("In dasrdldInSuffix ...");
   
    DistArraySecRefDimListNode dasrdn =
        ((ArraySectNode)n).dimList;
    DistArraySecRefDimListDescriptor dasrd = dasrdn.descriptor;

    //putStatementsList(dasrd);
    /*--------------------------------------------------------------------*/

    pre.namePrefix("tmp");
    pre.updatePrefix();
    NodeToken id = new NodeToken(pre.prefix());

    //System.out.println(".... Mixer.location(id): " + Mixer.location(id));
    //System.out.println(".... Mixer.group(id): " + Mixer.group(id));    

    MethodCallNode aexp_loc = new MethodCallNode((IdentifierNode)this.prefix,
                                                 new IdentifierNode(new NodeToken("rng")),
                                                 new ArgumentsNode());
    int[] _dim = new int[2];
    
    if(dasrd.isExpressionDim(0)){
      _dim[0] = 0;
      _dim[1] = 1;
    }
    else{
      _dim[0] = 1;
      _dim[1] = 0;
    }
    
    ArgumentsNode arg_loc;
    ExpressionNode priExprNd;
    
    if(_dim[0] == 0){
      priExprNd = (ExpressionNode)dasrd.dim(0);
    }
    else{
      priExprNd = (ExpressionNode)dasrd.dim(1);
    } 
    AbstractExpression iNode = (AbstractExpression)ExpressionManager.lookup(priExprNd.leaf);
    //System.out.println("..... " + priExprNd.getClass().getName());
    arg_loc = new ArgumentsNode(iNode);

    ArrayElemRefNode aerNd_loc = new ArrayElemRefNode(aexp_loc, 
                                                      new TripletOrExpressionListNode(new IdentifierNode(new NodeToken(Integer.toString(_dim[0])))));

    MethodCallNode aexp_loc_2 = new MethodCallNode(aerNd_loc,
                                                 new IdentifierNode(new NodeToken("location")),
                                                 arg_loc);
    
    Expression empty_loc = new Expression(null, new NodeOptional());
    //ExpressionManager.register(empty_loc, aexp_loc);
    //ExpressionManager.register(empty_loc, aerNd_loc);
    ExpressionManager.register(empty_loc, aexp_loc_2);

    BlockStatement blockStat_loc = locationDec(id, empty_loc);
    //blockStat_loc.show();

    MethodCallNode aexp_grp = new MethodCallNode((IdentifierNode)this.prefix,
                                                 new IdentifierNode(new NodeToken("grp")),
                                                 new ArgumentsNode());

    aexp_grp = new MethodCallNode(aexp_grp,
                                  new IdentifierNode(new NodeToken("clone")),
                                  new ArgumentsNode());

    Name name = new Name(new NodeToken("Group"), new NodeListOptional());
    NodeChoice Ndc = new NodeChoice(name, 1);
    Type _type = new Type(Ndc, new NodeListOptional());
    TypeNode typeNd = new TypeNode(_type);

    CastNode aexp_cast = new CastNode(aexp_grp, typeNd);

    Expression empty_grp = new Expression(null, new NodeOptional());
    //ExpressionManager.register(empty_grp, aexp_grp);
    ExpressionManager.register(empty_grp, aexp_cast);

    BlockStatement blockStat_grp = groupDec(id, empty_grp);
    //blockStat_grp.show();

    AbstractExpression tmp_group = new IdentifierNode(Mixer.group(id));
    AbstractExpression tmp_location = new ArgumentsNode(new IdentifierNode(Mixer.location(id)));
    
    MethodCallNode aexp = new MethodCallNode(tmp_group,
                                             new IdentifierNode(new NodeToken("restrict")),
                                             tmp_location);
    
    StatementExpression empty = new StatementExpression(new NodeChoice(null,2));
    ExpressionManager.register(empty, aexp);
    NodeSequence seq = new NodeSequence(1);
    seq.addNode(empty);
    seq.addNode(new NodeToken("; "));

    Statement stat = new Statement(new NodeChoice(seq, 3));
    BlockStatement blockStat = new BlockStatement(new NodeChoice(stat, 1));

    //blockStat.show();

    NodeListOptional optional = new NodeListOptional();
    optional.addNode(blockStat_loc);
    optional.addNode(blockStat_grp);
    optional.addNode(blockStat);
    
    Block block = new Block(new NodeToken(""), optional, new NodeToken(""));
    
    //block.show();
    
    blockStatementList.addLast(block);

    /*--------------------------------------------------------------------*/       
    ClassReflectorByDescriptor cr =
      (ClassReflectorByDescriptor)((ArraySectNode)n).arrayName.getClassReflector();

    ClassReflector base = cr.getComponentType();

    if (!base.isPrimitive())
      throw new Error("Not finished");

    // the following line should be changed to deal with none primitive distarray.
    //System.out.println("base.getName(): " + base.getName());
    
    PrimitiveTypeDescriptor baseDescriptor =
      new PrimitiveTypeDescriptor(base.getName());

    ArraySignaturesImage image =
      new ArraySignaturesImage(dasrd.secRefSignatureOf(cr));

    ClassReflectorDescriptor crd =
      new ArrayReflectorDescriptor(baseDescriptor,image.name);
   
    ClassReflector secReflector =
      new ClassReflectorByDescriptor(crd);

    ArgumentsDescriptor arg = new ArgumentsDescriptor();

    //addSecGroup(n, arg);
    
    arg.addArg(tmp_group);

    ArrayElemRefNode aerNd_rng = new ArrayElemRefNode(aexp_loc, new TripletOrExpressionListNode(new IdentifierNode(new NodeToken(Integer.toString(_dim[1])))));

    arg.addArg(aerNd_rng);
    
    MethodCallNode aexp_str = new MethodCallNode((IdentifierNode)this.prefix,
                                                 new IdentifierNode(new NodeToken("str")),
                                                 new ArgumentsNode());

    ArrayElemRefNode aerNd_str = new ArrayElemRefNode(aexp_str, new TripletOrExpressionListNode(new IdentifierNode(new NodeToken(Integer.toString(_dim[1])))));

    arg.addArg(aerNd_str);

    MethodCallNode aexp_dat = new MethodCallNode((IdentifierNode)this.prefix,
                                                 new IdentifierNode(new NodeToken("dat")),
                                                 new ArgumentsNode());

    arg.addArg(aexp_dat);

    MethodCallNode aexp_off = new MethodCallNode((IdentifierNode)this.prefix,
                                                 new IdentifierNode(new NodeToken("off")),
                                                 new ArgumentsNode());

    aerNd_str = new ArrayElemRefNode(aexp_str, new TripletOrExpressionListNode(new IdentifierNode(new NodeToken(Integer.toString(_dim[0])))));

    MethodCallNode aexp_offset = new MethodCallNode(aerNd_str,
                                                    new IdentifierNode(new NodeToken("offset")),
                                                    tmp_location);

    BinExprNode bin_offset = new BinExprNode(n.ADD, aexp_off, aexp_offset);

    arg.addArg(bin_offset);
    
    //addSecDimension(n, arg);
    //addSecDate(n, arg);
    //addSecOffset(n, arg);

    AllocationExpressionNode aedn =
      new AllocationExpressionNode(secReflector);
    AllocationExpressionDescriptor aed = aedn.descriptor;

    aed.addArguments(arg);

	n.replaceWith(aedn);

    //System.out.println(".... this.prefix: " + ((IdentifierNode)this.prefix).leaf);

  }
  // end for DistArraySecRefDimListDescriptor


  // this is for a simple distributed array.
  // here is a temp one.
  // it is much more complicated if you provide a complete solution
  // the AllocationExpression can not be translated into a single expression.
  // such as multiple allocated shapes.
  // private void aedAsPrefix (AllocationExpressionDescriptor n)
  private void aedAsPrefix (AllocationExpressionNode n) 
  {

    // AllocationExpression alloc = n.leaf;
    // AllocationExpression alloc = n.descriptor.leaf;

    // AllocationExpressionDescriptor allocDescriptor
    //  = new AllocationExpressionDescriptor(this.cr);

    ArgumentsDescriptor args = new ArgumentsDescriptor();

    // AbstractPrimaryExpression group = new IdentifierDescriptor("SpmdEnv");
    // group = group.append(new IdentifierDescriptor("apg"));

    AbstractExpression oName = new IdentifierNode("SpmdEnv");
    ClassFieldNode group =
      new ClassFieldNode( oName, (new IdentifierNode("apg")) );

    args.addArg(group);
 
    // AllocatedShapeDescriptor descriptor = n.allocatedShape(0);
    AllocatedShapeDescriptor descriptor = n.descriptor.allocatedShape(0);

    for (int i =0; i< descriptor.numberOfDims(); i++) {
      args.addArg(descriptor.dim(i));
    }

    AllocationExpressionNode allocNode
      = new AllocationExpressionNode(this.cr);

    AllocationExpressionDescriptor allocDescriptor = allocNode.descriptor;

    allocDescriptor.addArguments(args);

    // n.replaceWith(allocNode);
    n.replaceDescriptor(allocDescriptor);

    // System.out.println("parent of allocation expression !"
    //    + n.getParent().getClass().getName());

  }

  public void visit (AllocationExpressionNode n) 
  {
    if (status.peek()==PREFIX) {
      this.cr = n.getClassReflector();
      if (this.cr != null) { // not finished

        if (this.cr.isDistArray()) {
          aedAsPrefix(n);
        }

      }
    }
    else if (status.peek()==SUFFIX) {

    }
    else {
      throw new Error();
    }
  }

  private BlockStatement groupDec(NodeToken id,Expression group) {
    TypeDescriptor type = 
      new TypeDescriptor(new NodeToken("Group"));

    VariableDeclaratorDescriptor var = 
      new VariableDeclaratorDescriptor(new VariableDeclaratorId(Mixer.group(id), new NodeListOptional()));
    var.setInitializer(group);

    LocalVariableDeclarationDescriptor dec = 
      new LocalVariableDeclarationDescriptor(type.leaf, var.leaf);


    NodeSequence sequence = new NodeSequence(2);
    sequence.addNode(dec.leaf);
    sequence.addNode(new NodeToken(";"));

    return  new BlockStatement(new NodeChoice(sequence, 0));
  }
  
  private BlockStatement locationDec(NodeToken id,Expression location) {
    TypeDescriptor type = 
      new TypeDescriptor(new NodeToken("Location"));

    VariableDeclaratorDescriptor var = 
      new VariableDeclaratorDescriptor(new VariableDeclaratorId(Mixer.location(id), new NodeListOptional()));
    var.setInitializer(location);

    LocalVariableDeclarationDescriptor dec = 
      new LocalVariableDeclarationDescriptor(type.leaf, var.leaf);


    NodeSequence sequence = new NodeSequence(2);
    sequence.addNode(dec.leaf);
    sequence.addNode(new NodeToken(";"));

    return  new BlockStatement(new NodeChoice(sequence, 0));
  }
  

}
