/*
 * File         : ConstructTranslator.java
 * Author       : Guansong Zhang
 * Created      : Tue Oct 27 11:53:19 1998
 * Revision     : $Revision: 1.17 $
 * Updated      : $Date: 2000/01/05 22:07:30 $
 */

package hpjava.compiler.translator;

import hpjava.compiler.util.*;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

import hpjava.compiler.unparser.*;

import hpjava.compiler.reflection.*;
import hpjava.compiler.abstractexp.*;
import hpjava.compiler.wrapper.*;

public class ConstructTranslator {

  Node target;
  Prefix pre;

  public ConstructTranslator(OnStatement n) {
    target = n;    
  }

  public ConstructTranslator(OnStatement n, Prefix p) {
    target = n;
    this.pre = p;
  }

  public ConstructTranslator(OverallStatement n) {
    target = n;    
  }

  public ConstructTranslator(AtStatement n) {
    target = n;    
  }

  public void process() {
    //ConstructProcessor processor = new ConstructProcessor();
    ConstructProcessor processor = new ConstructProcessor(pre);
    target.accept(processor);
  }

}

class ConstructProcessor extends DepthFirstVisitor {
  final int DEFAULT =0;
  final int FORON = 1;

  IntStack status = new IntStack();

  Prefix pre;

  ConstructProcessor() {
  }

  ConstructProcessor(Prefix p) {
    this.pre = p;
  }

    // constructor has side effects on its input. 
    // a copy should be used?
    // if not, the node should be deleted.

  private BlockStatement groupDec(NodeToken id,Expression group) {
  //private BlockStatement groupDec(Expression group) {
    TypeDescriptor type = 
      new TypeDescriptor(new NodeToken("Group"));

    VariableDeclaratorDescriptor var = 
      new VariableDeclaratorDescriptor(new VariableDeclaratorId(Mixer.group(id), new NodeListOptional()));
   var.setInitializer(group);

    LocalVariableDeclarationDescriptor dec = 
      new LocalVariableDeclarationDescriptor(type.leaf, var.leaf);


    NodeSequence sequence = new NodeSequence(2);
    sequence.addNode(dec.leaf);
    sequence.addNode(new NodeToken(";"));

    return  new BlockStatement(new NodeChoice(sequence, 0));
  }

  private BlockStatement locationDecl(NodeToken id, Expression location) {
    TypeDescriptor type = new TypeDescriptor(new NodeToken("Location"));

    /* VariableDeclaratorDescriptor var = 
      new VariableDeclaratorDescriptor(new VariableDeclaratorId(Mixer.group(),
								new NodeListOptional())); */

    VariableDeclaratorDescriptor var = 
      new VariableDeclaratorDescriptor(new VariableDeclaratorId(id,
								new NodeListOptional()));
    var.setInitializer(location);

    LocalVariableDeclarationDescriptor dec = 
      new LocalVariableDeclarationDescriptor(type.leaf, var.leaf);

    NodeSequence sequence = new NodeSequence(2);
    sequence.addNode(dec.leaf);
    sequence.addNode(new NodeToken(";"));

    return  new BlockStatement(new NodeChoice(sequence, 0));
  }

  public void visit (AtStatement n) {
	//System.out.println("n.getParent(4): " + n.getParent(4).getClass().getName());
    NodeListOptional optional = new NodeListOptional();

    NamedLocationListDescriptor descriptor =
      new NamedLocationListDescriptor (n.f2);

	// To generate a list of location declaration statement. i.e.:
	// we have at(i = ... ,j = ... ,k = ...) in HPJava language
	// after translation, it becomes:
	// Location i_loc_hpj = ...; // Generated by this method;
	// Location j_loc_hpj = ...; // Generated by this method;
	// Location k_loc_hpj = ...; // Generated by this method;
    // if( i_loc_hpj.at() ) {
    //  if( j_loc_hpj.at() ) {
    //   if( k_loc_hpj.at() ) {
    //    ... 
    //   } k_loc_hpj.ta();
    //  } j_loc_hpj.ta(); 
    // } i_loc_hpj.ta();
	// Rules used here:
	// FieldDeclaration ::= Type VariableDeclarator
	// Variable Declarator ::= <IDENTIFIER> = VariableInitializer;
    for (int i = 0; i<descriptor.numberOfLocs(); i++) {
      // NodeToken id = descriptor.name(i);
      // NodeToken id = Mixer.index( descriptor.name(i) );
      NodeToken id = Mixer.location( descriptor.name(i) );
      PrimaryExpression locationExpr = descriptor.loc(i).f2;

      // AbstractPrimaryExpression apExpr = PrimaryExpressionManager.lookup(locationExpr);
      // optional.addNode(locationDecl(id, apExpr.toExpression()));

      AbstractExpression apExpr = ExpressionManager.lookup(locationExpr);
      Expression empty = new Expression(null, new NodeOptional());
      ExpressionManager.register(empty, apExpr);        

      //optional.addNode(locationDecl(id, apExpr.toExpression()));
      optional.addNode(locationDecl(id, empty));

    }

	// To generate nested "at" construct i.e. Before the translation, 
    // we have at(i = ... ,j = ... ,k = ...) {...} in HPJava language
    // after translation, it becomes:
	// Location i = ...;
    // Location j = ...;
    // Location k = ...;
	// if( i.at() ) {   // Generated by this method
	//	if( j.at() ) {  // Generated by this method
	//	 if( k.at() ) { // Generated by this method
	//	  ... 			// Original statements in "at" construct are kept.
    //   } k.ta(); 		// Generated by this method
    //  } j.ta();  		// Generated by this method
    // } i.ta();   		// Generated by this method
	// The nested construct is generated from the inner most level to outmost level

	Statement stat;
	BlockStatement blockStat; 
	NodeSequence sequence = new NodeSequence(1); 

	Statement nestedStmt = n.f4;
    for (int i = descriptor.numberOfLocs()-1; i >= 0; i--) {

      // NodeToken id = descriptor.name(i);
      // NodeToken id = Mixer.index( descriptor.name(i) );
      NodeToken id = Mixer.location( descriptor.name(i) );

	  // AbstractPrimaryExpression aexp = new IdentifierDescriptor(id);
	  // aexp = aexp.append(new IdentifierDescriptor(new NodeToken("at")));
	  // aexp = aexp.append(new ArgumentsDescriptor());

	  // AbstractPrimaryExpression aexp = new IdentifierNode(id);
	  // aexp = aexp.append(new IdentifierNode(new NodeToken("at")));
	  // aexp = aexp.append(new ArgumentsNode());

	  MethodCallNode aexp = new MethodCallNode(
		new IdentifierNode(id)
		, new IdentifierNode(new NodeToken("at"))
		, new ArgumentsNode());

      /* IfStatement ifStat = new IfStatement(new NodeToken("if"),
              n.f1,
              aexp.toAbstractExpression().toExpression(),
              n.f3,
              nestedStmt,
              new NodeOptional()); //NodeOptional is "else" here */

      Expression empty = new Expression(null, new NodeOptional());
      ExpressionManager.register(empty, aexp);        

      IfStatement ifStat = new IfStatement(new NodeToken("if"),
                                           n.f1, empty,
                                           //aexp.toExpression(),
                                           n.f3,
                                           nestedStmt,
                                           new NodeOptional()); //NodeOptional is "else" here

	  stat = new Statement(new NodeChoice(ifStat, 5));

	  stat = new Statement(new NodeChoice(ifStat, 5));
      blockStat = new BlockStatement(new NodeChoice(stat,1));
	  nestedStmt = new Statement(new NodeChoice(blockStat, 1));

      optional.addNode(nestedStmt);

      // aexp = new IdentifierDescriptor(id);
	  // aexp = aexp.append(new IdentifierDescriptor(new NodeToken("ta")));
	  // aexp = aexp.append(new ArgumentsDescriptor());

      /* aexp = new IdentifierNode(id);
	  aexp = aexp.append(new IdentifierNode(new NodeToken("ta")));
	  aexp = aexp.append(new ArgumentsNode()); */

      aexp = new MethodCallNode( new IdentifierNode(id)
		, new IdentifierNode(new NodeToken("ta"))
		, new ArgumentsNode() );

	  sequence = new NodeSequence(2);
      // sequence.addNode(aexp.toAbstractExpression().toStatementExpression());
      StatementExpression empty2 = new StatementExpression(new NodeChoice(null));
      ExpressionManager.register(empty2, aexp);        

      //sequence.addNode(aexp.toStatementExpression());
      sequence.addNode(empty2);
      sequence.addNode(new NodeToken(";"));

      stat = new Statement(new NodeChoice(sequence,3));
      blockStat = new BlockStatement(new NodeChoice(stat,1));

      optional.addNode(blockStat);

    }

    Block block = new Block ( new NodeToken("{"),
                              optional,
                              new NodeToken("}") );

    // block.show();

    n.getParent(4).insertAfter(block);
    n.getParent(4).destroy();
  }

  public void visit (OnStatement n) {
    NodeListOptional optional = new NodeListOptional();

    pre.namePrefix("tmp");
    pre.updatePrefix();

    NodeToken id = new NodeToken(pre.prefix());
    
    optional.addNode(groupDec(id, n.f2));

    // AbstractPrimaryExpression aexp = new IdentifierDescriptor(Mixer.group());
    // aexp = aexp.append(new IdentifierDescriptor(new NodeToken("on")));
    // aexp = aexp.append(new ArgumentsDescriptor());

    // AbstractPrimaryExpression aexp = new IdentifierNode(Mixer.group());
    // aexp = aexp.append(new IdentifierNode(new NodeToken("on")));
    // aexp = aexp.append(new ArgumentsNode());

	MethodCallNode aexp = new MethodCallNode(
        new IdentifierNode(Mixer.group(id))
        , new IdentifierNode(new NodeToken("on"))
        , new ArgumentsNode());

    /* IfStatement ifStat = 
      new IfStatement(new NodeToken("if"),
		      n.f1, 
		      aexp.toAbstractExpression().toExpression(),
		      n.f3, 
		      n.f4,
		      new NodeOptional()); */

    Expression empty = new Expression(null, new NodeOptional());
    ExpressionManager.register(empty, aexp);        

    IfStatement ifStat = 
      new IfStatement(new NodeToken("if"),
                      n.f1, empty,
                      //aexp.toExpression(),
                      n.f3, 
                      n.f4,
                      new NodeOptional());

    Statement stat = new Statement(new NodeChoice(ifStat, 5));
    BlockStatement blockStat = new BlockStatement(new NodeChoice(stat,1));
    
    optional.addNode(blockStat);

    // aexp = new IdentifierDescriptor(Mixer.group());
    // aexp = aexp.append(new IdentifierDescriptor(new NodeToken("no")));
    // aexp = aexp.append(new ArgumentsDescriptor());

    // aexp = new IdentifierNode(Mixer.group());
    // aexp = aexp.append(new IdentifierNode(new NodeToken("no")));
    // aexp = aexp.append(new ArgumentsNode());

	aexp = new MethodCallNode(
		new IdentifierNode(Mixer.group(id))
		, new IdentifierNode(new NodeToken("no"))
		, new ArgumentsNode() );

    NodeSequence sequence = new NodeSequence(2);
    // sequence.addNode(aexp.toAbstractExpression().toStatementExpression());

    StatementExpression empty2 = new StatementExpression(new NodeChoice(null));
    ExpressionManager.register(empty2, aexp);        

    sequence.addNode(empty2);
    sequence.addNode(new NodeToken(";"));

    stat = new Statement(new NodeChoice(sequence,3));
    blockStat = new BlockStatement(new NodeChoice(stat,1));

    optional.addNode(blockStat);

    Block block = new Block ( new NodeToken("{"),
                              optional,
                              new NodeToken("}") );
    //block.show();

    //n.getParent(4).insertAfter(block);
    //n.getParent(4).destroy();
    stat = new Statement(new NodeChoice(block, 1));
    n.getParent(2).replaceWith(stat);
  }

  private BlockStatement rangeDec(NodeToken id, Expression range) {
    TypeDescriptor type = 
      new TypeDescriptor(new NodeToken("Range"));

    VariableDeclaratorDescriptor var = 
      new VariableDeclaratorDescriptor(new VariableDeclaratorId(Mixer.range(id),
								new NodeListOptional()));

    var.setInitializer(range);

    LocalVariableDeclarationDescriptor dec = 
      new LocalVariableDeclarationDescriptor(type.leaf, var.leaf);


    NodeSequence sequence = new NodeSequence(2);
    sequence.addNode(dec.leaf);
    sequence.addNode(new NodeToken(";"));

    return  new BlockStatement(new NodeChoice(sequence, 0));
  }

  private NamedLocation locAt(NodeToken id) {
    // AbstractPrimaryExpression rng =
    //  new IdentifierDescriptor(Mixer.range(id));
    // AbstractPrimaryExpression rng =
    //  new IdentifierNode(Mixer.range(id));
    AbstractExpression rng =
      new IdentifierNode(Mixer.range(id));

    // IdentifierDescriptor loc = 
    //  new IdentifierDescriptor (new NodeToken("location"));
    IdentifierNode loc = 
      new IdentifierNode(new NodeToken("location"));

    // rng = rng.append(loc);

    // IdentifierDescriptor idx =
    //  new IdentifierDescriptor(Mixer.index(id));
    // IdentifierNode idx =
    //  new IdentifierNode(Mixer.index(id));
    IdentifierNode idx = new IdentifierNode(id);

    // ArgumentsDescriptor arg = 
    //  new ArgumentsDescriptor(idx);
    ArgumentsNode arg = new ArgumentsNode(idx);
    
    /*
    TripletOrExpressionListDescriptor idxExp =
      new TripletOrExpressionListDescriptor(new TripletOrExpressionList
                                            (new TripletOrExpression
                                             (new NodeChoice(idx.toExpression(), 1)),
                                             new NodeListOptional()));
    */

    // rng = rng.append(arg);

	MethodCallNode mthd = new MethodCallNode(rng, loc, arg);

    //return new NamedLocation(id, new NodeToken("="), rng.toPrimaryExpression(), new NodeOptional());
	PrimaryExpression empty = new PrimaryExpression(null, new NodeListOptional());
	ExpressionManager.register(empty, mthd);
    // return new NamedLocation(id, new NodeToken("="), empty, new NodeOptional());
    return new NamedLocation(id, new NodeToken("="), empty);
    // Changed to new NamedLocation Definition when Hanku 
    // changed the definition of NamedLocation and add NamedLocationRange
  }

  private NodeOptional forInit(NodeToken id) {
    TypeDescriptor type = 
      new TypeDescriptor("int");

    // VariableDeclaratorDescriptor var = 
    //  new VariableDeclaratorDescriptor
    // (new VariableDeclaratorId(Mixer.index(id), new NodeListOptional()));

    VariableDeclaratorDescriptor var =
      new VariableDeclaratorDescriptor
      (new VariableDeclaratorId(id, new NodeListOptional()));

    Literal literal = new Literal( new NodeChoice( new NodeToken("0"), 0));

    // LiteralDescriptor aexp = new LiteralDescriptor(literal);
    LiteralNode aexp = new LiteralNode(literal);

    Expression empty = new Expression(null, new NodeOptional());
    ExpressionManager.register(empty, aexp);        

    //var.setInitializer(aexp.toExpression());
    var.setInitializer(empty);

    LocalVariableDeclarationDescriptor dec = 
      new LocalVariableDeclarationDescriptor(type.leaf, var.leaf);

    NodeOptional optional = new NodeOptional();

    optional.addNode( new ForInit(new NodeChoice(dec.leaf, 0)) );

    return optional;
  }

 private NodeOptional forInit(NodeToken id , Expression initValExp) {
    TypeDescriptor type = 
      new TypeDescriptor("int");

    // VariableDeclaratorDescriptor var = 
    //  new VariableDeclaratorDescriptor
    // (new VariableDeclaratorId(Mixer.index(id), new NodeListOptional()));

    VariableDeclaratorDescriptor var =
      new VariableDeclaratorDescriptor
      (new VariableDeclaratorId(id, new NodeListOptional()));

    var.setInitializer(initValExp);

    LocalVariableDeclarationDescriptor dec = 
      new LocalVariableDeclarationDescriptor(type.leaf, var.leaf);

    NodeOptional optional = new NodeOptional();

    optional.addNode( new ForInit(new NodeChoice(dec.leaf, 0)) );

    return optional;
  }

  private NodeOptional forExpression(NodeToken id) {
    // AbstractPrimaryExpression idx = 
    //  new IdentifierDescriptor(Mixer.index(id));
    // AbstractPrimaryExpression idx = 
    //  new IdentifierNode(Mixer.index(id));
    // AbstractExpression idx = 
    //  new IdentifierNode(Mixer.index(id));
    AbstractExpression idx = new IdentifierNode(id);

    // AbstractPrimaryExpression rng = 
    //  new IdentifierDescriptor(Mixer.range(id));
    // AbstractPrimaryExpression rng = 
    //  new IdentifierNode(Mixer.range(id));
    AbstractExpression rng = 
      new IdentifierNode(Mixer.range(id));

    // IdentifierDescriptor size 
	// = new IdentifierDescriptor(new NodeToken("size"));
    // IdentifierNode size = new IdentifierNode(new NodeToken("size"));
    // rng = rng.append(size);
    IdentifierNode size = new IdentifierNode(new NodeToken("size"));

    // ArgumentsDescriptor arg = new ArgumentsDescriptor();
    // ArgumentsNode arg = new ArgumentsNode();
    // rng = rng.append(arg);
	ArgumentsNode arg = new ArgumentsNode();

	MethodCallNode rng1 = new MethodCallNode(rng, size, arg);

    // AbstractExpression condition = idx.toAbstractExpression();

    // condition = condition.less(rng.toAbstractExpression());    
    //AbstractExpression condition = new BinExprNode( 9, idx, rng1 );
    
    // Modified by Han-Ku because, in Overall Statement, the range
    // has to be like i <= N-2 ...
    AbstractExpression condition = new BinExprNode( 11, idx, rng1 );    

    Expression empty = new Expression(null, new NodeOptional());
    ExpressionManager.register(empty, condition);        

    NodeOptional optional = new NodeOptional();
    optional.addNode(empty);

    return optional;
  }

 private NodeOptional forExpression(NodeToken id, Expression upperBound) {
    // AbstractPrimaryExpression idx = 
    //  new IdentifierDescriptor(Mixer.index(id));
    // AbstractPrimaryExpression idx = new IdentifierNode(Mixer.index(id));
    // AbstractExpression idx = new IdentifierNode(Mixer.index(id));
    AbstractExpression idx = new IdentifierNode(id);

    /* AbstractExpression condition = idx.toAbstractExpression();

    condition = condition.less(upperBound.toAbstractExpression());

    NodeOptional optional = new NodeOptional();
	optional.addNode(condition.toExpression()); */

	/* ShiftExpression shftExpr0 = idx.toAbstractExpression().toExpression();
	NodeListOptional lessThan = new NodeListOptional(new NodeToken("<="));
	ShiftExpression shftExpr1 = upperBound.f0;
	lessThan.addNode(shftExpr1);
	RelationalExpression relatExp = new RelationalExpression(shftExpr0, lessThan);
	InstanceOfExpression instOfExpr = new InstanceOfExpression(relatExp,null);
	EqualityExpression eqExp = new EqualityExpression(instOfExpr, null);
	AndExpression andExpr = new AndExpression(eqExp, null);
	ExclusiveOrExpression exclOrExp = new ExclusiveOrExpression(andExpr, null);
	InclusiveOrExpression inclOrExp = new InclusiveOrExpression(exclOrExp, null);
	ConditionalAndExpression condAndExp = new ConditionalAndExpression(inclOrExp,null); 
	ConditionalOrExpression condOrExp = new ConditionalOrExpression(condAndExp,null);
	ConditionalExpression condExpression = new ConditionalExpression(condOrExp,null);
	Expression	expr = new Expression(condExpression, null); */

	// AbstractExpression condition = idx.toAbstractExpression();

    // condition = condition.less(ExpressionManager.lookup(upperBound));

	//AbstractExpression condition = new BinExprNode(9, idx, ExpressionManager.lookup(upperBound));
    // Modified by Han-Ku because, in Overall Statement, the range
    // has to be like i <= N-2 ...
    AbstractExpression condition = new BinExprNode(11, idx, ExpressionManager.lookup(upperBound));    

    Expression empty = new Expression(null, new NodeOptional());
    ExpressionManager.register(empty, condition);        

    NodeOptional optional = new NodeOptional();
    optional.addNode(empty);

    return optional;
  }

  private NodeOptional forUpdate(NodeToken id) {
    // AbstractPrimaryExpression idx = 
    //  new IdentifierDescriptor(Mixer.index(id));
    // AbstractPrimaryExpression idx = new IdentifierNode(Mixer.index(id));

    // AbstractExpression aexp = idx.toAbstractExpression();

    // AbstractExpression idx = new IdentifierNode(Mixer.index(id));
    AbstractExpression idx = new IdentifierNode(id);
    
    // aexp = aexp.postfixAddition();
	AbstractExpression aexp = new UnaryExprNode(29, idx);

    StatementExpression empty = new StatementExpression(new NodeChoice(null));
    ExpressionManager.register(empty, aexp);        

    StatementExpressionList list = 
      new StatementExpressionList(empty,
                                  new NodeListOptional());   

    ForUpdate update = new ForUpdate(list);

    NodeOptional optional = new NodeOptional();
    optional.addNode(update);

    return optional;

  }

 private NodeOptional forUpdate(NodeToken id, Expression updtExp) {
    // AbstractPrimaryExpression idx = 
    //  new IdentifierDescriptor(Mixer.index(id));
    // AbstractPrimaryExpression idx = new IdentifierNode(Mixer.index(id));
    // AbstractExpression idx = new IdentifierNode(Mixer.index(id));
    AbstractExpression idx = new IdentifierNode(id);

    // AbstractExpression aexp = idx.toAbstractExpression();

	AbstractExpression aexp1 = ExpressionManager.lookup(updtExp);

    AbstractExpression aexp = new AdditionAssignmentNode(idx, aexp1);

    StatementExpression empty = new StatementExpression(new NodeChoice(null));
    ExpressionManager.register(empty, aexp);        

    StatementExpressionList list = 
      new StatementExpressionList(empty,
                                  new NodeListOptional());

    ForUpdate update = new ForUpdate(list);

    NodeOptional optional = new NodeOptional();
    optional.addNode(update);

    return optional;

  }

  // This method is modified by Qiang Zheng in order to fit the new grammar
  // of Overall statement.
  // Syntax of Overall Statement:
  // overall:  "overall" "(" NamedLocationList() ")" Statement()
  // NamedLocationList: NamedLocation() ( "," NamedLocation() )*
  // NamedLocation : <IDENTIFIER> "=" PrimaryExpression() [ "|" Triplet() ]

  // The strategy here is to translate "overall" statement
  // into "at" statement inside "for" loops.

  public void visit (OverallStatement n) {
    NodeListOptional optional = new NodeListOptional();
    NodeListOptional lstOptional = new NodeListOptional();

    NamedLocationListDescriptor locList = 
      new NamedLocationListDescriptor(new NamedLocationList(null, new NodeListOptional())); 

	// Constructing the at statement
    AtStatement atStatement = new AtStatement (new NodeToken("at"),
                                               new NodeToken("("),
                                               locList.leaf,
                                               new NodeToken(")"),
                                               n.f4);
    /* Changed this statement after Hanku modified the definition
     * of NamedLocation, NamedLocationList, NamedLocationListDescriptor
     * and newly add the definition of NamedLocationRange,
     * NamedLocationRangeList, NamedLocationRangeListDescriptor */

    /* AtStatement atStatement = new AtStatement (new NodeToken("at"),
                               new NodeToken("("),
                               new NamedLocationList(locRngList.leaf),
                               new NodeToken(")"),
                               n.f4); */

    Statement statement = new Statement (new NodeChoice(atStatement, 16));
	BlockStatement blckStmt = new BlockStatement(new NodeChoice(statement,1));
	lstOptional.addNode(blckStmt);
	Block		blck = new Block(new NodeToken("{")
								, lstOptional
								, new NodeToken("}"));
	Statement	stmt = new Statement(new NodeChoice(blck,1));

    ForStatement forStatement = null;

    /* Changed this statement after Hanku modified the definition
     * of NamedLocation, NamedLocationList, NamedLocationListDescriptor
     * and newly add the definition of NamedLocationRange,
     * NamedLocationRangeList, NamedLocationRangeListDescriptor
     * By Qiang
     */

    /* NamedLocationListDescriptor descriptor =
      new NamedLocationListDescriptor (n.f2);
     */

    NamedLocationRangeListDescriptor descriptor =
      new NamedLocationRangeListDescriptor (n.f2);

    for (int i = 0; i<descriptor.numberOfLocs(); i++) {
      NodeToken id = descriptor.name(i);
      PrimaryExpression range = descriptor.base(i);

      // NamedLocation ithLoc = descriptor.loc(i);
      NamedLocationRange ithLocRng = descriptor.loc(i);
      NamedLocation ithLoc = new NamedLocation( ithLocRng );

      // AbstractPrimaryExpression apExp = PrimaryExpressionManager.lookup(range);
      AbstractExpression apExp = ExpressionManager.lookup(range);

      Expression empty = new Expression(null, new NodeOptional());
      ExpressionManager.register(empty, apExp);        

      optional.addNode(rangeDec(id, empty));

      locList.addLoc(locAt(id));

	  // This is revised by Qiang Zheng

	  // If the statement is without triplet...
	  // if (!ithLoc.f3.present()) {
      /*  forStatement = new ForStatement (new NodeToken("for"),
                                         new NodeToken("("),
                                         forInit(id),
                                         new NodeToken(";"),
                                         forExpression(id),
                                         new NodeToken(";"),
                                         forUpdate(id),
                                         new NodeToken(")"),
                                         stmt);
       */
      // } else {
	    // If the overall statement is with triplet...
        // NodeSequence namedLocOptSeq = (NodeSequence) ithLoc.f3.node;
		// Triplet	namedLocTrip = (Triplet) namedLocOptSeq.elementAt(1);
		Triplet	namedLocTrip = ithLocRng.f4;

		java.util.Vector childrenVec = namedLocTrip.getChildren();

		NodeOptional initOpt, upperValOpt, updateOpt;

		NodeOptional ndOpt = (NodeOptional) childrenVec.elementAt(0);
		if ( ndOpt.present() ) {
			initOpt = forInit( id,(Expression) ndOpt.node );
		} else {
			initOpt = forInit(id);
		}

		ndOpt = (NodeOptional) childrenVec.elementAt(2);
		//Expression upperValExp = (Expression) childrenVec.elementAt(2).node;
		if ( ndOpt.present() ) {
			upperValOpt = forExpression( id, (Expression) ndOpt.node );
		} else {
			upperValOpt = forExpression( id );
		}

		ndOpt = (NodeOptional) childrenVec.elementAt(3);
		//NodeSequence updtSeqns = (NodeSequence) childrenVec.elementAt(3);
		NodeSequence updtSeqns = (NodeSequence) ndOpt.node;
        if ( ndOpt.present() ) {
            updateOpt = forUpdate( id, (Expression) updtSeqns.elementAt(1));
        } else {
            updateOpt = forUpdate( id );
        }


        forStatement = new ForStatement (new NodeToken("for"),
                                         new NodeToken("("),
										 initOpt,
                                         new NodeToken(";"),
										 upperValOpt,
                                         new NodeToken(";"),
										 updateOpt,
                                         new NodeToken(")"),
                                         stmt);
      /* } */

      statement = new Statement (new NodeChoice(forStatement, 8));
    }

    BlockStatement blockStatement = 
      new BlockStatement( new NodeChoice(statement, 1));

    optional.addNode(blockStatement);

    Block block = new Block ( new NodeToken("{"),
                              optional,
                              new NodeToken("}") );

    //block.show();

    // n.getParent(4).insertAfter(block);
    // n.getParent(4).destroy();
	n.replaceWith(block);
	visit(atStatement);
  }

}

    /*
    System.err.println(n.parent.toString());
    System.err.println(n.getParent(1).toString());
    System.err.println(n.parent.parent.toString());
    System.err.println(n.getParent(2).toString());
    System.err.println(n.parent.parent.parent.toString());
    System.err.println(n.getParent(3).toString());
    System.err.println(n.parent.parent.parent.parent.toString());
    System.err.println(n.getParent(4).toString());
    System.err.println(n.parent.parent.parent.parent.parent.toString());
    System.err.println(n.getParent(5).toString());
    System.err.println(n.parent.parent.parent.parent.parent.parent.toString());
    System.err.println(n.getParent(6).toString());
    */
	      


/* In fact, we don't need visitor here. */
