/*
 * File         : DeclarationTranslator.java
 * Author       : Guansong Zhang
 * Created      : Tue Oct 27 11:53:19 1998
 * Revision     : $Revision: 1.8 $
 * Updated      : $Date: 1999/11/11 01:17:45 $
 */



package hpjava.compiler.translator;

import hpjava.compiler.util.*;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

import hpjava.compiler.unparser.*;

import hpjava.compiler.reflection.*;
import hpjava.compiler.abstractexp.*;
import hpjava.compiler.wrapper.*;

public class DeclarationTranslator {

  Node target;

  public DeclarationTranslator(FieldDeclaration n) {
    target = n;    
  }

  public DeclarationTranslator(MethodDeclaration n) {
    target = n;    
  }

  public DeclarationTranslator(MethodDeclarator n) {
    target = n;    
  }

  public DeclarationTranslator(LocalVariableDeclaration n) {
    target = n;    
  }

  public void process() {
    DeclarationProcessor processor = new DeclarationProcessor();
    target.accept(processor);
  }

}

class DeclarationProcessor extends DepthFirstVisitor {
  
  static final int FORFIELD = 0;
  static final int FORMETHOD = 1;
  static final int FORLOCALVAR = 2;

  IntStack status = new IntStack();

  DeclarationProcessor() {
  }

  public void visit (FieldDeclaration n) {
    status.push(FORFIELD);
    super.visit(n);
    status.pop();
  }

  public void visit (MethodDeclaration n) {
    status.push(FORMETHOD);
    super.visit(n);
    status.pop();
  }

  // Following method is added by Qiang Zheng
  // in order to insert `MPI initialization' statement
  // to main method
  public void visit (MethodDeclarator n) {
    super.visit(n);
  }

  public void visit (LocalVariableDeclaration n) {
    status.push(FORLOCALVAR);

    LocalVariableDeclarationDescriptor descriptor
      = new LocalVariableDeclarationDescriptor(n);

    ClassReflectorDescriptor crd;

    for (int i = descriptor.numberOfVars()-1; i >= 0; i--) {
      VariableDeclarator varDeclarator = descriptor.var(i);
      
      NodeToken var = varDeclarator.f0.f0;
      
      try {
	crd =LocalVariableManager.classReflectorDescriptorOf(var);
      }
      catch (NoSuchNameException e) {
	throw new Error("This should be a local var.");
      }

      if (crd.isDistArray()) {
	Type type = classTypeOf(crd);

	if (i==0) {
	  if (descriptor.numberOfVars() != 1) {
	    LocalVariableDeclaration declaration = 
	      new LocalVariableDeclaration(n.f0, 
					   type, 
					   varDeclarator, 
					   new NodeListOptional());

	    // insertAfter.
	    declaration.accept(new PrettyPrinter(System.err));
	    System.err.println();
	    System.err.println(n.getParent(4).toString());
	  }
	  else {
	    VariableDeclaratorId vdi = 
	      new VariableDeclaratorId(varDeclarator.f0.f0,
				  new NodeListOptional());

	    varDeclarator.f0 = vdi;
	    vdi.setParent(varDeclarator);

	    n.f1 = type;
	    type.setParent(n);
	  }
	}
	else {
	  throw new Error("Not finished.");
	}
      }
    }
    status.pop();
  }

  public void visit (FormalParameter n) {
    NodeToken var = n.f2.f0;

    try {
      ClassReflectorDescriptor crd =
	LocalVariableManager.classReflectorDescriptorOf(var);
      
      if (crd.isDistArray()) {
	Type type = classTypeOf(crd);

	// n.f1.replaceWith(type); 
	// it is complicated to implement this function.
	// first, a virtual function Node sonAs(Node n) has to be implemented
	// to find the reference in parent.

	n.f1=type;
	type.setParent(n);
      }
    }
    catch (NoSuchNameException e) {
      throw new Error("Formal parameter should be a local var.");
    }
  }

  private Type classTypeOf(ClassReflectorDescriptor crd) {
    if (crd.isDistArray()) {
      String cname = crd.getName();
	
      // this is not quite good, change to id concatenation.

      NodeToken id  = new NodeToken(cname);
      Name name = new Name(id, new NodeListOptional());
      NodeChoice choice = new NodeChoice(name, 1);
      Type type = new Type(choice, new NodeListOptional());
      return type;
    }
    else {
      throw new Error("This is used for dist array.");
    }
  }
}



