/*
 * File         : Main.java
 * Author       : Guansong Zhang
 * Created      : Thu Oct  8 13:00:47 1998
 * Revision     : $Revision: 1.7 $
 * Updated      : $Date: 1999/11/11 01:17:08 $
 */

package hpjava.compiler.translator;

import java.io.*;
import java.util.*;

import hpjava.compiler.util.*;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

import hpjava.compiler.reflection.*;
import hpjava.compiler.abstractexp.*;

public class Main {

  static Runtime runtime = Runtime.getRuntime();

  static CMDline cmd;
  static Project prj;
  
  static ScopeManager scopeMgr;

  static NameSpaceManager nsMgr;
  
  static ClassReflectorManager clsMgr;
  static LocalVariableManager symbMgr;

  static ExpressionManager expMgr;
  // static PrimaryExpressionManager primExpMgr;
  
  static void usage() {
    System.out.println("hpj2j usage ...");
  }

  public static void main(String[] argv) {
    if (argv.length==0) {
      usage();
      runtime.exit(1);
    } 
    else {       
      try {
        cmd = new CMDline(argv);

        prj= new Project (cmd);
        
        scopeMgr = new ScopeManager (prj);

        nsMgr = new NameSpaceManager (prj);

        clsMgr = new ClassReflectorManager (prj);
        
        symbMgr = new LocalVariableManager (prj);

        expMgr = new ExpressionManager(prj);

	// primExpMgr = new PrimaryExpressionManager(prj);

        // primExpMgr.typeChecking();
        expMgr.typeChecking();

        if (CompilingError.number() !=0 )
          throw new NotCompleteException ("hpj2j translator:", CompilingError.number());

	// begin translation, only deal with named distarray.
        
        if (!prj.isVerbose()){
          Translator translator = new Translator(prj);
          translator.process();
        };
        
        prj.output();
      }
      catch(InvalidCMDflagException e) {
        System.err.println("Flag: "+e.getMessage()+" invalid.");
        System.err.println();
        usage();
        runtime.exit(1);
      }
      catch(InvalidFileNameException e) {
        System.err.println("Filename: "+e.getMessage()+" invalid.");
        System.err.println();
        usage();
        runtime.exit(1);
      }
      catch(MissingFileException e) {
        System.err.println("Missing input file.");
        System.err.println();
        usage();
        runtime.exit(1);
      }
      catch (FileNotFoundException e) {
        System.err.println("File: "+e.getMessage()+" not found.");
        runtime.exit(1);
      }
      catch (NotCompleteException e) {
        System.err.println();
        System.err.println(e.getMessage());
        runtime.exit(1);
      }
      
    }
  }
}                                                                              

/* 
   // code for testing memory
   Runtime rt = Runtime.getRuntime();
   rt.gc();
   System.out.println("Total Memory in VM: " +
   Long.toString(rt.totalMemory()));
   System.out.println("Total Free Memory:  " +
   Long.toString(rt.freeMemory()));
   System.in.read();
*/

