/*
 * File         : Translator.java
 * Author       : Guansong Zhang
 * Created      : Thu May 14 12:03:02 1998
 * Revision     : $Revision: 1.23 $
 * Updated      : $Date: 1999/12/02 20:38:47 $
 */

package hpjava.compiler.translator;

import java.io.*;
import java.util.*;

import hpjava.compiler.util.*;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

import hpjava.compiler.reflection.*;
import hpjava.compiler.abstractexp.*;

import hpjava.compiler.unparser.*;

// Following line is added by Qiang Zheng after
// the addition of wrapper package 
import hpjava.compiler.wrapper.*;

public class Translator {
  static public Node current;

  private Project prj;

  public Translator (Project prj) {
    this.prj=prj;
  }

  public void process() {
    Processor processor = new Processor();
    processor.visit(prj);
  }
}                                                                              

class Processor extends DepthFirstVisitor {
  LinkedList blockStatementList = new LinkedList();
  
  // prefix for group like tmp_1_grp_hpj
  Prefix pre = new Prefix();

  public Processor() {

  }

  public void visit(Project prj) {
    
    Enumeration enum = prj.rootFileTable.keys();
    
  LOOP: while (enum.hasMoreElements()) {
      
      Node root = (Node) enum.nextElement();      

      root.accept(this);
      
    }

  }

  private ImportDeclaration importDec(String pkg) {
    StringTokenizer tokenizer = 
      new StringTokenizer(pkg, ".");

    NameDescriptor name = new NameDescriptor();
    
    while (tokenizer.hasMoreTokens()) {
      name.addID(new NodeToken(tokenizer.nextToken()));
    }

    NodeSequence sequence = new NodeSequence(2);
    sequence.addNode(new NodeToken("."));
    sequence.addNode(new NodeToken("*"));

    NodeOptional optional = new NodeOptional();
    optional.addNode(sequence);

    ImportDeclaration result =
      new ImportDeclaration(new NodeToken("import"),
                            name.leaf,
                            optional,
                            new NodeToken(";"));

    return result;
  }

  public void visit(CompilationUnit n) {
    n.f1.addNode(importDec("hpjava.lang"));
    n.f1.addNode(importDec("hpjava.spmd"));

    super.visit(n);
  } 

  /* private LocalVariableDeclaration mpiInitStmt( NodeToken paraToken ) {
    NodeToken nToken = Mixer.env();
    Name typeName = new Name( new NodeToken("hpjava.spmd.SpmdEnv"), 
                              new NodeListOptional() );
    Type mpiType = new Type( new NodeChoice( typeName ,1 ),
                             new NodeListOptional() );

    // constructing right subtree of the initializing mpi environment
    ClassReflector cr;
    try {
      cr = ClassReflector.forName( "hpjava.spmd.SpmdEnv" );
    } catch (ClassNotFoundException e) {
      throw new Error("SpmdEnv Class can not be found");
    }

    AllocationExpressionNode mpiInitMethodNode =
      new AllocationExpressionNode( cr );

    IdentifierNode arguId = new IdentifierNode(paraToken);
    ArgumentList argsLst = new ArgumentList( arguId.toExpression(),
                                   new NodeListOptional() );
    Arguments args = new Arguments( new NodeToken("("),
                                    new NodeOptional( argsLst ),
                                    new NodeToken(")"));

    ArgumentsDescriptor argd = new ArgumentsDescriptor( args );

    mpiInitMethodNode.descriptor.addArguments( argd );

    VariableInitializer mpiInitMethod
      = new VariableInitializer(new NodeChoice(
              mpiInitMethodNode.toExpression(),1));

    VariableDeclaratorId mpiVar = new VariableDeclaratorId (
                                    nToken,
                                    new NodeListOptional() );

    VariableDeclarator mpiVarDcl = new VariableDeclarator(
                                  mpiVar,
                                  new NodeOptional( mpiInitMethod ) );

    LocalVariableDeclaration result = 
      new LocalVariableDeclaration(new NodeOptional(),
                                   mpiType,
                                   mpiVarDcl,
                                   new NodeListOptional() );
    ScopeManager.add(nToken);
    try {
      LocalVariableManager.add(nToken, Class.forName("hpjava.spmd.SpmdEnv"));
    } catch (ClassNotFoundException e) {
      throw new Error("SpmdEnv Class can not be found");
    }
    return result;
  } */

  // Following is added by Qiang Zheng to insert
  // a statement to initialize MPI environment
  public void visit(MethodDeclaration n) {

    /* if ( n.f2.f0.tokenImage.equals(new String("main") ) ) {
      Flags.setInMainFlag();
      // In a `main' method 
      if ( !Flags.isInit() ) {
        Flags.setInitFlag();
        FormalParameters para = n.f2.f1;
        if ( para.f1.present() ) {
          NodeSequence paraSeq = (NodeSequence) para.f1.node;
          VariableDeclaratorId para0
            = ((FormalParameter)paraSeq.elementAt(0)).f2;
          ((Block)n.f4.choice).f1.insertNode( mpiInitStmt( para0.f0 ) );
        } else {
          System.out.println("Must specify an argument in main method");
        }
      }
    } */

    DeclarationTranslator translator 
      = new DeclarationTranslator(n);

    translator.process();

    super.visit(n);

    // Flags.resetInMainFlag();
  }

  public void visit(FieldDeclaration n) {
    DeclarationTranslator translator 
      = new DeclarationTranslator(n);

    translator.process();

    super.visit(n);
  }

  public void visit(LocalVariableDeclaration n) {
    DeclarationTranslator translator 
      = new DeclarationTranslator(n);

    translator.process();

    super.visit(n);
  }


  public void visit(Expression n) {
    AbstractExpression exp 
      = ExpressionManager.lookup(n);

    if (exp == null)
      n.getParent(1).accept( new ITreePrinter());
    
    AbstractExpressionTranslator translator = 
      new AbstractExpressionTranslator(exp, blockStatementList, pre);
    
    translator.process();

  }


  public void visit(StatementExpression n) {
    AbstractExpression exp 
      = ExpressionManager.lookup(n);

    if (exp == null)
      n.getParent(1).accept( new ITreePrinter());

    AbstractExpressionTranslator translator = 
      new AbstractExpressionTranslator(exp, blockStatementList, pre);
    
    translator.process();

  }

  public void visit(GroupExpression n) {
    AbstractExpression exp 
      = ExpressionManager.lookup(n);

    if (exp == null)
      n.getParent(1).accept( new ITreePrinter());

    AbstractExpressionTranslator translator = 
      new AbstractExpressionTranslator(exp, blockStatementList, pre);
    
    translator.process();

  }

  public void visit (OnStatement n) { 
    ConstructTranslator translator = 
      new ConstructTranslator (n, pre); 

    translator.process();

    super.visit(n);
  }

  public void visit (AtStatement n) { 
    ConstructTranslator translator = 
      new ConstructTranslator (n); 

    translator.process();

    super.visit(n);
  }

  public void visit (OverallStatement n) { 
    ConstructTranslator translator = 
      new ConstructTranslator (n); 

    translator.process();

    super.visit(n);
  }

  // Override the method visit()
   public void visit( NodeList n ) {
	int i = 0;
	while(  i < n.size() ) {
		((Node)n.elementAt(i)).accept(this);
		i++;
	}	
      // for ( Enumeration e = n.elements(); e.hasMoreElements(); )
      //   ((Node)e.nextElement()).accept(this);
   }

   public void visit(NodeListOptional n) {
	int i = 0;

      if ( n.present() )
		while(i < n.size() ) {
			((Node)n.elementAt(i)).accept(this);
			i++;
		}	
         // for ( Enumeration e = n.elements(); e.hasMoreElements(); )
         //   ((Node)e.nextElement()).accept(this);
   }

  // Created by Han-Ku
  // to generate ArraySectNode
  public void visit(BlockStatement n) {
    n.f0.accept(this);

    //System.out.println("blockStatementList.size() = " + blockStatementList.size());

    if(blockStatementList.size() != 0){
      //blockStatementList.removeLast();
      
      //System.out.println("n.getParent(1): " + n.getParent(1).getClass().getName());
      //System.out.println("n.getParent(2): " + n.getParent(2).getClass().getName());    
 
      NodeListOptional optional = new NodeListOptional();
      Statement stat;
      Block blockStat;

      while(blockStatementList.size() != 0){
        blockStat = (Block)blockStatementList.getLast();
        optional.addNode(blockStat);
        blockStatementList.removeLast();
      }
      
      BlockStatement blockStat2 = new BlockStatement(n.f0);
      optional.addNode(blockStat2);
      Block block = new Block(new NodeToken(""), optional, new NodeToken(""));
    
      //blockStat.show();
      //block.show();

      //n.insertBefore(blockStat);
      //n.insertAfter(blockStat);
      n.replaceWith(block);

      //blockStatementList.removeFirst();
      
    }
    
  }
  
  
}


