package hpjava.compiler.unparser;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.abstractexp.*;
import hpjava.compiler.util.*;

import java.util.*;
import java.io.*;

public class AbstractExpressionPrettyPrinter 
  extends AbstractExpressionVisitor {

  final int PREFIX = 0;
  final int SUFFIX = 1;

  protected PrintWriter out;
  protected Space spc = new Space(3); 

  protected IPrettyPrinter ipp;  
  protected NodeToken previous = null;  

  private IntStack status=new IntStack(); {
    // for simple expression with out concatenation, it is a prefix.
    status.push(PREFIX);
  }

  public AbstractExpressionPrettyPrinter(IPrettyPrinter ipp){
    this.ipp = ipp;
    this.out = ipp.out;
    this.spc = ipp.spc;
  }

  public AbstractExpressionPrettyPrinter() { 
    this.out = new PrintWriter(System.out, true); 
  }

  public AbstractExpressionPrettyPrinter(Writer o) { 
    this.out = new PrintWriter(o, true); 
  }

  public AbstractExpressionPrettyPrinter(OutputStream o) { 
    this.out = new PrintWriter(o, true); 
  }

  public void flushWriter(){ out.flush(); }


  // write your core from here.

  public void visit(TypeNode n) {

    flushWriter();
    n.leaf.accept(ipp);

  }


   public void visit(AssignmentNode n) {
      String image = "=";

      n.left.accept(this);
      out.print(" " + image + " ");
      n.right.accept(this);
   }

   public void visit(AdditionAssignmentNode n) {
      String image = "+=";

      n.left.accept(this);
      out.print(" " + image + " ");
      n.right.accept(this);
   }

   public void visit(SubtractionAssignmentNode n) {
      String image = "-=";

      n.left.accept(this);
      out.print(" " + image + " ");
      n.right.accept(this);
   }

   public void visit(MultiplicationAssignmentNode n) {
      String image = "*=";

      n.left.accept(this);
      out.print(" " + image + " ");
      n.right.accept(this);
   }

   public void visit(DivisionAssignmentNode n) {
      String image = "/=";

      n.left.accept(this);
      out.print(" " + image + " ");
      n.right.accept(this);
   }

   public void visit(AndAssignmentNode n) {
      String image = "&=";

      n.left.accept(this);
      out.print(" " + image + " ");
      n.right.accept(this);
   }

   public void visit(InclusiveOrAssignmentNode n) {
      String image = "|=";

      n.left.accept(this);
      out.print(" " + image + " ");
      n.right.accept(this);
   }

   public void visit(ExclusiveOrAssignmentNode n) {
      String image = "^=";

      n.left.accept(this);
      out.print(" " + image + " ");
      n.right.accept(this);
   }

   public void visit(LeftShiftAssignmentNode n) {
      String image = "<<=";

      n.left.accept(this);
      out.print(" " + image + " ");
      n.right.accept(this);
   }

   public void visit(RightShiftAssignmentNode n) {
      String image = ">>=";

      n.left.accept(this);
      out.print(" " + image + " ");
      n.right.accept(this);
   }

   public void visit(ModulusAssignmentNode n) {
      String image = "%=";

      n.left.accept(this);
      out.print(" " + image + " ");
      n.right.accept(this);
   }

   public void visit(ExtensionRightShiftAssignmentNode n) {
      String image = ">>>=";

      n.left.accept(this);
      out.print(" " + image + " ");
      n.right.accept(this);
   }

   //
   // f0 -> ConditionalOrExpression()
   // f1 -> [ "?" Expression() ":" ConditionalExpression() ]
   //
   public void visit(ConditionalNode n) {
      String image1 = "?";
      String image2 = ":";

      n.condition.accept(this);

      if(n.trueResult != null || n.falseResult != null){
        out.print(" " + image1 + " ");
        n.trueResult.accept(this);

        out.print(" " + image2 + " ");
        n.falseResult.accept(this);
      }
   }

   //
   // f0 -> "(" Type() ")" UnaryExpression()
   //       | "(" Type() ")" UnaryExpressionNotPlusMinus()
   //
   public void visit(CastNode n) {
      String image1 = "(";
      String image2 = ")";

      out.print(image1);
      n.type.accept(this);

      out.print(image2);
      n.leaf.accept(this);
   }

   //
   // f0 -> Operator
   // f1 -> AbstractExpression()
   // f2 -> AbstractExpression()
   //
   public void visit(BinExprNode n) {
      String image;
      switch(n.binaryOperator){
        case n.CONDITIONAL_OR:
          image = "||";
          break;
        case n.CONDITIONAL_AND:
          image = "&&";
          break;
        case n.INCLUSIVE_OR:
          image = "|";
          break;
        case n.EXCLUSIVE_OR:
          image = "^";
          break;
        case n.AND:
          image = "&";
          break;
        case n.EQUALITY:
          image = "==";
          break;
        case n.INEQUALITY:
          image = "!=";
          break;
        case n.INSTANCE_OF:
          image = "instanceof ";
          break;
        case n.GREATER:
          image = ">";
          break;
        case n.LESS:
          image = "<";
          break;
        case n.GREATER_THAN_EQ:
          image = ">=";
          break;
        case n.LESS_THAN_EQ:
          image = "<=";
          break;
        case n.LEFT_SHIFT:
          image = "<<";
          break;
        case n.RIGHT_SHIFT:
          image = ">>";
          break;
        case n.EXTENSION_RIGHT_SHIFT:
          image = ">>>";
          break;
        case n.ADD:
          image = "+";
          break;
        case n.SUBTRACT:
          image = "-";
          break;
        case n.MULTIPLE:
          image = "*";
          break;
        case n.DIVIDE:
          image = "/";
          break;
        case n.MODULE:
          image = "%";
          break;
        case n.GROUP:
          image = "/";
          break;
      default:
        throw new Error("Unexpected case.");
      }
      n.left.accept(this);
      if(n.right != null){
        out.print(" " + image + " ");
        //posCounter[1] += image.length()+1; 
        n.right.accept(this);
      }

   }

   public void visit(UnaryExprNode n) {
      String image = "";
      String post_image = "";

      switch(n.uOperator){
        case n.PLUS:
          image = "+";
          break;
        case n.MINUS:
          image = "-";
          break;
        case n.PRE_INCREMENT:
          image = "++";
          break;
        case n.PRE_DECREMENT:
          image = "--";
          break;
        case n.COMPLEMENT:
          image = "~";
          break;
        case n.LOGICAL_NOT:
          image = "!";
          break;
        case n.POSTFIX:
          post_image = "++";
          break;
        case n.POSTFIX_ADDITION:
          post_image = "++";
          break;
        case n.POSTFIX_SUBTRACTION:
          post_image = "--";
          break;
        case n.POSTFIX_INDEX_LOCATION:
          post_image = "`";
          break;
      default:
        throw new Error("Unexpected case.");
      }
      //out.print(" " + image);
      out.print(image);
      //posCounter[1] += image.length()+1; 
      n.operand.accept(this);
      out.print(post_image + "");
   }

   public void visit (MethodCallNode n) {
    status.push(PREFIX);   
	  n.objName.accept(this);
    status.pop();

    status.push(SUFFIX);
	  n.methodName.accept(this);
	  n.args.accept(this);
    status.pop();
   }

   public void visit (ClassFieldNode n) {
    status.push(PREFIX);   
	  n.objName.accept(this);
    status.pop();

    status.push(SUFFIX);
	  n.fieldName.accept(this);
    status.pop();
   }

   public void visit (ArrayElemRefNode n) {
    status.push(PREFIX);   
	  n.arrayName.accept(this);
    status.pop();

    status.push(SUFFIX);
	  n.eList.accept(this);
    status.pop();
   }

   public void visit (ArraySectNode n) {
    status.push(PREFIX);   
	  n.arrayName.accept(this);
    status.pop();

    status.push(SUFFIX);
	  n.dimList.accept(this);
    status.pop();
   }

   public void visit (InnerThisNode n) {
    status.push(PREFIX);   
	  n.objName.accept(this);
    status.pop();

    status.push(SUFFIX);
	  n.fieldName.accept(this);
    status.pop();
   }
  
   public void visit (InnerAllocationExpressionNode n) {
    status.push(PREFIX);   
	  n.objName.accept(this);
    status.pop();

    status.push(SUFFIX);
	  n.fieldName.accept(this);
    status.pop();
   }

  public void visit (ThisNode n) {
    String image = "this";

    if (status.peek()==PREFIX) {
      //out.print(image);
      out.print(" " + image);
    }
    else {
      out.print("."+image);
    }
  }

  private int flag = 0;

  public void visit (NullNode n) {
    String image = "null";
    flag++;

    if (status.peek()==PREFIX) {
      out.print(image);
    }
    else {
      out.print(""+image);
    }
  }

  public void visit (SuperNode n) {
    String image = "super";

    if (status.peek()==PREFIX) {
      out.print(image);
    }
    else {
      throw new Error();
    }
  }
  
  public void visit (LiteralNode n) {
    if (status.peek()==PREFIX) {
      n.leaf.accept(new IPrettyPrinter(this));
    }
    else {
      throw new Error();
    }
  }

  public void visit (ClassLiteralNode n) {
    String image = ".class";

    if (status.peek()==PREFIX) {
      n.leaf.accept(new IPrettyPrinter(this));
      out.print(image);
    }
    else {
      throw new Error();
    }
  }

  public void visit (TripletOrExpressionListNode n) {
    if (status.peek()==PREFIX) {
      throw new Error();
    }
    else {
      out.print("[");
      n.descriptor.leaf.accept(new IPrettyPrinter(this));
      out.print("]");
    }
  }

  public void visit (DistArraySecRefDimListNode n) {
    if (status.peek()==PREFIX) {
      throw new Error();
    }
    else {
      out.print("[[");
      n.descriptor.leaf.accept(new IPrettyPrinter(this));
      out.print("]]");
    }
  }

  public void visit (IdentifierNode n) {
    if (status.peek()==PREFIX) {
      n.leaf.accept(new IPrettyPrinter(this));
    }
    else {
      if(flag == 0)
        out.print(".");
      else
        flag--;

      n.leaf.accept(new IPrettyPrinter(this));
    }
  }

  public void visit (AllocationExpressionNode n) {
    if (status.peek()==PREFIX) {
      n.descriptor.leaf.accept(new IPrettyPrinter(this));
    }
    else {
      out.print(".");
      n.descriptor.leaf.accept(new IPrettyPrinter(this));
    }
  }

  public void visit (ExpressionNode n) {
    if (status.peek()==PREFIX) {
      out.print("(");
      n.leaf.accept(new IPrettyPrinter(this));
      out.print(")");
    }
    else {
      throw new Error();
    }
  }

  public void visit (ArgumentsNode n) {
    if (status.peek()==PREFIX) {
      throw new Error();
    }
    else {
      n.descriptor.leaf.accept(new IPrettyPrinter(this));
    }
  }

}
