/*
 * File         : ITreePrinter.java
 * Author       : Han-Ku Lee
 * Created      : Fri Oct  9 15:20:09 1998
 * Revision     : $Revision: 1.9 $
 * Updated      : $Date: 1999/10/25 22:14:55 $
 */


package hpjava.compiler.unparser;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.abstractexp.*;

import java.util.*;
import java.io.*;

public class AbstractExpressionTreePrinter extends AbstractExpressionVisitor {
  protected PrintWriter out;
  protected Space spc = new Space(2); 

  protected ITreePrinter tp;

  public AbstractExpressionTreePrinter(ITreePrinter itp){
    this.tp = itp;
    this.out = itp.out;
    this.spc = itp.spc;
  }

  public AbstractExpressionTreePrinter() {
    out = new PrintWriter(System.out, true);
  }

  public AbstractExpressionTreePrinter(Writer o) {
    out = new PrintWriter(o, true);
  }

  public AbstractExpressionTreePrinter(OutputStream o) {
    out = new PrintWriter(o, true);
  }

  public void flushWriter() { 
    out.flush(); 
  }

  //
  // Auto class visitors--probably don't need to be overridden.
  //

  //
  // User-generated visitor methods below
  //

  //
  // f0 -> "="
  //       | "*="
  //       | "/="
  //       | "%="
  //       | "+="
  //       | "-="
  //       | "<<="
  //       | ">>="
  //       | ">>>="
  //       | "&="
  //       | "^="
  //       | "|="
  //
  public void visit(AssignmentNode n) {
    out.print(spc.spc);
    out.print("AssignmentNode");
    out.println();

    spc.updateSpc(+1);
    n.left.accept(this);
    n.right.accept(this);
    spc.updateSpc(-1);
  }

  public void visit(AdditionAssignmentNode n) {
    out.print(spc.spc);
    out.print("AdditiveAssignmentNode");
    out.println();

    spc.updateSpc(+1);
    n.left.accept(this);
    n.right.accept(this);
    spc.updateSpc(-1);
  }

  public void visit(SubtractionAssignmentNode n) {
    out.print(spc.spc);
    out.print("SubtractionAssignmentNode");
    out.println();

    spc.updateSpc(+1);
    n.left.accept(this);
    n.right.accept(this);
    spc.updateSpc(-1);
  }

  public void visit(MultiplicationAssignmentNode n) {
    out.print(spc.spc);
    out.print("MultiplicationAssignmentNode");
    out.println();

    spc.updateSpc(+1);
    n.left.accept(this);
    n.right.accept(this);
    spc.updateSpc(-1);
  }

  public void visit(DivisionAssignmentNode n) {
    out.print(spc.spc);
    out.print("DivisionAssignmentNode");
    out.println();

    spc.updateSpc(+1);
    n.left.accept(this);
    n.right.accept(this);
    spc.updateSpc(-1);
  }

  public void visit(AndAssignmentNode n) {
    out.print(spc.spc);
    out.print("AndAssignmentNode");
    out.println();

    spc.updateSpc(+1);
    n.left.accept(this);
    n.right.accept(this);
    spc.updateSpc(-1);
  }

  public void visit(InclusiveOrAssignmentNode n) {
    out.print(spc.spc);
    out.print("InclusiveOrAssignmentNode");
    out.println();

    spc.updateSpc(+1);
    n.left.accept(this);
    n.right.accept(this);
    spc.updateSpc(-1);
  }

  public void visit(ExclusiveOrAssignmentNode n) {
    out.print(spc.spc);
    out.print("ExclusiveOrAssignmentNode");
    out.println();

    spc.updateSpc(+1);
    n.left.accept(this);
    n.right.accept(this);
    spc.updateSpc(-1);
  }

  public void visit(LeftShiftAssignmentNode n) {
    out.print(spc.spc);
    out.print("LeftShiftAssignmentNode");
    out.println();

    spc.updateSpc(+1);
    n.left.accept(this);
    n.right.accept(this);
    spc.updateSpc(-1);
  }

  public void visit(RightShiftAssignmentNode n) {
    out.print(spc.spc);
    out.print("RightShiftAssignmentNode");
    out.println();

    spc.updateSpc(+1);
    n.left.accept(this);
    n.right.accept(this);
    spc.updateSpc(-1);
  }

  public void visit(ExtensionRightShiftAssignmentNode n) {
    out.print(spc.spc);
    out.print("ExtensionRightShiftAssignmentNode");
    out.println();

    spc.updateSpc(+1);
    n.left.accept(this);
    n.right.accept(this);
    spc.updateSpc(-1);
  }

  //
  // f0 -> ConditionalOrExpression()
  // f1 -> [ "?" Expression() ":" ConditionalExpression() ]
  //
  public void visit(ConditionalNode n) {
    out.print(spc.spc);
    out.print("ConditionalNode");
    out.println();

    spc.updateSpc(+1);
    n.condition.accept(this);
    n.trueResult.accept(this);
    n.falseResult.accept(this);
    spc.updateSpc(-1);
  }

  //
  // f0 -> "(" Type() ")" UnaryExpression()
  //       | "(" Type() ")" UnaryExpressionNotPlusMinus()
  //
  public void visit(CastNode n) {
    out.print(spc.spc);
    out.print("CastNode");
    out.println();

    spc.updateSpc(+1);
    n.type.accept(this);
    n.leaf.accept(this);
    spc.updateSpc(-1);
  }

   //
   // f0 -> Operator
   // f1 -> AbstractExpression()
   // f2 -> AbstractExpression()
   //
   public void visit(BinExprNode n) {
      out.print(spc.spc);
      
      switch(n.binaryOperator){
        case n.CONDITIONAL_OR:
          out.print("ConditionalOrNode");
          break;
        case n.CONDITIONAL_AND:
          out.print("ConditionalAndNode");
          break;
        case n.INCLUSIVE_OR:
          out.print("InclusiveOrNode");
          break;
        case n.EXCLUSIVE_OR:
          out.print("ExclusiveOrNode");
          break;
        case n.AND:
          out.print("AndNode");
          break;
        case n.EQUALITY:
          out.print("EqualityNode");
          break;
        case n.INEQUALITY:
          out.print("InEqualityNode");
          break;
        case n.INSTANCE_OF:
          out.print("InstanceOfNode");
          break;
        case n.GREATER:
          out.print("GreaterNode");
          break;
        case n.LESS:
          out.print("LessNode");
          break;
        case n.GREATER_THAN_EQ:
          out.print("GreaterThanNode");
          break;
        case n.LESS_THAN_EQ:
          out.print("LessThanNode");
          break;
        case n.LEFT_SHIFT:
          out.print("LeftShiftNode");
          break;
        case n.RIGHT_SHIFT:
          out.print("RightShiftNode");
          break;
        case n.EXTENSION_RIGHT_SHIFT:
          out.print("ExtensionRightShiftNode");
          break;
        case n.ADD:
          out.print("AdditiveNode");
          break;
        case n.SUBTRACT:
          out.print("SubtractiveNode");
          break;
        case n.MULTIPLE:
          out.print("MultiplicativeNode");
          break;
        case n.DIVIDE:
          out.print("DivideNode");
          break;
        case n.MODULE:
          out.print("ModuleNode");
          break;
        case n.GROUP:
          out.print("GroupNode");
          break;
      default:
        throw new Error("Unexpected case.");
      }
      out.println();

      spc.updateSpc(+1);
      n.left.accept(this);
      n.right.accept(this);
      spc.updateSpc(-1);
   }

   public void visit(UnaryExprNode n) {
      out.print(spc.spc);
      
      switch(n.uOperator){
        case n.PLUS:
          out.println("PlusNode");
          break;
        case n.MINUS:
          out.println("MinusNode");
          break;
        case n.PRE_INCREMENT:
          out.println("PreIncrementNode");
          break;
        case n.PRE_DECREMENT:
          out.println("PreDecrementNode");
          break;
        case n.COMPLEMENT:
          out.println("ComplementNode");
          break;
        case n.LOGICAL_NOT:
          out.println("LogicalNotNode");
          break;
        case n.POSTFIX:
          out.println("PostfixNode");
          break;
        case n.POSTFIX_ADDITION:
          out.println("PostfixAdditionNode");
          break;
        case n.POSTFIX_SUBTRACTION:
          out.println("PostfixSubtractionNode");
          break;
        case n.POSTFIX_INDEX_LOCATION:
          out.println("IndexOfLocationNode");
          break;
      default:
        throw new Error("Unexpected case.");
      }
      //out.println();
      spc.updateSpc(+1);
      n.operand.accept(this);
      spc.updateSpc(-1);
   }

   public void visit (ArraySectNode n) {
    out.print(spc.spc);
    out.println("ArraySectNode");
    //out.println();
    spc.updateSpc(+1);
	  n.arrayName.accept(this);
	  n.dimList.accept(this);
    spc.updateSpc(-1);
   }

   public void visit (MethodCallNode n) {
    out.print(spc.spc);
    out.println("MethodCallNode");
    //out.println();
    spc.updateSpc(+1);
	  n.objName.accept(this);
	  n.methodName.accept(this);
	  n.args.accept(this);
    spc.updateSpc(-1);
   }

   public void visit (ClassFieldNode n) {
    out.print(spc.spc);
    out.println("ClassFieldNode");
    //out.println();
    spc.updateSpc(+1);
	  n.objName.accept(this);
	  n.fieldName.accept(this);
    spc.updateSpc(-1);
   }

   public void visit (ArrayElemRefNode n) {
    out.print(spc.spc);
    out.println("ArrayElemRefNode");
    //out.println();
    spc.updateSpc(+1);
	  n.arrayName.accept(this);
	  n.eList.accept(this);
    spc.updateSpc(-1);
   }

   public void visit (InnerThisNode n) {
    out.print(spc.spc);
    out.println("InnerThisNode");
    //out.println();
    spc.updateSpc(+1);
	  n.objName.accept(this);
	  n.fieldName.accept(this);
    spc.updateSpc(-1);
   }
  
   public void visit (InnerAllocationExpressionNode n) {
    out.print(spc.spc);
    out.println("InnerAllocationExpressionNode");
    //out.println();
    spc.updateSpc(+1);
	  n.objName.accept(this);
	  n.fieldName.accept(this);
    spc.updateSpc(-1);
   }


  //
  // f0 -> PrimaryExpression()
  // f1 -> ( "/" PrimaryExpression() )*
  //
  public void visit(GroupNode n) {
    out.print(spc.spc);
    out.print("GroupNode");
    out.println();

    spc.updateSpc(+1);
    n.left.accept(this);
    n.right.accept(this);
    spc.updateSpc(-1);
  }

  public void visit(TypeNode n) {
    out.print(spc.spc);
    out.print("TypeNode");
    out.println();

    spc.updateSpc(+1);
    n.leaf.accept(this.tp);
    spc.updateSpc(-1);
  }


  // From here, PrimaryExpression is controlled

  public void visit (ThisNode n) {
    out.print(spc.spc);
    out.print("ThisNode");
    out.println();
  }

  public void visit (SuperNode n) {
    out.print(spc.spc);
    out.print("SuperNode");
    out.println();
  }

  public void visit (LiteralNode n) {
    out.print(spc.spc);
    out.print("LiteralNode");
    out.println();

    spc.updateSpc(+1);
    n.leaf.accept(new ITreePrinter(this));
    spc.updateSpc(-1);
  }

  public void visit (ClassLiteralNode n) {
    out.print(spc.spc);
    out.print("ClassLiteralNode");
    out.println();

    spc.updateSpc(+1);
    n.leaf.accept(new ITreePrinter(this));
    spc.updateSpc(-1);
  }

  public void visit (TripletOrExpressionListNode n) {
    out.print(spc.spc);
    out.print("TripletOrExpressionListNode");
    out.println();

    spc.updateSpc(+1);
    n.descriptor.leaf.accept(new ITreePrinter(this));
    spc.updateSpc(-1);
  }

  public void visit (DistArraySecRefDimListNode n) {
    out.print(spc.spc);
    out.print("DistArraySecRefDimListNode");
    out.println();

    spc.updateSpc(+1);
    n.descriptor.leaf.accept(new ITreePrinter(this));
    spc.updateSpc(-1);
  }

  public void visit (IdentifierNode n) {
    out.print(spc.spc);
    out.print("IdentifierNode");
    out.println();

    spc.updateSpc(+1);
    out.print(spc.spc);
    n.leaf.accept(new ITreePrinter(this));
    out.println();
    spc.updateSpc(-1);
  }

  public void visit (AllocationExpressionNode n) {
    out.print(spc.spc);
    out.print("AllocationExpressionNode");
    out.println();

    spc.updateSpc(+1);
    n.descriptor.leaf.accept(new ITreePrinter(this));
    spc.updateSpc(-1);
  }

  public void visit (ExpressionNode n) {
    out.print(spc.spc);
    out.print("ExpressionNode");
    out.println();

    spc.updateSpc(+1);
    n.leaf.accept(new ITreePrinter(this));
    spc.updateSpc(-1);
  }

  public void visit (ArgumentsNode n) {
    out.print(spc.spc);
    out.print("ArgumentsNode");
    out.println();

    spc.updateSpc(+1);
    n.descriptor.leaf.accept(new ITreePrinter(this));
    spc.updateSpc(-1);
  }

}
