/*
 * File         : AbstractExpressionUnparsePrinter.java
 * Author       : Han-Ku Lee
 * Created      : Thu Oct  8 16:10:59 1998
 * Revision     : $Revision: 1.9 $
 * Updated      : $Date: 1999/10/25 22:14:56 $
 */

package hpjava.compiler.unparser;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.abstractexp.*;
import hpjava.compiler.util.*;

//import hpjava.compiler.unparser.*;

import java.util.*;
import java.io.*;

public class AbstractExpressionUnparsePrinter 
  extends AbstractExpressionVisitor {
    
  final int PREFIX = 0;
  final int SUFFIX = 1;

  protected PrintWriter out;

  protected int posCounter [];

//  public int lineCounter = 1;
//  public int columnCounter = 1;

  protected IUnparsePrinter iup;  

  protected NodeToken previous = null;  

  private IntStack status=new IntStack(); {
    // for simple expression with out concatenation, it is a prefix.
    status.push(PREFIX);
  }

  public AbstractExpressionUnparsePrinter(IUnparsePrinter iup){
    this.iup = iup;
    this.out = iup.out;
    this.previous = iup.previous;
    this.posCounter = iup.posCounter;
  }

  public AbstractExpressionUnparsePrinter() { 
    this.out = new PrintWriter(System.out, true); 

    this.posCounter = new int [2];

    this.posCounter[0]=1;
    this.posCounter[1]=1;    
  }

  public AbstractExpressionUnparsePrinter(Writer o) { 
    this.out = new PrintWriter(o, true); 

    this.posCounter = new int [2];

    this.posCounter[0]=1;
    this.posCounter[1]=1;    
  }

  public AbstractExpressionUnparsePrinter(OutputStream o) { 
    this.out = new PrintWriter(o, true); 

    this.posCounter = new int [2];

    this.posCounter[0]=1;
    this.posCounter[1]=1;    
  }

  public void flushWriter(){ out.flush(); }


  // write unparse here.

  public void visit(TypeNode n) {

    n.leaf.accept(iup);

  }


   public void visit(AssignmentNode n) {
      String image = "=";

      n.left.accept(this);
      out.print(" " + image);
      posCounter[1] += image.length()+1; 
      n.right.accept(this);
   }

   public void visit(AdditionAssignmentNode n) {
      String image = "+=";

      n.left.accept(this);
      out.print(" " + image);
      posCounter[1] += image.length()+1; 
      n.right.accept(this);
   }

   public void visit(SubtractionAssignmentNode n) {
      String image = "-=";

      n.left.accept(this);
      out.print(" " + image);
      posCounter[1] += image.length()+1; 
      n.right.accept(this);
   }

   public void visit(MultiplicationAssignmentNode n) {
      String image = "*=";

      n.left.accept(this);
      out.print(" " + image);
      posCounter[1] += image.length()+1; 
      n.right.accept(this);
   }

   public void visit(DivisionAssignmentNode n) {
      String image = "/=";

      n.left.accept(this);
      out.print(" " + image);
      posCounter[1] += image.length()+1; 
      n.right.accept(this);
   }

   public void visit(AndAssignmentNode n) {
      String image = "&=";

      n.left.accept(this);
      out.print(" " + image);
      posCounter[1] += image.length()+1; 
      n.right.accept(this);
   }

   public void visit(InclusiveOrAssignmentNode n) {
      String image = "|=";

      n.left.accept(this);
      out.print(" " + image);
      posCounter[1] += image.length()+1; 
      n.right.accept(this);
   }

   public void visit(ExclusiveOrAssignmentNode n) {
      String image = "^=";

      n.left.accept(this);
      out.print(" " + image);
      posCounter[1] += image.length()+1; 
      n.right.accept(this);
   }

   public void visit(LeftShiftAssignmentNode n) {
      String image = "<<=";

      n.left.accept(this);
      out.print(" " + image);
      posCounter[1] += image.length()+1; 
      n.right.accept(this);
   }

   public void visit(RightShiftAssignmentNode n) {
      String image = ">>=";

      n.left.accept(this);
      out.print(" " + image);
      posCounter[1] += image.length()+1; 
      n.right.accept(this);
   }

   public void visit(ModulusAssignmentNode n) {
      String image = "%=";

      n.left.accept(this);
      out.print(" " + image);
      posCounter[1] += image.length()+1; 
      n.right.accept(this);
   }

   public void visit(ExtensionRightShiftAssignmentNode n) {
      String image = ">>>=";

      n.left.accept(this);
      out.print(" " + image);
      posCounter[1] += image.length()+1; 
      n.right.accept(this);
   }

   //
   // f0 -> ConditionalOrExpression()
   // f1 -> [ "?" Expression() ":" ConditionalExpression() ]
   //
   public void visit(ConditionalNode n) {
      String image1 = "?";
      String image2 = ":";

      n.condition.accept(this);

      if(n.trueResult != null || n.falseResult != null){
        out.print(" " + image1);
        posCounter[1] += image1.length()+1; 
        n.trueResult.accept(this);

        out.print(" " + image2);
        posCounter[1] += image2.length()+1; 
        n.falseResult.accept(this);
      }
   }

   //
   // f0 -> Operator
   // f1 -> AbstractExpression()
   // f2 -> AbstractExpression()
   //
   public void visit(BinExprNode n) {
      String image;
      switch(n.binaryOperator){
        case n.CONDITIONAL_OR:
          image = "||";
          break;
        case n.CONDITIONAL_AND:
          image = "&&";
          break;
        case n.INCLUSIVE_OR:
          image = "|";
          break;
        case n.EXCLUSIVE_OR:
          image = "^";
          break;
        case n.AND:
          image = "&";
          break;
        case n.EQUALITY:
          image = "==";
          break;
        case n.INEQUALITY:
          image = "!=";
          break;
        case n.INSTANCE_OF:
          image = "instanceof ";
          break;
        case n.GREATER:
          image = ">";
          break;
        case n.LESS:
          image = "<";
          break;
        case n.GREATER_THAN_EQ:
          image = ">=";
          break;
        case n.LESS_THAN_EQ:
          image = "<=";
          break;
        case n.LEFT_SHIFT:
          image = "<<";
          break;
        case n.RIGHT_SHIFT:
          image = ">>";
          break;
        case n.EXTENSION_RIGHT_SHIFT:
          image = ">>>";
          break;
        case n.ADD:
          image = "+";
          break;
        case n.SUBTRACT:
          image = "-";
          break;
        case n.MULTIPLE:
          image = "*";
          break;
        case n.DIVIDE:
          image = "/";
          break;
        case n.MODULE:
          image = "%";
          break;
        case n.GROUP:
          image = "/";
          break;
      default:
        throw new Error("Unexpected case.");
      }
      n.left.accept(this);
      if(n.right != null){
        out.print(" " + image);
        posCounter[1] += image.length()+1; 
        n.right.accept(this);
      }
   }

   public void visit(UnaryExprNode n) {
      String image = "";
      String post_image = "";

      switch(n.uOperator){
        case n.PLUS:
          image = "+";
          break;
        case n.MINUS:
          image = "-";
          break;
        case n.PRE_INCREMENT:
          image = "++";
          break;
        case n.PRE_DECREMENT:
          image = "--";
          break;
        case n.COMPLEMENT:
          image = "~";
          break;
        case n.LOGICAL_NOT:
          image = "!";
          break;
        case n.POSTFIX:
          post_image = "++";
          break;
        case n.POSTFIX_ADDITION:
          post_image = "++";
          break;
        case n.POSTFIX_SUBTRACTION:
          post_image = "--";
          break;
        case n.POSTFIX_INDEX_LOCATION:
          post_image = "`";
          break;
      default:
        throw new Error("Unexpected case.");
      }
      //out.print(" " + image);
      out.print(image);
      posCounter[1] += image.length()+1; 
      n.operand.accept(this);
      out.print(post_image + "");
   }

   public void visit (MethodCallNode n) {
    status.push(PREFIX);   
	  n.objName.accept(this);
    status.pop();

    status.push(SUFFIX);
	  n.methodName.accept(this);
	  n.args.accept(this);
    status.pop();
   }

   public void visit (ClassFieldNode n) {
    status.push(PREFIX);   
	  n.objName.accept(this);
    status.pop();

    status.push(SUFFIX);
	  n.fieldName.accept(this);
    status.pop();
   }

   public void visit (ArrayElemRefNode n) {
    status.push(PREFIX);   
	  n.arrayName.accept(this);
    status.pop();

    status.push(SUFFIX);
	  n.eList.accept(this);
    status.pop();
   }

   public void visit (ArraySectNode n) {
    status.push(PREFIX);   
	  n.arrayName.accept(this);
    status.pop();

    status.push(SUFFIX);
	  n.dimList.accept(this);
    status.pop();
   }

   public void visit (InnerThisNode n) {
    status.push(PREFIX);   
	  n.objName.accept(this);
    status.pop();

    status.push(SUFFIX);
	  n.fieldName.accept(this);
    status.pop();
   }
  
   public void visit (InnerAllocationExpressionNode n) {
    status.push(PREFIX);   
	  n.objName.accept(this);
    status.pop();

    status.push(SUFFIX);
	  n.fieldName.accept(this);
    status.pop();
   }

   //
   // f0 -> "(" Type() ")" UnaryExpression()
   //       | "(" Type() ")" UnaryExpressionNotPlusMinus()
   //
   public void visit(CastNode n) {
      String image1 = "(";
      String image2 = ")";

      out.print(" " + image1);
      posCounter[1] += image1.length()+1; 
      n.type.accept(this);

      out.print(image2);
      posCounter[1] += image2.length()+1; 
      n.leaf.accept(this);
   }


  // From here PrimaryExpression is taken care

  public void visit (ThisNode n) {
    String image = "this";

    if (status.peek()==PREFIX) {
      //out.print(image);
      out.print(" " + image);
      posCounter[1] += image.length();
    }
    else {
      out.print("."+image);
      posCounter[1] += image.length()+1;
    }
  }

  private int flag = 0;

  public void visit (NullNode n) {
    String image = "";
    flag++;

    if (status.peek()==PREFIX) {
      out.print(image);
      posCounter[1] += image.length();
    }
    else {
      out.print(""+image);
      posCounter[1] += image.length()+1;
    }
  }

  public void visit (SuperNode n) {
    String image = "super";

    if (status.peek()==PREFIX) {
      out.print(image);
      posCounter[1] += image.length();
    }
    else {
      throw new Error();
    }
  }
  
  public void visit (LiteralNode n) {
    if (status.peek()==PREFIX) {
      n.leaf.accept(new IUnparsePrinter(this));
    }
    else {
      throw new Error();
    }
  }

  public void visit (ClassLiteralNode n) {
    String image = ".class";

    if (status.peek()==PREFIX) {
      n.leaf.accept(new IUnparsePrinter(this));
      out.print(image);
      posCounter[1] += image.length();
    }
    else {
      throw new Error();
    }
  }

  public void visit (TripletOrExpressionListNode n) {
    if (status.peek()==PREFIX) {
      throw new Error();
    }
    else {
      out.print("[");

      previous = new NodeToken("[");
      previous.kind = hpjava.compiler.parser.HPJavaParserConstants.LBRACKET;
      posCounter[1]++;

      n.descriptor.leaf.accept(new IUnparsePrinter(this));

      out.print("]");

      previous = new NodeToken("]");
      previous.kind = hpjava.compiler.parser.HPJavaParserConstants.RBRACKET;
      posCounter[1]++;

    }
  }

  public void visit (DistArraySecRefDimListNode n) {
    if (status.peek()==PREFIX) {
      throw new Error();
    }
    else {
      out.print("[[");

      //previous = new NodeToken("[");
      previous = new NodeToken("[[");
      previous.kind = hpjava.compiler.parser.HPJavaParserConstants.LBRACKET;
      posCounter[1]++;      
      posCounter[1]++;

      n.descriptor.leaf.accept(new IUnparsePrinter(this));

      out.print("]]");

      //previous = new NodeToken("]");
      previous = new NodeToken("]]");
      previous.kind = hpjava.compiler.parser.HPJavaParserConstants.RBRACKET;
      posCounter[1]++;
      posCounter[1]++;
    }
  }

  public void visit (IdentifierNode n) {
    if (status.peek()==PREFIX) {
      n.leaf.accept(new IUnparsePrinter(this));
    }
    else {
      if(flag == 0)
        out.print(".");
      else
        flag--;

      previous = new NodeToken(".");
      previous.kind = hpjava.compiler.parser.HPJavaParserConstants.DOT;
      posCounter[1] ++;

      n.leaf.accept(new IUnparsePrinter(this));
    }
  }

  public void visit (AllocationExpressionNode n) {
    if (status.peek()==PREFIX) {
      n.descriptor.leaf.accept(new IUnparsePrinter(this));
    }
    else {
      out.print(".");

      previous = new NodeToken(".");
      previous.kind = hpjava.compiler.parser.HPJavaParserConstants.DOT;
      posCounter[1] ++;

      n.descriptor.leaf.accept(new IUnparsePrinter(this));
    }
  }

  public void visit (ExpressionNode n) {
    if (status.peek()==PREFIX) {
      out.print("(");

      previous = new NodeToken("(");
      previous.kind = hpjava.compiler.parser.HPJavaParserConstants.LPAREN;
      posCounter[1] ++;

      n.leaf.accept(new IUnparsePrinter(this));

      out.print(")");

      previous = new NodeToken(")");
      previous.kind = hpjava.compiler.parser.HPJavaParserConstants.RPAREN;
      posCounter[1] ++;
    }
    else {
      throw new Error();
    }
  }

  public void visit (ArgumentsNode n) {
    if (status.peek()==PREFIX) {
      throw new Error();
    }
    else {
      n.descriptor.leaf.accept(new IUnparsePrinter(this));
      //out.print("|");
    }
  }

}
