/*
 * File         : ITreePrinter.java
 * Author       : Han-Ku Lee
 * Created      : Fri Oct  9 15:20:09 1998
 * Revision     : $Revision: 1.7 $
 * Updated      : $Date: 1999/06/08 19:37:36 $
 */


package hpjava.compiler.unparser;

//import misc.Spacing;
import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

import hpjava.compiler.abstractexp.*;

import java.util.*;
import java.io.*;

public class ITreePrinter extends TreePrinter {
  //   protected PrintWriter out;
  //   protected Space spc = new Space(1);   // see below for class Space

  public ITreePrinter()    { 
    super(); 

    // discard the one created by its super class;
    this.spc = new Space(2);
  }
  
  public ITreePrinter(Writer o) { 
    super(o); 

    // discard the one created by its super class;
    this.spc = new Space(2);
  }
  
  public ITreePrinter(OutputStream o)  { 
    super(o); 

    // discard the one created by its super class;
    this.spc = new Space(2);
  }

  public ITreePrinter(AbstractExpressionTreePrinter aetp) { 
    super(aetp.out);

    // discard the one created by its super class;
    this.spc = aetp.spc;
  }
  /*
  public ITreePrinter(AbstractPrimaryExpressionTreePrinter apetp) { 
    super(apetp.out);

    // discard the one created by its super class;
    this.spc = apetp.spc;
  }
  */

   //
   // f0 -> PrimaryExpression()
   // f1 -> ( "/" PrimaryExpression() )*
   //
  public void visit(GroupExpression n) {
    out.print(spc.spc);
    out.print("GroupExpression");
    out.println();

    flushWriter();

    spc.updateSpc(+1);
    AbstractExpressionTreePrinter aetp =
      new AbstractExpressionTreePrinter(this);

    AbstractExpression aeNode = ExpressionManager.lookup(n);

    aeNode.accept(aetp);

    spc.updateSpc(-1);
  }


   //
   // f0 -> PreIncrementExpression()
   //       | PreDecrementExpression()
   //       | PrimaryExpression() [ "++" | "--" | AssignmentOperator() Expression() ]
   //
  public void visit(StatementExpression n) {
    out.print(spc.spc);
    out.print("StatementExpression");
    out.println();

    flushWriter();

    spc.updateSpc(+1);
    AbstractExpressionTreePrinter aetp =
      new AbstractExpressionTreePrinter(this);

    AbstractExpression aeNode = ExpressionManager.lookup(n);
    aeNode.accept(aetp);

    spc.updateSpc(-1);
  }

   //
   // f0 -> ConditionalExpression()
   // f1 -> [ AssignmentOperator() Expression() ]
   //
  public void visit(Expression n) {
    out.print(spc.spc);
    out.print("Expression");
    out.println();

    flushWriter();

    spc.updateSpc(+1);

    AbstractExpressionTreePrinter aetp =
      new AbstractExpressionTreePrinter(this);

    AbstractExpression aeNode = ExpressionManager.lookup(n);
    aeNode.accept(aetp);

    spc.updateSpc(-1);
  }

   //
   // f0 -> PrimaryPrefix()
   // f1 -> ( PrimarySuffix() )*
   //
  public void visit(PrimaryExpression n) {
    out.print(spc.spc);
    out.print("PrimaryExpression");
    out.println();

    flushWriter();

    spc.updateSpc(+1);

    AbstractExpressionTreePrinter aetp =
      new AbstractExpressionTreePrinter(this);

    AbstractExpression aeNode = ExpressionManager.lookup(n);
    aeNode.accept(aetp);

    spc.updateSpc(-1);
  }

}


