//
// A modified version of the main() method found in the Java 1.1 grammar.
// To add your visitors to the program, go to near the end, and copy or
// replace the line "root.accept(new Visitor());" with your own visitor
//

package hpjava.compiler.unparser;

import hpjava.compiler.parser.*;
import hpjava.compiler.abstractexp.*;
import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

import java.util.*;
import java.io.*;

public class IUnparser {

  static private boolean iTree = false;

  public static void main(String args[]) {
    HPJavaParser parser;
    if (args.length == 0) {
      System.err.println("HPJava Parser Version 1.1:  Reading from standard " +
                         "input . . .");
      parser = new HPJavaParser(System.in);
    }
    else if (args.length == 1) {
      System.err.println("HPJava Parser Version 1.1:  Reading from file " +
                         args[0] + " . . .");
      try {
        parser = new HPJavaParser(new java.io.FileInputStream(args[0]));
      } catch (java.io.FileNotFoundException e) {
        System.err.println("HPJava Parser Version 1.1:  File " + args[0] +
                           " not found.");
        return;
      }
    }
    else if (args.length == 2) {
      if (args[0].equals("-t"))
        iTree = true;

      System.err.println("HPJava Parser Version 1.1:  Reading from file " +
                         args[1] + " . . .");
      try {
        parser = new HPJavaParser(new java.io.FileInputStream(args[1]));
      } catch (java.io.FileNotFoundException e) {
        System.err.println("HPJava Parser Version 1.1:  File " + args[1] +
                           " not found.");
        return;
      }
    }
    else {
      System.err.println("HPJava Parser Version 1.1:  Usage is one of:");
      System.err.println("         java HPJavaParser < inputfile");
      System.err.println("OR");
      System.err.println("         java HPJavaParser inputfile");
      return;
    }

    try {
      //
      // Here's where the AST actions and visiting take place.
      //
      Node root = parser.CompilationUnit();
      System.err.println("HPJava Parser Version 1.1:  HPJava program parsed " +
                         "successfully.");

      /**** Memory tests--you can ignore this section ****
            Runtime rt = Runtime.getRuntime();
            rt.gc();
            System.out.println("Total Memory in VM: " +
            Long.toString(rt.totalMemory()));
            System.out.println("Total Free Memory:  " +
            Long.toString(rt.freeMemory()));
            System.in.read();
            **** Memory tests--you can ignore this section ****/

      ExpressionManager em = new ExpressionManager(root);

      //PrimaryExpressionManager pem = new PrimaryExpressionManager(root);
      
      if (iTree){
        root.accept(new ITreePrinter());
      }

      root.accept(new IUnparsePrinter());

    } catch (ParseException e) {
      System.err.println(e.getMessage());
      System.err.println("HPJava Parser Version 1.1:  Encountered errors " +
                         "during parse.");
    }
  }
}
