/*
 * File         : Project.java
 * Author       : Guansong Zhang
 * Created      : Thu May 14 12:03:02 1998
 * Revision     : $Revision: 1.8 $
 * Updated      : $Date: 1999/12/02 20:42:00 $
 */

package hpjava.compiler.util;

import java.io.*;
import java.util.*;

import hpjava.compiler.parser.*;
import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;
import hpjava.compiler.unparser.*;

public class Project {

  // data dealed in this package
  static protected OutputStream out;

  static final int FILENUM = 8;

  // here is public ?
  static private CMDline cmd;  
  static public Hashtable rootFileTable = new Hashtable (FILENUM);  

  // data dealed in separate packages.

  /*******************************/      

  // method in global
  public boolean isVerbose() {
    return cmd.isVerbose();
  }
      
  public void output() throws NotCompleteException {

    boolean OutputNotComplete=false;
            
    Enumeration enum = rootFileTable.keys();
    
    LOOP: while (enum.hasMoreElements()) {

      File srcPath,srcDirPath; 
      File dstPath,dstDirPath;
      
      String srcFilePathName, srcFileName;
      String dstFilePathName, dstFileName;

      Node root = (Node) enum.nextElement();      
      srcFilePathName = (String) rootFileTable.get(root);

      srcPath = new File ( (new File (srcFilePathName)).getAbsolutePath() );      

      srcFileName = srcPath.getName();

      srcDirPath=new File ( srcPath.getParent() );
        
      if (cmd.hasDir()) 
        dstDirPath=new File( (new File(cmd.getDir())).getAbsolutePath() ); 
      else
        dstDirPath=srcDirPath;

      if (cmd.isCat()) {
        out = System.out;
        dstFilePathName="System.out";
      }
      else {
        // here is a simple method to prevent overwrite source.      
        if (srcFileName.endsWith("java")) {
          out = null;
          dstFilePathName="null";          
        }
        else {
          int lastIndexOfhpj=srcFileName.lastIndexOf("hpj");
          dstFileName = srcFileName.substring(0,lastIndexOfhpj)+"java";

          dstPath = new File (dstDirPath, dstFileName);
          dstFilePathName = dstPath.getAbsolutePath();

//          System.err.println(" dst file "+ dstPath);
          try {
            out = new FileOutputStream(dstPath);
          }
          catch (IOException e) {    
            System.err.println("Can not create output file "+dstFilePathName+".");
            OutputNotComplete=true;
            continue LOOP;
          }
        }
      }
      
      if (out != null)
        root.accept(new IUnparsePrinter(out));     
        // root.accept(new UnparsePrinter(out));     
        // The above line was changed by Qiang Zheng because
        // the IUnparsePrinter cannot do correct job
        // It doesn't produce a blank between "return' and "this"
        // And sometimes "instance" and an object array
    
      if (cmd.isVerbose()) {
        System.err.println("Java file generated to "+dstFilePathName+".");
        System.err.println();
      }  
    }

  if (OutputNotComplete)
     throw new NotCompleteException("Output not complete.");
  }

  void createProject() throws FileNotFoundException, NotCompleteException {
    boolean ParseNotComplete=false;
  
    Enumeration enum = cmd.getFiles().elements();    
    
    LOOP: while (enum.hasMoreElements()) {    
	    String name = (String)enum.nextElement();
	    
      try {    
	      FileInputStream fileInputStream = new FileInputStream(name);
	      HPJavaParser parser = new HPJavaParser(fileInputStream);

	      Node root = parser.CompilationUnit();

	      //root.accept(new PrettyPrinter());
	      rootFileTable.put(root, name);

	      if (cmd.isVerbose()) {
	        System.err.println("File: "+name+" parsed successfully.");
	        System.err.println();
        }     

	    }	  
	    catch (ParseException e) {
	      System.err.println(e.getMessage());
	      System.err.println("File: "+name+" encountered errors during parse.");
	      ParseNotComplete=true;
	      continue LOOP;
      }     
    }
    
    if (ParseNotComplete)
      throw new NotCompleteException ("Parsing not complete.");
  }

  public Project (CMDline cmd) throws FileNotFoundException, NotCompleteException {
    this.cmd=cmd;

    createProject();

  }  
}

