// Modified by Qiang Zheng on April 21, 1999
// Removed accept method after move the descriptor from 
// abstractexp directory to wrapper directory

/*
 * File         : AllocationExpressionDescriptor.java
 * Author       : Guansong Zhang
 * Created      : Wed Aug 12 14:23:07 1998
 * Revision     : $Revision: 1.2 $
 * Updated      : $Date: 1999/11/11 02:12:47 $
 */

package hpjava.compiler.wrapper;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

import hpjava.compiler.reflection.*;

import hpjava.compiler.unparser.*;

import hpjava.compiler.abstractexp.*;

public class AllocationExpressionDescriptor {
  public AllocationExpression leaf;
  
  public AllocationExpressionDescriptor (AllocationExpression leaf) {
    this.leaf = leaf;
  }

  public AllocationExpressionDescriptor (ClassReflector cr) {
    if (cr.isDistArray()) {
        // this one is most inner choice.
        NodeChoice choice = new NodeChoice (null,-1);
        
        NodeSequence sequence = new NodeSequence(3);

        sequence.addNode(new NodeToken("new"));
        sequence.addNode(getNameNode(cr));
        sequence.addNode(choice);

        //sequence.accept(new TreePrinter());
        
        NodeChoice allocationChoice = new NodeChoice(sequence, 1);

        this.leaf = new AllocationExpression(allocationChoice);
    } else if (cr.isArray()) {
      throw new Error("HPJava: native array not supported");
    } else {
      // throw new Error("HPJava: Types other than array not finished");
        NodeChoice choice = new NodeChoice( null, -1 );
        NodeSequence sequence = new NodeSequence( 3 );
        sequence.addNode(new NodeToken("new"));
        sequence.addNode(getNameNode(cr));
        sequence.addNode(choice);
        NodeChoice allocationChoice = new NodeChoice(sequence, 1);

        this.leaf = new AllocationExpression(allocationChoice);
    }
  }

  private Name getNameNode(ClassReflector cr) {
    // rewrite this later.

    String name = cr.getName();

    NodeToken id = new NodeToken(name);

    NodeListOptional listOptional = new NodeListOptional();

    return new Name(id, listOptional);
  }

  public boolean isPrimitiveTypeArrayAllocator() {
    return ( this.leaf.f0.which == 0 );
  }

  public boolean isClassAllocator() {
    return ( this.leaf.f0.which == 1 );
  }

  public PrimitiveTypeDescriptor primitiveType() {
    NodeSequence sequence = (NodeSequence) this.leaf.f0.choice;
    return new PrimitiveTypeDescriptor((PrimitiveType) sequence.elementAt(1));
  }

  public Name className() {
    NodeSequence sequence = (NodeSequence) this.leaf.f0.choice;
    return (Name) sequence.elementAt(1); // return a descriptor later.
  }

  private ArrayShapes arrayShapes() {
    if (isPrimitiveTypeArrayAllocator()) {
      NodeSequence sequence = (NodeSequence) leaf.f0.choice;
      ArrayShapes shapes = (ArrayShapes) sequence.elementAt(2);
      return shapes;
    } 
    else {
      NodeSequence sequence = (NodeSequence) leaf.f0.choice;

      NodeChoice choice = (NodeChoice) sequence.elementAt(2);

      if (choice.which == 1) {
        return null;
      }
      else {
        ArrayShapes shapes = (ArrayShapes) choice.choice;
        return shapes;
      }
    }
  }

  public boolean isGSReference() {
    ArrayShapes shapes = arrayShapes();

    if (shapes != null) {
      return (shapes.f0.which == 2);
    }
    else {
      return false;
    }
  }

  public int numberOfAllocatedShapes() {
    ArrayShapes shapes = arrayShapes();

    if (shapes == null)
      return 0;
    
    if (shapes.f0.which !=0 )
      return 0;

    NodeSequence sequence = (NodeSequence) shapes.f0.choice;
    NodeList list = (NodeList) sequence.elementAt(0);

    return list.size();
  }

  public AllocatedShapeDescriptor allocatedShape(int i) {
    if (numberOfAllocatedShapes() != 0 ) {
      ArrayShapes shapes = arrayShapes();

      NodeSequence sequence = (NodeSequence) shapes.f0.choice;
      NodeList list = (NodeList)sequence.elementAt(0);

      AllocatedShape allocatedShape = (AllocatedShape) list.elementAt(i);

      return new AllocatedShapeDescriptor(allocatedShape);
    }
    else
      return null;
  }

  public int numberOfSignatures() {
    ArrayShapes shapes = arrayShapes();
    if (shapes == null)
      return 0;

    if (shapes.f0.which !=0 )
      return 0;

    NodeSequence sequence = (NodeSequence) shapes.f0.choice;
    NodeListOptional list = (NodeListOptional) sequence.elementAt(1);

    return list.size();
  }

  public void addArguments(ArgumentsDescriptor ad) {
    if (this.leaf.f0.which==0)
      throw new Error();
    else {
      NodeOptional optional= new NodeOptional();

      NodeSequence sequence = new NodeSequence(2);
      sequence.addNode(ad.leaf);
      sequence.addNode(optional);
      
      NodeSequence allocationSequence =(NodeSequence) 
        ((NodeChoice)this.leaf.f0).choice;

      NodeChoice choice = (NodeChoice) allocationSequence.elementAt(2);

      if (choice.which!=-1) {
        throw new Error("Added twice.");
      }
      else {
        /* choice.choice = ad.leaf;
        choice.which = 2; */
        choice.choice = sequence;
        choice.which = 1;
      }

    }
  }
  
  // public void accept(AbstractPrimaryExpressionVisitor v) {
  //   v.visit(this);
  // }
  
}
