/*
 * File         : ArgumentsDescriptor.java
 * Author       : Guansong Zhang
 * Created      : Wed Aug 12 14:23:07 1998
 * Revision     : $Revision: 1.2 $
 * Updated      : $Date: 1999/06/08 14:58:45 $
 */

/* Modified by Qiang Zheng
 * Move ArgumentsDescriptor to package wrapper from package abstractexp
 * In this file, I comment off the accept function
 * And this class no longer implements PrimaryItem
 */ 

package hpjava.compiler.wrapper;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

import hpjava.compiler.abstractexp.*;
import hpjava.compiler.unparser.*;

public class ArgumentsDescriptor {
  public Arguments leaf;
  
  public ArgumentsDescriptor (Arguments leaf) {
    this.leaf = leaf;
  }

  /*
 
   //
   // f0 -> "("
   // f1 -> [ ArgumentList() ]
   // f2 -> ")"
   //
   public void visit(Arguments n) {
      n.f0.accept(this);
      n.f1.accept(this);
      n.f2.accept(this);
   }

   //
   // f0 -> Expression()
   // f1 -> ( "," Expression() )*
   //
   public void visit(ArgumentList n) {
      n.f0.accept(this);
      n.f1.accept(this);
   }

   */

  public ArgumentsDescriptor () {
    NodeToken left = new NodeToken("(");
    left.kind = hpjava.compiler.parser.HPJavaParserConstants.LPAREN;

    NodeToken right = new NodeToken(")");
    right.kind =hpjava.compiler.parser.HPJavaParserConstants.RPAREN;

    NodeOptional optional = new NodeOptional();

    this.leaf = new Arguments(left,optional,right);
  }

  public ArgumentsDescriptor (Expression n) {
    this();
    this.addArg(n);
  }

  public ArgumentsDescriptor (AbstractExpression n) {
    this();
    this.addArg(n);
  }

  /* public ArgumentsDescriptor (AbstractPrimaryExpression n) {
    this();
    this.addArg(n);
  } */

  /* public void accept(AbstractPrimaryExpressionVisitor v) {
    v.visit(this);
  } */

  public void addArg(Expression exp) {
    NodeOptional nodeOptional = this.leaf.f1;

    if (nodeOptional.present()) { 
      ArgumentList args = (ArgumentList)nodeOptional.node;

      NodeSequence sequence = new NodeSequence(2);

      NodeToken comma = new NodeToken(",");
      comma.kind = hpjava.compiler.parser.HPJavaParserConstants.COMMA;

      sequence.addNode(comma);
      sequence.addNode(exp);

      args.f1.addNode(sequence);
    } 
    else { 
      NodeListOptional nodeListOptional = new NodeListOptional();
      
      ArgumentList args = new ArgumentList(exp, nodeListOptional);

      nodeOptional.addNode(args);        
    } 

  }
  
  
  public void addArg(AbstractExpression exp) {
    Expression empty = new Expression (null, null);

    ExpressionManager.register(empty, exp);

    addArg(empty);

  }
  
  /* public void addArg(AbstractPrimaryExpression exp) {
    PrimaryExpression emptyPrimary = new PrimaryExpression (null, null);
    PrimaryExpressionManager.register(emptyPrimary, exp);

    AbstractExpression aexp = new PrimaryExpressionNode(emptyPrimary);

    Expression empty = new Expression(null, null);
    ExpressionManager.register(empty, aexp);

    addArg(empty);

  } */

  public int numberOfArgs() {
    NodeOptional nodeOptional = this.leaf.f1;

    if (!nodeOptional.present()) {
      return 0;
    }
    else {
      ArgumentList args = (ArgumentList) nodeOptional.node;
      
      if (!args.f1.present()) {
        return 1;
      }
      else {
        int size = args.f1.size();
        return 1+size;
      }
    }
  }

  public Expression arg(int i) {

    NodeOptional nodeOptional = this.leaf.f1; 
    
    ArgumentList args;

    if (!nodeOptional.present()) {
      throw new Error("No element inside.");
    }
    else {
      args = (ArgumentList)nodeOptional.node;
    }

    if (i==0) {
      return (Expression) args.f0;
    }
    else {
      NodeSequence sequence = (NodeSequence) args.f1.elementAt(i-1);
      return (Expression)sequence.elementAt(1);
    }
  }

  public AbstractExpression abstractArg(int i) {
    Expression exp = arg(i);
    return ExpressionManager.lookup(exp);
  }

}
