/*
 * File         : NamedLocationRangeListDescriptor.java
 * Author       : Han-Ku Lee
 * Created      : Fri Oct 29 17:46:33 1999
 * Revision     : $Revision: 1.1 $
 * Updated      : $Date: 1999/10/29 22:00:16 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1999
 */

package hpjava.compiler.wrapper;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

import hpjava.compiler.unparser.*;

/**
 * Grammar production:
 * f0 -> NamedLocationRange()
 * f1 -> ( "," NamedLocationRange() )*
 */

public class NamedLocationRangeListDescriptor {
  public NamedLocationRangeList leaf;

  public NamedLocationRangeListDescriptor (NamedLocationRangeList leaf) {
    this.leaf = leaf;
  }

  public int numberOfLocs() {
    return 1 + leaf.f1.size();
  }

  public NamedLocationRange loc(int i) {
    if (i==0) {
      return leaf.f0;
    }
    else {
      return (NamedLocationRange) leaf.f1.elementAt(i-1);
    }
  }

  public NodeToken name(int i) {
    NamedLocationRange loc = loc(i);

    return loc.f0;
  }

  public PrimaryExpression base(int i) {
    NamedLocationRange loc = loc(i);

    return loc.f2;
  } 

  // We don't need this method any more...
  /*
  public boolean hasRangeSpecifier(int i) {
    NamedLocationRange loc = loc(i);

    return loc.f3.present();
  }
  */

  public Triplet range(int i) {
    /*
    if (hasRangeSpecifier(i)) {
      NamedLocationRange loc = loc(i);      
      return (Triplet) loc.f4.node;
    }
    else 
      throw new Error("No range specified.");
    */

    NamedLocationRange loc = loc(i);
    return (Triplet) loc.f4;
  }
  
  public void addLoc(NamedLocationRange loc) {
    if (leaf.f0 == null)
      leaf.f0=loc;
    else 
      leaf.f1.addNode(loc);
  }

}
