/*
 * File         : TypeDescriptor.java
 * Author       : Guansong Zhang
 * Created      : Wed Aug 12 14:23:07 1998
 * Revision     : $Revision: 1.1 $
 * Updated      : $Date: 1999/05/03 15:52:39 $
 */

package hpjava.compiler.wrapper;

import hpjava.compiler.syntaxtree.*;
import hpjava.compiler.visitor.*;

import hpjava.compiler.reflection.*;

import hpjava.compiler.unparser.*;
import hpjava.compiler.abstractexp.*;

public class TypeDescriptor {
  public Type leaf;
  
  public TypeDescriptor (Type leaf) {
    this.leaf = leaf;
  }

  public TypeDescriptor (NodeToken id) {
    Name name = new Name(id, new NodeListOptional());
    NodeChoice choice = new NodeChoice(name, 1);
    Type type = new Type(choice, new NodeListOptional());
    leaf = type;
  }

  public TypeDescriptor (String str) {
    PrimitiveType primitive;
    Name name;

    NodeChoice choice ;
    Type type;


    if (str.equals("boolean")) {
      primitive = new PrimitiveType(new NodeChoice(new NodeToken(str), 0));
      choice = new NodeChoice(primitive, 0);

    } else if (str.equals("char")) {
      primitive = new PrimitiveType(new NodeChoice(new NodeToken(str), 1));
      choice = new NodeChoice(primitive, 0);

    } else if (str.equals("byte")) {
      primitive = new PrimitiveType(new NodeChoice(new NodeToken(str), 2));
      choice = new NodeChoice(primitive, 0);

    } else if (str.equals("short")) {
      primitive = new PrimitiveType(new NodeChoice(new NodeToken(str), 3));
      choice = new NodeChoice(primitive, 0);

    } else if (str.equals("int")) {
      primitive = new PrimitiveType(new NodeChoice(new NodeToken(str), 4));
      choice = new NodeChoice(primitive, 0);

    } else if (str.equals("long")) {
      primitive = new PrimitiveType(new NodeChoice(new NodeToken(str), 5));
      choice = new NodeChoice(primitive, 0);

    } else if (str.equals("float")) {
      primitive = new PrimitiveType(new NodeChoice(new NodeToken(str), 6));
      choice = new NodeChoice(primitive, 0);

    } else if (str.equals("double")) {
      primitive = new PrimitiveType(new NodeChoice(new NodeToken(str), 7));
      choice = new NodeChoice(primitive, 0);
    } else {
      name = new Name(new NodeToken(str), new NodeListOptional());
      choice = new NodeChoice(name, 1);
    }

    type = new Type(choice, new NodeListOptional());
    leaf = type;

  }

  public TypeDescriptor (ClassReflectorDescriptor crd) {
    if (crd.isDistArray()) {
      String cname = crd.getName();
	
      // this is not quite good, change to id concatenation.

      NodeToken id  = new NodeToken(cname);
      Name name = new Name(id, new NodeListOptional());
      NodeChoice choice = new NodeChoice(name, 1);
      Type type = new Type(choice, new NodeListOptional());
      leaf = type;
    }
    else {
      throw new Error("This is used for dist array.");
    }
  }

}

